# Author: Bruce Tesar
# 
# Called from the factorial typology output sheet, this
# generates comparative tableaux for each selected language,
# along with the skeletal basis for each.

require_relative '../lib/excel'
require_relative '../lib/comparative_tableau'
require_relative '../lib/fred'
require_relative '../lib/epo.rb'
require_relative '../lib/moat.rb'
require 'REXML/syncenumerator' # Ruby 1.9 moved syncenumerator to REXML.

# Create a ruby-excel session, and connect it to the existing
# (running) instance of excel.
sess = Excel_session.new
sess.connect_to_excel

# Get a list of selected language labels and associated optimum labels.
lang_list = sess.get_ft_lang_and_opts

# Get the lists of possibly optimal candidates for each competition
competition_list = sess.get_ft_competitions

lang_list.each do |lang|
  lang_label, lang_opt_labels = lang
  # Adjust the optimal candidate flags
  gen = REXML::SyncEnumerator.new(lang_opt_labels, competition_list)
  gen.each do |opt,comp|
    comp.winners.each{|c| c.opt = false}
    (comp.find{|c| c.label == opt}).opt = true
  end  
  # Convert competitions to CT format
  ct = Comparative_tableau.new(lang_label)
  ct.add_competition_list(competition_list)
  # Call FRed on ct
  fred_result = Fred.new(ct)
  # Create new ws, put ct and skb info to it
  sess.put_ft_cand_skb(fred_result)
end

# This call only works on Mink Sum competitions
# sess.put_moat(Moat.new(competition_list[0]))
# sess.put_moat(Moat.get_test_moat())

#apply VBA routine RCDpretty to all RnkRpts
sess.fixRnkRptsAll

# Pop up a button box indicating where RUBOT, when finished, has left you.
sess.display_ok_box('You are now on the first page of the requested ranking info.')


