# Author: Bruce Tesar
# Main project executable.
#
# Runs the standard RUBOT routines on an input violation tableau:
# * canonically formatted Violation Tableau (VT)
# * Comparative Tableau (CT)
# * RCD
# * MIB
# * SKB
# * Residues
 
require_relative '../lib/excel'
require_relative '../lib/comparative_tableau'
require_relative '../lib/rcd'
require_relative '../lib/fred'

# Create a ruby-excel session, and connect it to the existing
# (running) instance of excel.
sess = Excel_session.new
sess.connect_to_excel

# Read in the active worksheet, which is assumed to be
# a properly formatted violation tableau (VT), and store
# the data as a competition list. Then create a new worksheet
# in the active workbook, and write the data as a canonically
# formatted violation tableau.
competition_list = sess.get_input_sheet_data
sess.put_competition_list_as_vt(competition_list)

# Generate a comparative tableau from the competition list,
# and write it to a new worksheet in excel.
ct = Comparative_tableau.new(competition_list.label)
ct.add_competition_list(competition_list)
sess.put_ct_raw(ct)

# Run RCD on the comparative tableau, and put the results
# to a new worksheet in excel.
rcd_result = Rcd.new(ct)
sess.put_rcd_results(rcd_result)

# If RCD determines that the data are inconsistent, pop
# up a message box to that effect. Otherwise, run FRED,
# and put the components of the FRED result to several
# new worksheets.
if !rcd_result.consistent? then
  sess.display_ok_box('Data are inconsistent!')
else
  fred_result = Fred.new(ct)
  sess.put_mib(fred_result)
  sess.put_skb(fred_result)
  sess.put_kept_residues(fred_result)
end

# Pop up a botton box indicating that RUBOT is finished.
sess.display_ok_box('Done!')
