
require_relative 'comparative_tableau'
require_relative 'constraint'
require_relative 'win_lose_pair'
require_relative 'fred'
require_relative 'mojo'
require_relative 'ot_utils'

class CJoin
  
  def initialize(constraint_list=nil)
    @constraints = constraint_list
    @hiers = OT_Utilities.create_all_permutations_as_hierarchies(@constraints)
  end

  def c_join_counts (freds)
    join = Mojo.new(freds)
    join_count = count_consistent_total_orders(join.fred_result.mib)
    union_count = 0
    freds.each {|f| union_count += count_consistent_total_orders(f.mib)}
    
    return join_count, union_count
  end

  def count_consistent_total_orders(ct)
    count = 0
    @hiers.each do |h|
      ct_n = create_ct(h)
      ct_n.concat(ct)
      rcd = Rcd.new(ct_n)
      count += 1 if rcd.consistent?
    end
    return count
  end

  # Assumes hier is a total order
  def create_ct(hier)
    ct = Comparative_tableau.new(@constraints)
    for i in 0..hier.length-2
      cand1 = Candidate.new("a", "out1", true, @constraints)
      cand2 = Candidate.new("a", "out2", false, @constraints)
      
      @constraints.each {|c| cand1.set_viols(c, 0); cand2.set_viols(c, 0)}
      cand1.set_viols(hier[i+1][0], 1)
      cand2.set_viols(hier[i][0], 1)
      
      wlp = Win_lose_pair.new(cand1, cand2)
      ct << wlp
    end
    return ct
  end

end

=begin
c1 = Constraint.new("ML", ":")
c2 = Constraint.new("MR", ":")
c3 = Constraint.new("WSP", ":")
c4 = Constraint.new("NoL", ":")
c5 = Constraint.new("IdS", ":")
c6 = Constraint.new("IdL", ":")

cons = [c1, c2, c3, c4, c5, c6]
#cons = [c1, c2, c3]

# Need candidates
# L3    LLLLWL, WLLeee, eeLLeW
ca1 = Candidate.new("in1", "out1", true, cons)
ca2 = Candidate.new("in1", "out2", false, cons)
ca3 = Candidate.new("in2", "out1", true, cons)
ca4 = Candidate.new("in2", "out2", false, cons)
ca5 = Candidate.new("in3", "out1", true, cons)
ca6 = Candidate.new("in3", "out2", false, cons)

ca11 = Candidate.new("in3", "out2", true, cons)
ca12 = Candidate.new("in3", "out2", false, cons)
ca13 = Candidate.new("in3", "out2", true, cons)
ca14 = Candidate.new("in3", "out2", false, cons)

cons.each do |c| 
  ca1.set_viols(c, 0)
  ca2.set_viols(c, 0)
  ca3.set_viols(c, 0)
  ca4.set_viols(c, 0)
  ca5.set_viols(c, 0)
  ca6.set_viols(c, 0)
  
  ca11.set_viols(c, 0)
  ca12.set_viols(c, 0)
  ca13.set_viols(c, 0)
  ca14.set_viols(c, 0)
end
  
# LLLLWL
ca1.set_viols(c1, 1)
ca1.set_viols(c2, 1)
ca1.set_viols(c3, 1)
ca1.set_viols(c4, 1)
ca1.set_viols(c5, 0); ca2.set_viols(c5, 1)
ca1.set_viols(c6, 1)

# WLLeee
ca3.set_viols(c1, 0); ca4.set_viols(c1, 1)
ca3.set_viols(c2, 1)
ca3.set_viols(c3, 1)

# eeLLeW
ca5.set_viols(c3, 1)
ca5.set_viols(c4, 1)
ca5.set_viols(c6, 0); ca6.set_viols(c6, 1)


#L11 WLLLLL
#    eWeLLL

# WLLLLL
ca11.set_viols(c1, 0); ca12.set_viols(c1, 1)
ca11.set_viols(c2, 1)
ca11.set_viols(c3, 1)
ca11.set_viols(c4, 1)
ca11.set_viols(c5, 1)
ca11.set_viols(c6, 1)

# eWeLLL
ca13.set_viols(c2, 0); ca14.set_viols(c2, 1)
ca13.set_viols(c4, 1)
ca13.set_viols(c5, 1)
ca13.set_viols(c6, 1)

# Need winner-loser pairs
wlp1 = Win_lose_pair.new(ca1, ca2)
wlp2 = Win_lose_pair.new(ca3, ca4)
wlp3 = Win_lose_pair.new(ca5, ca6)

wlp11 = Win_lose_pair.new(ca11, ca12)
wlp12 = Win_lose_pair.new(ca13, ca14)

# Need comp tableau
ct1 = Comparative_tableau.new(cons)
ct2 = Comparative_tableau.new(cons)
ct1 << wlp1 << wlp2 << wlp3
ct2 << wlp11 << wlp12

# Need Fred on ct
fred1 = Fred.new(ct1)
fred2 = Fred.new(ct2)
freds = Array.new
freds << fred1 << fred2

mojo = Mojo.new(freds)
fredj = mojo.fred_result
mib1 = fred1.mib
mib2 = fred2.mib
mib3 = fredj.mib
cj = CJoin.new(cons)

puts "Fred result \n#{mib1.to_s}"
puts "Fred result \n#{mib2.to_s}"
puts "Fred result \n#{mib3.to_s}"

puts "Count of consistent (L3): \t#{cj.count_consistent_total_orders(mib1)}"
puts "Count of consistent (L11): \t#{cj.count_consistent_total_orders(mib2)}"
puts "Count of consistent (join): \t#{cj.count_consistent_total_orders(mib3)}"


=end


