# Author: Nazarre Merchant

require_relative 'candidate'

# An HS Candidate differs from its OT brethren by having ref to its constructing
# operation, so to allow opvec creation.

class HS_Candidate < Candidate

  def initialize(input, output, optimal, constraints, operation)
    super(input, output, optimal, constraints)
    @operation = operation
    @base = false
  end

  # Is this candidate the base candidate (for opvec construction purposes) in its candidate set?
  def base?
    return @base
  end
  
  def output_as_input
    return @output
  end
  
  def base=(val)
    @base = val
  end
  
  def operation
    @operation
  end

  # Returns a copy of the candidate, containing duplicates of the
  # input, the output, label, remark, and the list of merged candidates
  # (the merged candidates themselves are not duplicated).
  # The copy candidate also gets a duplicate of the constraint violations.
  def dup
    copy = HS_Candidate.new(@input.dup, @output.dup, @opt, @constraints, @operation)
    copy.base=@base
    @constraints.each{|con| copy.set_viols(con,get_viols(con))}
    copy.label = @label.dup unless @label.nil? # cannot call nil.dup()
    copy.remark = @remark.dup unless @remark.nil?
    # set @merged of the copy to be the same as @merged of the current candidate.
    copy.instance_variable_set(:@merged, @merged)
    # set @merged_candidates of the copy to be A DUPLICATE of
    # @merged_candidates of the current candidate.
    copy.instance_variable_set(:@merge_candidates, @merge_candidates.dup)
    return copy
  end

end # class HS_Candidate
