# Author: Bruce Tesar
#

# Stores the candidates of a competition (i.e., candidates
# that have the same input).
class Competition < Array

  # Returns an empty competition.
  def initialize
  end

  # Returns true if at least one candidate is marked as optimal;
  # false otherwise.
  def optima?
    (find_all{|c| c.opt?}).size > 0
  end

  # Returns true if more than one candidate is marked as optimal;
  # false otherwise.
  def mult_optima?
    (find_all{|c| c.opt?}).size > 1
  end

  # Returns a list of the candidates marked as optimal.
  def winners
    find_all{|c| c.opt?}
  end

  # Returns a list of the candidates marked as not optimal.
  def losers
    find_all{|c| !c.opt?}
  end

  # Returns a reference to the constraint list of the candidates. Returns
  # an empty array if the competition is empty (contains no candidates).
  #--
  # The idea is to have a single constraint list object shared by all other
  # objects in the system (more efficient).
  #++
  def constraint_list
    return [] if empty?
    return self[0].constraint_list
  end
  
  # Identify sets of candidates with identical violation profiles.
  # Each set with more than one candidate is merged together: the first
  # candidate is duplicated (leaving the original unaffected),
  # and the other violation-matching candidates are added to the duplicate
  # as merge candidates. The merged duplicate is then added to the revised list.
  # At the end, the competition contains one candidate
  # for each distinct violation profile of the original competition.
  def merge_identical_candidates
    nonmatch = self.dup
    self.clear
    while !nonmatch.empty? do
      match, nonmatch = nonmatch.partition{|cand| nonmatch.first.ident_viols?(cand)}
      merge_cand = match.shift.dup # remove first candidate, duplicate it.
      match.each {|c| merge_cand.add_merge_candidate(c)} # merge the matching candidates.
      self << merge_cand
    end
    self
  end
  
  # NNM: need dup to duplicate the candidates it contains for HS work.
  def dup
    dup_self = Competition.new
    self.each {|c| dup_self << c.dup}
    return dup_self
  end

  # Returns a string consisting of the to_s() for each candidate, separated
  # by newlines, and terminated by a newline.
  def to_s
    "#{join("\n")}\n"
  end

end # class Competition
