# Author: Bruce Tesar
#

# Stores a list of competitions comprising a dataset. 
class Competition_list < Array
  attr_accessor :label
  def initialize(label = "NoLabel")
    @label = label
  end

  # Returns a list of the constraints.
  def constraint_list
    return [] if empty?
    return self[0].constraint_list    
  end

  # Auto-numbers the candidates of each competition.
  # The numbering takes the form of {competition_#}.{candidate_#},
  # where both competition and candidate numbering start from 1.
  #
  # The first candidate of the first competition is numbered 1.1;
  # The fifth candidate of the ninth competition is numbered 9.5.
  def auto_number_candidates
    comp_number = '0'
    self.each do |competition|
      comp_number = comp_number.succ
      cand_number = '0'
      competition.each do |cand|
        cand_number = cand_number.succ
        cand.label = "#{comp_number}.#{cand_number}"
      end
    end
    return true
  end

  # Calls Competition#merge_identical_candidates() for each competition in
  # the list, thus merging (within each competition) candidates with
  # identical violation profiles.
  def merge_identical_candidates
    self.each{|comp| comp.merge_identical_candidates}
  end

end # class Competition_list
