#!/usr/bin/env ruby -w

def corr_origin(input0, output0, epenthesis=nil)
  # raise "Error: input.size>output.size. ['#{input0}', '#{output0}']" if input0.size>output0.size
  epenthesis ||= ['e', 'p']
  replaces = Hash.new('?').update({ 'X'=>'', 'e'=>'V', 'p'=>'C' })
  
  o2_sym = []
  o2_matrix = []
  i1 = i2 = 1
  input = ' ' + input0
  output = ' ' + output0
  while input[i1] != nil || output[i2] != nil
    if output[i2] == 'X'
      output[i2] = replaces[output[i2]]
      o2_matrix << "#{i1}:_"
      i1 += 1
      # i2 not change
    elsif epenthesis.include?(output[i2])
      o2_sym << output[i2]
      o2_matrix << "_:#{i2}"
      i2 += 1
    else
      if input[i1] == nil && output[i2] == nil
        i2 = output.size # end loop
      elsif input[i1] == nil
        o2_matrix << "_:#{i2}"
      elsif output[i2] == nil
        o2_matrix << "#{i1}:_"
      elsif input[i1] == output[i2]
        o2_sym << output[i2]
        o2_matrix << "#{i1}:#{i2}"
      else
        o2_sym << output[i2] if output0[i2] != 'X'
        o2_matrix << "#{i1}:#{i2}"
      end
      i1 += 1
      i2 += 1
    end
  end
  
  "#{o2_sym.join} #{o2_matrix.join(', ')}".strip
end

def corr(input, output, epenthesis=nil)
  epenthesis ||= []
  o2_matrix = []
  i22 = 0
  output.split('$').each_with_index do |oo, i1|
    i2 = 0
    while oo[i2] != nil
      if oo[i2] != 'X'
        if epenthesis.include?(oo[i2])
          o2_matrix << "_:#{i22+i2+1}"
        else
          o2_matrix << "#{i1+1}:#{i22+i2+1}"
        end
      else
        o2_matrix << "#{i1+1}:_"
        # o2_matrix << "_:#{i22+i2+1}"
        oo[i2] = ''
      end
      i2 += 1
    end
    i22 += oo.size
  end
  "#{output.gsub('$','').gsub('X','')} #{o2_matrix.join(', ')}".strip
end

def test
  p '------------- test -------------'
  DATA.each_line do |line|
    line.strip!
    next if line.empty?
    next if line[0] == '#'
  
    input, output, expect = line.split('-').map(&:strip)
    v1 = corr(input, output)
    if v1 == expect
      puts "OK     corr(#{input}\t, #{output}\t) == #{expect}"
    else
      puts "Error  corr(#{input}\t, #{output}\t) == #{v1}\t   expect: #{expect}"
    end
  end
end

def test_has_epenthesis
  data = <<-EOS
    CV    -    C$V       -    CV 1:1, 2:2
    CV    -    C$X       -    C 1:1, 2:_
    CV    -    Cp$V      -    CpV 1:1, _:2, 2:3
    CV    -    X$V       -    V 1:_, 2:1
    CV    -    X$X       -    1:_, 2:_
    CV    -    X$pV      -    pV 1:_, _:1, 2:2
    CV    -    Ce$V      -    CeV 1:1, _:2, 2:3
    CV    -    Ce$X      -    Ce 1:1, _:2, 2:_
    CV    -    Ce$pV     -    CepV 1:1, _:2, _:3, 2:4
    EOS

  p '------------- test have epenthesis param -------------'
  data.each_line do |line|
    line.strip!
    next if line.empty?
    next if line[0] == '#'
  
    input, output, expect = line.split('-').map(&:strip)
    v1 = corr(input, output, ['e', 'p'])
    if v1 == expect
      puts "OK     corr(#{input}\t, #{output}\t) == #{expect}"
    else
      puts "Error  corr(#{input}\t, #{output}\t) == #{v1}\t   expect: #{expect}"
    end
  end
end

if __FILE__ == $0
  test
  test_has_epenthesis
end

__END__
#  INPUT     OUTPUT         EXPECT VALUE
    # CV    -    CV       -    CV 1:1, 2:2
    # CV    -    CX       -    C 1:1, 2:_
    # CV    -    CpV      -    CpV 1:1, _:2, 2:3
    # CV    -    XV       -    V 1:_, 2:1
    # CV    -    XX       -    1:_, 2:_
    # CV    -    XpV      -    pV 1:_, _:1, 2:2
    # CV    -    CeV      -    CeV 1:1, _:2, 2:3
    # CV    -    CeX      -    Ce 1:1, _:2, 2:_
    # CV    -    CepV     -    CepV 1:1, _:2, _:3, 2:4

    CV    -    C$V       -    CV 1:1, 2:2
    CV    -    C$X       -    C 1:1, 2:_
    CV    -    C$pV      -    CpV 1:1, 2:2, 2:3
    CV    -    X$V       -    V 1:_, 2:1
    CV    -    X$X       -    1:_, 2:_
    CV    -    X$pV      -    pV 1:_, 2:1, 2:2
    CV    -    Ce$V      -    CeV 1:1, 1:2, 2:3
    CV    -    Ce$X      -    Ce 1:1, 1:2, 2:_
    CV    -    Ce$pV     -    CepV 1:1, 1:2, 2:3, 2:4

    BD - t$d  -  td 1:1, 2:2
    BD - d$d  -  dd 1:1, 2:2
    BD - d$t  -  dt 1:1, 2:2
    BD - t$t  -  tt 1:1, 2:2

    BD -  t$t     - tt 1:1, 2:2
    BD -  t$da    - tda 1:1, 2:2, 2:3
    BD -  da$t    - dat 1:1, 1:2, 2:3
    BD -  da$da   - dada 1:1, 1:2, 2:3, 2:4

    tdd - t$d$d   - tdd 1:1, 2:2, 3:3
    tdd - t$da$d  - tdad 1:1, 2:2, 2:3, 3:4
    tdd - t$da$da - tdada 1:1, 2:2, 2:3, 3:4, 3:5
    tdd - t$d$da  - tdda 1:1, 2:2, 3:3, 3:4
