# Author: Naz Merchant
#

# A derivation is an input and a set of paths.

require_relative 'path.rb'
require_relative 'operation.rb'
require_relative 'hierarchy.rb'
require_relative 'competition.rb'
require_relative 'competition_list.rb'
require_relative 'comparative_tableau.rb'
require_relative 'opvec.rb'
require_relative 'rcd.rb'

class Derivation

    attr_reader :input, :paths, :constraints

    def initialize(input)
      @input = input
      @paths = Array.new
      @constraints = nil
      
      #@hier = hier
      #@ops = ops
      #p = Path.new
      #@paths << p
      #derive_paths(p, @input)
    end
    
    def add_path(path)
      @paths << path
      @constraints = path.get_competition_list.constraint_list if @constraints == nil
    end
    
    def contains_path?(path)
      return @paths.include?(path)
    end
    
    def dup
      dderiv = Derivation.new(@input)
      @paths.each {|p| dderiv.add_path(p)}
      return dderiv
    end

=begin
    def derive_paths(path, input)
      cand_set = Competition.new
      @ops.each{|op| op.create_candidates(input).each{|cand| cand_set << cand} }
      return if cand_set.length == 0
      
      # Mod to remove HB'ed candidates
      cand_set = remove_harmonically_bounded(cand_set)
      
      winners = @hier.winners(cand_set)
      orig_path = path.dup
      new_path = nil
      
      winners.each do |w|
        if new_path == nil
          new_path = path
        else
          new_path = orig_path.dup
          @paths << new_path
        end
        w.opt=true
        new_path << cand_set.dup
        #So that there is only one optimum for cand sets in paths
        w.opt=false
        derive_paths(new_path, w.output_as_input) unless w.is_faithful?
      end  
    end
    
    def remove_harmonically_bounded(cand_set)
      hb_free_cand_set = Competition.new
      cand_set.each do |cand|
        ct = Comparative_tableau.new("check_with_rcd", cand.constraint_list)
        winner_comp = construct_competition_with_winner(cand, cand_set)
        # check to see if winner is possibly optimal
        ct.add_competition(winner_comp)
        rcd_result = Rcd.new(ct)
        # If the CT is inconsistent, then the candidate is harmonically bound
        hb_free_cand_set << cand if rcd_result.consistent?
      end
      
      return hb_free_cand_set
    end
    
    # method below is code taken from BT's factorial typology (now defunct)
    
    # Returns a copy of the competition _competition_, altered so that the
    # candidate <em>candidate_winner</em> is the winner.
    # A duplicate is made of each of the candidates so that the altered
    # opt field doesn't affect objects referenced elsewhere.
    def construct_competition_with_winner(candidate_winner, competition)
      winner = candidate_winner.dup
      losers = competition.reject{|c| c==candidate_winner}
      losers = losers.map{|c| c.dup}
      winner.opt=true; losers.each{|c| c.opt=false}
      new_comp = Competition.new
      new_comp.push(winner)
      losers.each{|loser| new_comp.push(loser)}
      return new_comp
    end
=end

    def is_consistent?(path)
      ct = get_comparative_tableau
      ct.concat(path.get_comparative_tableau)
      rcd = Rcd.new(ct)
      rcd.consistent? ? true : false
    end
    
    def derivation_consistent?(deriv)
      ct = get_comparative_tableau
      ct.concat(deriv.get_comparative_tableau)
      rcd = Rcd.new(ct)
      rcd.consistent? ? true : false
    end
     
    def get_competition_list
      comp_list = Competition_list.new
      @paths.each {|p| p.each{|c| comp_list << c}}
      return comp_list
    end
    
    def get_opvecs
      vecs = Array.new
      get_competition_list.each do |comp|
        base = get_base(comp)
        #NB: this doesn't produce the order the derivation imposes on the opvecs
        comp.each {|cand| vecs << OpVec.new(cand, base, cand.operation, cand.constraint_list) }
      end
      return vecs
    end
    
    def get_base(competition)
      base = nil
      competition.each {|cand| base = cand if cand.base?}
      # Arb. select first candidate for the base if none are marked (faith cand is default marked)
      base = competition[0] if base.nil?
      base.base=true
      return base
    end
    
    def get_comparative_tableau
      comp_list = get_competition_list
      ct = Comparative_tableau.new("", comp_list.constraint_list)
      ct.add_competition_list(comp_list)
      return ct
    end

    # Returns a set of outputs that are the final outputs of each path
    def get_final_optima
      outs = Array.new
      @paths.each {|p| outs << p.get_final_optima}
      return outs
    end

    def contains_path?(path)
      @paths.each {|p| return true if p == path} 
      false
    end

    # Two derivations are equivalent if their paths are the same
    def ==(other)
      return false if @input != other.input
      return false if @paths.length != other.paths.length
      @paths.each {|p| return false unless other.contains_path?(p)}
      return true
    end
  
    # The same as ==
    def eql?(other)
      self == other
    end

    def to_s
      s = "Deriv Input: \t#{input}\n"
      @paths.each {|p| s += "\t\t\t#{p.to_s}\n"}
      return s
    end

end
