# Author: Naz Merchant
#

# A derivation list is a consistent set of derivations

require_relative 'path.rb'
require_relative 'derivation.rb'
require_relative 'operation.rb'
require_relative 'hierarchy.rb'
require_relative 'competition.rb'
require_relative 'competition_list.rb'
require_relative 'comparative_tableau.rb'
require_relative 'opvec.rb'
require_relative 'rcd.rb'

class Derivation_Hash < Hash
    
  # Assumes that the order of constraints in cons is the desired total order
  def create_derivations_for_total_order(inputs, ops, cons)
    @cons = cons
    hier = create_total_order(@cons)
    inputs.reverse.each do |inp|  
      paths = Array.new
      Path.spawn_paths(inp, paths, ops, hier)
      dv = Derivation.new(inp)
      paths.each {|p| dv.add_path(p)}
      self[inp] = dv
    end  
  end
  
  def create_total_order(cons)
    hier = Hierarchy.new
    cons.each {|c| hier << [c]}
    return hier
  end
    
  def get_comparative_tableau    
    ct = Comparative_tableau.new("", @cons)
    self.values.each {|d| d.get_competition_list.each {|c| ct.add_competition(c)}}
    return ct
  end
    
end
