# Author: Bruce Tesar
# Other: N. Merchant, A. Prince
# The Excel interface for RUBOT.
#As of May 2013, puts MOAT page; and MIB onto rnkrpt page

require 'win32ole' # need this to use OLE automation
require 'dl' # Windows API access (for messagebox)
require_relative 'c_join_check'
require_relative 'constraint'
require_relative 'candidate'
require_relative 'competition'
require_relative 'competition_list'
require_relative 'comparative_tableau'
require_relative 'skb_expansion'
require_relative 'ranking_diagram'
require_relative 'epo_diagram'
require_relative 'erc'
require_relative 'constraint_hs'
require_relative 'operation'
require_relative 'operation_faith'
require_relative 'opvec_po'
require_relative 'opvec_po_diagram'

require_relative '../usr_lib/merchant/constraint_bst_hs.rb'
require_relative '../usr_lib/merchant/operation_bst.rb'
require_relative '../usr_lib/merchant/operation_bst_faith.rb'
require_relative '../usr_lib/merchant/operation_bee_faith.rb'
require_relative '../usr_lib/merchant/operation_bst_epen.rb'
require_relative '../usr_lib/merchant/bst_parser.rb'
require_relative '../usr_lib/merchant/bee_parser.rb'
require_relative '../usr_lib/merchant/est1_parser.rb'


require_relative '../usr_lib/constraints/constraint_generator.rb'

require 'REXML/syncenumerator' # Ruby 1.9 moved syncenumerator to REXML.

# A session for interacting with excel. Once an Excel_session object
# is created, it can connect to an already running instance of Excel
# via connect_to_excel(), or it can launch a new excel session via
# start_excel(). Other methods of the class provide RUBOT-specific
# ways of reading from and writing to Excel.
#
# Methods with names beginning with "put_" create one or more new
# worksheets, and fully determine the content of those worksheets.
# Methods with names ending with "_to_ws" add content to existing
# worksheets; the worksheet to be written to and the starting location
# to write to on the sheet are passed in as parameters.
class Excel_session
  # Excel Constants
  # Load built-in Excel constants (Xl*); for a list, see
  # http://msdn.microsoft.com/en-us/library/aa221100(office.11).aspx
  # The constants of the variety Xl* can be obtained from WIN32OLE.
  # const_load(<excel object>, Excel_session) obtains the
  # defined Excel constaints and makes them constants of the class Excel_session.
  # That way, they can be used in the methods of this class, which makes it
  # much easier to use the MSDN Reference Library for the Excel VBA calls.
  # Reference: http://msdn.microsoft.com/en-us/library/aa220733(office.11).aspx
  begin
    WIN32OLE.const_load(WIN32OLE.connect("excel.application"), Excel_session) # Load excel constants
  rescue WIN32OLERuntimeError # error exception thrown if Excel isn't running.
    # Create a temporary excel app, so that constants can be loaded from it.
    excel_temp = WIN32OLE.new("excel.application")
    WIN32OLE.const_load(excel_temp, Excel_session) # Load excel constants
    excel_temp.ole_free # Terminate the temporary excel app
  end
  
  #--
  # The color index constants aren't defined as colors in Excel, because they
  # actually point to positions in the default color palette, which can be
  # dynamically changed by the user. The color assignments given below
  # as constants reflect the Excel defaults: the colors that by default are
  # in the numbered positions in the color palette.
  # Color index constant usage: obj.colorindex = CONST
  RED = 3 #:nodoc:
  BRIGHTGREEN = 4 #:nodoc:
  BLUE = 5 #:nodoc:
  PALEYELLOW = 19 #:nodoc:
  BRIGHTCYAN = 28 #:nodoc:
  LIGHTCYAN = 34 #:nodoc:
  LIGHTGREEN = 35 #:nodoc:
  MIDYELLOW = 36 #:nodoc:
  MAGENTA = 38 #:nodoc:
  GOLD = 44 #:nodoc:

  # Nothing "instance-ish" to initialize.
  def initialize
  end
  
  # Connects to a running Excel server.
  def connect_to_excel
    @excel = WIN32OLE.connect("excel.application") # connects to an already running Excel
  rescue WIN32OLERuntimeError # error exception thrown if Excel isn't running.
    puts "RuntimeError: Excel probably isn't running." # print an error message (in English) to Console
    display_ok_box("RuntimeError: Excel probably isn't running.")
    exit
  end
  
  # Starts a new excel session with a fresh workbook.
  def start_excel
    @excel = WIN32OLE.new("excel.application")
    # Load the RUBOT excel add-in (not automatically loaded when a new
    # excel session is started via OLE).
    # Loadpath uses the RUBOT environment variable, which points to Rubot\bin,
    # and looks in Rubot\vb.
#    wb = @excel.workbooks.open("#{ENV['RUBOT']}\\..\\vb\\Rubot.xla")
    # Run the automacros in the add-in.
#    wb.runautomacros 1
    add_output_workbook('New')
    @excel.visible=true
  end
    
  # Display a simple pop-up window with an OK button and
  # the message _msg_.
  # This uses a standard Windows message window via the Windows API.
  def display_ok_box(msg)
    message = msg
    title = 'RUBOT'
    buttons = 0 # 0 indicates that an OK button should be displayed.
    user32 = DL.dlopen('user32')
    # The DL library differs between Ruby 1.8 and Ruby 1.9, so different
    # versions of the messagebox code are provided for each version.
    # The code was obtained from posts on the Ruby forum (www.ruby-forum.com),
    # which attribute the 1.9 code to the Ruby development team.
    # TODO: either eliminate the 1.8 code, or relocate the key differing
    #       code to a method in the Compatibility1_9 project.
    if RUBY_VERSION =~ /^1\.9/ then
      # version for Ruby 1.9 version of DL.
      msgbox = DL::CFunc.new(user32['MessageBoxA'], DL::TYPE_LONG, 'MessageBox')
      r, rs = msgbox.call([0, message, title, buttons].pack('L!ppL!').unpack('L!*'))
      return r
    else
      # version for Ruby 1.8 version of DL.
      msgbox = user32['MessageBoxA', 'ILSSI']
      r, rs = msgbox.call(0, message, title, buttons)
      return r
    end
  end
  
  # Display a simple pop-up window with Yes and No buttons and
  # the message _msg_.
  # This uses a standard Windows message window via the Windows API.
  def display_yesno_box(msg)
    message = msg
    title = 'RUBOT'
    buttons = 4 # 4 indicates that yes/no buttons should be displayed.
    user32 = DL.dlopen('user32')
    # The DL library differs between Ruby 1.8 and Ruby 1.9, so different
    # versions of the messagebox code are provided for each version.
    # The code was obtained from posts on the Ruby forum (www.ruby-forum.com),
    # which attribute the 1.9 code to the Ruby development team.
    # TODO: either eliminate the 1.8 code, or relocate the key differing
    #       code to a method in the Compatibility1_9 project.
    if RUBY_VERSION =~ /^1\.9/ then
      # version for Ruby 1.9 version of DL.
      msgbox = DL::CFunc.new(user32['MessageBoxA'], DL::TYPE_LONG, 'MessageBox')
      r, rs = msgbox.call([0, message, title, buttons].pack('L!ppL!').unpack('L!*'))
    else
      # version for Ruby 1.8 version of DL.
      msgbox = user32['MessageBoxA', 'ILSSI']
      r, rs = msgbox.call(0, message, title, buttons)
    end
    
    return true if r == 6 # 6 is code for clicked yes
    return false
  end
  
  # Gets raw violation tableau data (candidates and violation counts)
  # from the currently active worksheet,
  # converts it to a Competition_list, and returns that list.
  def get_input_sheet_data
    # create an object for the currently active worksheet
    input_sheet = @excel.activeworkbook.activesheet
    # If the first cell is empty, put a default name in that cell.
    # This will ensure that the first row is included in the usedrange.  -AP
    first_cell = input_sheet.range("a1")
    first_cell.value ||= "Nameless"
    @proj_name = first_cell.value
    # Load the usedrange of the input sheet into a 2-dim array.
    raw_input = input_sheet.usedrange.value
    # delete empty rows
    raw_input = raw_input[0..-1].select{|row| row.find{|e|e}}
    # Convert the raw input data to a competition list
    return convert_input_data(raw_input)
  end

  # Gets operations, inputs, and constraints for a HS system
  def get_hs_op_input_sheet_data
    # create an object for the currently active worksheet
    input_sheet = @excel.activeworkbook.activesheet
    # If the first cell is empty, put a default name in that cell.
    # This will ensure that the first row is included in the usedrange.  -AP
    first_cell = input_sheet.range("a1")
    first_cell.value ||= "Nameless"
    @proj_name = first_cell.value
    # Load the usedrange of the input sheet into a 2-dim array.
    raw_input = input_sheet.usedrange.value
    # delete empty rows
    raw_input = raw_input[0..-1].select{|row| row.find{|e|e}}
    # Convert the raw input data to a competition list
    return convert_hs_input_data(raw_input)
  end
  
  
  # Converts the raw data from the worksheet, raw_input,
  # into three arrays, an input, op, and constraint array.
  # TODO: more actual input verification: are the format assumptions obeyed?
  def convert_hs_input_data(raw_input)
    # The label for the datasheet
    data_sheet_label = raw_input[0][0]
    # NB: sheet numbering starts at *1*, BUT raw_input numbering starts at *0*
    op_col = 0
    input_col = 4
    con_col = 6
    user_col = 9
    
    con_first_row = 2
    con_last_row = con_first_row-1
    # Use 'or' short-circuit eval, otherwise raw_input[x] not defined for x outside of range.
    con_last_row += 1 until (con_last_row == raw_input.size-1) or raw_input[con_last_row+1][con_col].nil?
    
    input_first_row = 2 
    input_last_row = input_first_row-1
    input_last_row += 1 until (input_last_row == raw_input.size-1) or raw_input[input_last_row+1][input_col].nil?
    
    op_first_row = 2
    op_last_row = op_first_row - 1
    op_last_row += 1 until (op_last_row == raw_input.size-1) or raw_input[op_last_row][op_col] == "Faith"
   
    con_range = (con_first_row..con_last_row)
    input_range = (input_first_row..input_last_row)
    op_range = (op_first_row..op_last_row)
    
    #Operations
    #Inputs
    #Constraints

    return convert_hs_parser_data(raw_input, con_range, input_range, op_range) if !raw_input[1][user_col].nil?
    
    constraints = Array.new
    con_gen = Constraint_Generator.new()
    con_range.each {|row| constraints << con_gen.gen_constraint(raw_input[row][con_col], "", raw_input[row][con_col+1])}
    
    #con_range.each {|col| constraints << Constraint_HS.new(raw_input[1][col], "", Regexp.new(raw_input[0][col]))}
    
    inputs = Array.new
    input_range.each {|row| inputs << raw_input[row][input_col]}
      
    ops = Array.new
    op_range.each do |row|
      op_name = nil
      op_name = raw_input[row][op_col] unless raw_input[row][op_col] == nil
      if op_name == nil
        # do nothing, proceed to next row 
      else
        # find end of operation definition
        if op_name == "Faith"
          ops << Faithful_Operation.new("Faith", [""], [""], constraints)
        else
          op_end = row
          op_end += 1 until (raw_input[op_end][op_col] != nil && raw_input[op_end][op_col] != op_name)
          op_end -= 1
  
          curr_op_range = (row..op_end)
          source = Array.new
          target = Array.new
          
          puts "Current op range: " + curr_op_range.to_s
          
          curr_op_range.each do |r2|
            source << raw_input[r2][op_col+1]
            raw_input[r2][op_col+2] == nil ? target << "" : target << raw_input[r2][op_col+2] 
          end
          ops << Operation.new(op_name, source, target, constraints)
        end
      end
    end

    return inputs, ops, constraints
  end
  
  def convert_hs_parser_data(raw_input, con_range, input_range, op_range)
    op_col = 0
    input_col = 4
    con_col = 6
    user_col = 9
    
    # Everything uses the parser, so build it first
    parser = build_bst_parser(raw_input, user_col) if raw_input[1][user_col] == "BST"
    parser = build_bee_parser(raw_input, user_col) if raw_input[1][user_col] == "BEE"
    parser = build_est1_parser(raw_input, user_col) if raw_input[1][user_col] == "EST1"
    
    # Now put together inputs -- same for any parser
    inputs = Array.new
    input_range.each {|row| inputs << raw_input[row][input_col]}
         
    # Constraints -- def varies by parser
    constraints = build_bst_constraints(raw_input, parser, con_range, con_col) if raw_input[1][user_col] == "BST" 
    constraints = build_bee_constraints(raw_input, parser, con_range, con_col) if raw_input[1][user_col] == "BEE" 
    constraints = build_est1_constraints(raw_input, parser, con_range, con_col) if raw_input[1][user_col] == "EST1" 
      
    # Ops
    ops = build_bst_ops(raw_input, parser, op_range, op_col, constraints) if raw_input[1][user_col] == "BST" 
    ops = build_bee_ops(raw_input, parser, op_range, op_col, constraints) if raw_input[1][user_col] == "BEE" 
    ops = build_est1_ops(raw_input, parser, op_range, op_col, constraints) if raw_input[1][user_col] == "EST1" 
    return inputs, ops, constraints   
  end

  def build_bst_parser(raw_input, user_col)
    nuc_row_beg = 3
    nuc_row_end = nuc_row_beg-1
    nuc_row_end += 1 until raw_input[nuc_row_end+1][user_col] == "Margins"
    nuc_range = (nuc_row_beg..nuc_row_end)
    
    mar_row_beg = nuc_row_end+2
    mar_row_end = mar_row_beg-1
    
    # Use 'or' short-circuit eval, otherwise raw_input[x] not defined for x outside of range.
    mar_row_end += 1 until (mar_row_end == raw_input.size-1) or raw_input[mar_row_end+1][user_col].nil?
    mar_range = (mar_row_beg..mar_row_end)
       
    nucs = Array.new
    nucs_r = Array.new
    mars = Array.new
    mars_r = Array.new
    
    (nuc_range).each {|r| nucs << raw_input[r][user_col]; nucs_r << raw_input[r][user_col+1]}
    (mar_range).each {|r| mars << raw_input[r][user_col]; mars_r << raw_input[r][user_col+1]} 
    
    return BST_Parser.new(nucs, mars, nucs_r, mars_r)
  end
  
  def build_est1_parser(raw_input, user_col)
    nuc_row_beg = 3
    nuc_row_end = nuc_row_beg-1
    nuc_row_end += 1 until raw_input[nuc_row_end+1][user_col] == "Margins"
    nuc_range = (nuc_row_beg..nuc_row_end)
    
    mar_row_beg = nuc_row_end+2
    mar_row_end = mar_row_beg-1
    
    mar_row_end += 1 until raw_input[mar_row_end+1][user_col] == "Unsyll"
    mar_range = (mar_row_beg..mar_row_end)
    
    unsyl_row_beg = mar_row_end+2
    unsyl_row_end = unsyl_row_beg-1
    
    # Use 'or' short-circuit eval, otherwise raw_input[x] not defined for x outside of range.
    unsyl_row_end += 1 until (unsyl_row_end == raw_input.size-1) or raw_input[unsyl_row_end+1][user_col].nil?
    unsyl_range = (unsyl_row_beg..unsyl_row_end)
       
    nucs = Array.new
    nucs_r = Array.new
    mars = Array.new
    mars_r = Array.new
    unsyl = Array.new
    unsyl_r = Array.new
    
    (nuc_range).each {|r| nucs << raw_input[r][user_col]; nucs_r << raw_input[r][user_col+1]}
    (mar_range).each {|r| mars << raw_input[r][user_col]; mars_r << raw_input[r][user_col+1]} 
    (unsyl_range).each {|r| unsyl << raw_input[r][user_col]; unsyl_r << raw_input[r][user_col+1]} 
    
    return EST1_Parser.new(nucs, mars, unsyl, nucs_r, mars_r, unsyl_r)
  end
  
  def build_bee_parser(raw_input, user_col)
    seg_row_beg = 2
    seg_row_end = seg_row_beg-1
    seg_row_end += 1 until (seg_row_end == raw_input.size-1) or raw_input[seg_row_end+1][user_col].nil?
    seg_range = (seg_row_beg..seg_row_end)

    segs = Array.new
    segs_r = Array.new
    
    (seg_range).each {|r| segs << raw_input[r][user_col]; segs_r << raw_input[r][user_col+1]}
    
    return BEE_Parser.new(segs, segs_r)
  end
  
  def build_bst_constraints(raw_input, parser, con_range, con_col)
    constraints = Array.new
    con_range.each do |row|   
      f_con = raw_input[row][con_col].start_with?("f.")
      constraints << Constraint_BST_HS.new(raw_input[row][con_col], "", Regexp.new(raw_input[row][con_col+1]), parser, f_con)
    end
    return constraints
  end
  
  def build_est1_constraints(raw_input, parser, con_range, con_col)
    return build_bst_constraints(raw_input, parser, con_range, con_col)
  end
  
  
  def build_bee_constraints(raw_input, parser, con_range, con_col)
    constraints = Array.new
    con_gen = Constraint_Generator.new()
    con_range.each {|row| constraints << con_gen.gen_constraint(raw_input[row][con_col], "", raw_input[row][con_col+1])}
    return constraints
  end
  
  def build_bst_ops(raw_input, parser, op_range, op_col, constraints) 
    ops = Array.new
    op_range.each do |row|
      op_name = nil
      op_name = raw_input[row][op_col] unless raw_input[row][op_col] == nil
      if op_name == nil
        # do nothing, proceed to next row 
      else
        # find end of operation definition
        if op_name == "Faith"
          ops << Faithful_Operation_BST.new("Faith", [""], [""], constraints, parser)
        else
          op_end = row
          op_end += 1 until (raw_input[op_end][op_col] != nil && raw_input[op_end][op_col] != op_name)
          op_end -= 1
  
          curr_op_range = (row..op_end)
          source = Array.new
          target = Array.new
          
          #puts "Current op range: " + curr_op_range.to_s
          
          curr_op_range.each do |r2|
            source << raw_input[r2][op_col+1]
            target << raw_input[r2][op_col+2]
          end
          ops << Epenthesis_Operation_BST.new(op_name, source, target, constraints, parser) if op_name.start_with?("e.")
          ops << Operation_BST.new(op_name, source, target, constraints, parser) if !op_name.start_with?("e.")          
        end
      end
    end
    return ops
  end
    
  def build_est1_ops(raw_input, parser, op_range, op_col, constraints) 
    return build_bst_ops(raw_input, parser, op_range, op_col, constraints) 
  end
    
  # This method is currently nearly identical to build_bst_ops -- TODO: need to cut the ties
  # Now uses a BEE faith operation (that is built for PAKA - it produces no faith candidate
  # when there are not exactly one capital letter in the input. Capitalization in PAKA reps. stress
  # and only one segment can have stress in PAKA.)
  def build_bee_ops(raw_input, parser, op_range, op_col, constraints) 
    ops = Array.new
    op_range.each do |row|
      op_name = nil
      op_name = raw_input[row][op_col] unless raw_input[row][op_col] == nil
      if op_name == nil
        # do nothing, proceed to next row 
      else
        # find end of operation definition
        if op_name == "Faith"
          ops << Faithful_Operation_BEE.new("Faith", [""], [""], constraints, parser)
        else
          op_end = row
          op_end += 1 until (raw_input[op_end][op_col] != nil && raw_input[op_end][op_col] != op_name)
          op_end -= 1
  
          curr_op_range = (row..op_end)
          source = Array.new
          target = Array.new
          
          #puts "Current op range: " + curr_op_range.to_s
          
          curr_op_range.each do |r2|
            source << raw_input[r2][op_col+1]
            target << raw_input[r2][op_col+2]
          end
          ops << Epenthesis_Operation_BST.new(op_name, source, target, constraints, parser) if op_name.start_with?("e.")
          ops << Operation_BST.new(op_name, source, target, constraints, parser) if !op_name.start_with?("e.")          
        end
      end
    end
    return ops
  end
  
  # Returns a list of languages, where each language is represented as an
  # ordered pair:
  # * a language label
  # * a list of the labels of the optimal candidates of the language
  #
  # This method should be invoked when the active worksheet is the output
  # of a factorial typology computation. The selected cells should all
  # be in the column of language labels. For each selected language label,
  # the optimal candidates are read from the row containing that label.
  def get_ft_lang_and_opts
    ws = @excel.activesheet
    # Compute number of competitions as number of columns in row 3, minus
    # the first column, divided by two (there are two columns per competition:
    # candidate label and output string).
    column_count = ws.cells(3,1).end(XlToRight).Column
    competition_count = (column_count-1)/2
    # The selected cells contain the names of the languages
    selected_cells = @excel.selection             
    #removed: .specialcells(XlCellTypeVisible)                     -ASP FIX
    #ref to specialcells eliminated in favor of better VBA prep
    #specialcells screws up here if any rows are hidden anywhere, as by filtering.
    # For each selected cell, read the language name and
    # optimal candidate labels from that cell's row.
    lang_and_opt = []
    selected_cells.each do |cell|
      lang_label = cell.value
      opts = []
      (1..competition_count).each{|col| opts << ws.cells(cell.row,col*2).value}
      lang_and_opt << [lang_label, opts]
    end
    return lang_and_opt
  end
  
  def get_ft_lang_and_orders
    ws = @excel.activesheet
    # Compute number of competitions as number of columns in row 3, minus
    # the first column
    column_count = ws.cells(3,1).end(XlToRight).Column
    competition_count = column_count-1
    # The selected cells contain the names of the languages
    selected_cells = @excel.selection             
    #removed: .specialcells(XlCellTypeVisible)                     -ASP FIX
    #ref to specialcells eliminated in favor of better VBA prep 
    #specialcells screws up here if any rows are hidden anywhere, as by filtering.
    # For each selected cell, read the language name and
    # optimal candidate labels from that cell's row.
    lang_and_orders = []
    selected_cells.each do |cell|
      lang_label = cell.value
      orders = ws.cells(cell.row, column_count + 2).value
      lang_and_orders << [lang_label, orders]
    end
    return lang_and_orders
  end

  # Returns a Competition_list of the competitions, as previously pared down
  # to possibly optimal (non-harmonically bound) candidates.
  #
  # This method should be invoked when the active worksheet is the output
  # of a factorial typology computation. The immediately preceding worksheet
  # must be a violation tableau of the non-harmonically bound candidates
  # of the system (typically, the two sheets should be generated by the same
  # routine).
  def get_ft_competitions
    ft_sheet = @excel.activesheet # currently active ws should be fact. typ. list
    comp_sheet = @excel.Sheets(ft_sheet.index - 1) # get the preceding ws
    comp_sheet.activate # make the preceding sheet active
    competition_list = get_input_sheet_data # read in the VT data
    ft_sheet.activate # reactivate the factorial typology list sheet
    return competition_list
  end
  
  def get_one_lg_competitions              #RBShort project -AP
    competition_list = get_input_sheet_data # read in the VT data
    return competition_list
  end
  
  def move_active_sheet_back
    @excel.Sheets(@excel.activesheet.index - 1).Activate 
  end
  
  def move_active_sheet_forward
    @excel.Sheets(@excel.activesheet.index + 1).Activate 
  end
  
  # Write the data in <em>competition_list</em> to a new excel worksheet
  # in the form of a violation tableau.
  # The new worksheet is inserted after the currently active one, and the sheet
  # name is the label of <em>competition_list</em> with a prefixed "VT.".
  def put_competition_list_as_vt(competition_list)
    @excel.screenUpdating=false # turn off updating while writing to a worksheet
    # Add new worksheet after the current one  -AP
    ws = @excel.sheets.add('After'=>@excel.activesheet)
    ws.name = name_sheet("VT." + competition_list.label)
    # Put the competition label in the first cell
    range(ws,1,1,1,1).value = competition_list.label
    # The violation tableau begins on row 2
    viol_counts_to_ws(ws, competition_list, 2, 1)
    return ws
  ensure # make sure screen updating is turned back on, even if an exception is raised.
    @excel.screenUpdating=true
  end

  # Write violation data to a new excel worksheet, with colors indicating
  # which candidates are harmonically bounded.
  # The new worksheet is inserted after the currently active one, and the sheet
  # name is the label of <em>hb_result</em> with a prefixed "allHB.".
  # * Harmonically bounded candidates are colored Magenta.
  # * Non-harmonically bounded candidates are colored Light Green.
  # The input <em>hb_result</em> is a FactorialTypology
  # object, which includes analysis of harmonic bounding for all candidates
  # in a competition list.
  def put_inputs_all_hbound(hb_result)
    competition_list = hb_result.competition_list
    @excel.screenUpdating=false # turn off updating while writing to a worksheet
    # Add new worksheet after the current one  -AP
    ws = @excel.sheets.add('After'=>@excel.activesheet)
    ws.name = name_sheet("allHB." + competition_list.label)
    # Put the competition label in the first cell
    range(ws,1,1,1,1).value = competition_list.label
    # The violation tableau begins on row 2
    row_last, col_last = viol_counts_to_ws(ws, competition_list, 2, 1)
    # color the candidates
    label_col = 1; output_col = 3
    (3..row_last).each do |row|
      clabel = ws.cells(row,label_col).value
      if hb_result.hbound?(clabel) then
        ws.cells(row,label_col).interior.colorindex = MAGENTA
        ws.cells(row,output_col).interior.colorindex = MAGENTA
      else
        ws.cells(row,label_col).interior.colorindex = LIGHTGREEN
        ws.cells(row,output_col).interior.colorindex = LIGHTGREEN
      end
    end
    return ws
  ensure # make sure screen updating is turned back on, even if an exception is raised.
    @excel.screenUpdating=true
  end

  # Write violation data for non-harmonically-bound candidates.
  # The new worksheet is inserted after the currently active one, and the sheet
  # name is the label of <em>hb_result</em> with a prefixed "nonHB.".
  # The candidates are colored Light Green, to match the format of the
  # output sheet using color to distinguish HB from non-HB candidates;
  # see put_inputs_all_hbound().
  #
  # The input <em>hb_result</em> is a FactorialTypology
  # object, which includes analysis of harmonic bounding for all candidates
  # in a competition list.
  def put_inputs_non_hbound(hb_result)
    competition_list = hb_result.non_hb_competition_list
    @excel.screenUpdating=false # turn off updating while writing to a worksheet
    # Add new worksheet after the current one  -AP
    ws = @excel.sheets.add('After'=>@excel.activesheet)
    ws.name = name_sheet("nonHB." + competition_list.label)
    # Put the competition label in the first cell
    range(ws,1,1,1,1).value = competition_list.label
    # The violation tableau begins on row 2
    row_last, col_last = viol_counts_to_ws(ws, competition_list, 2, 1)
    # color the candidates light green (all are non-HB)
    label_col = 1; output_col = 3
    (3..row_last).each do |row|
      ws.cells(row,label_col).interior.colorindex = LIGHTGREEN
      ws.cells(row,output_col).interior.colorindex = LIGHTGREEN
    end
    return ws
  ensure # make sure screen updating is turned back on, even if an exception is raised.
    @excel.screenUpdating=true
  end

  # Writes a comparative tableau in CT Raw format to a new Excel worksheet.
  # The new worksheet is inserted after the currently active one, and the sheet
  # name is the label of <em>ct</em> with a prefixed "CT.".
  def put_ct_raw(ct)
    @excel.screenUpdating=false # turn off updating while writing to a worksheet
    ws = @excel.sheets.add('After'=>@excel.activesheet)
    ws.name = name_sheet("CT." + ct.label)
    range(ws,1,1,1,1).value = ct.label
    wl_pairs_to_ws(ws,ct.constraint_list,ct,2,1)
    return ws
  ensure # make sure screen updating is turned back on, even if an exception is raised.
    @excel.screenUpdating=true
  end

  # Sort the ercs of an RCD result in several ways.
  #
  # :call-seq:
  #   sort_rcd_results(rcd_result) -> [sorted_ercs, ercs_by_stratum, explained_ercs]
  #
  # The returned array contains three objects, each a list of ercs.
  # sorted_ercs:: all of the ercs in a flat array, sorted with respect to the
  #               constraint hierarchy constructed by RCD.
  # ercs_by_stratum:: all of the ercs in an array of strata, with each
  #                   stratum corresponding to a stratum of the hierarchy
  #                   constructed by RCD. The first erc stratum contains those
  #                   ercs assigned a W by a constraint in the first stratum
  #                   of the hierarchy, and so forth.
  # explained_ercs:: only the explained ercs (those that are properly accounted
  #                  for by RCD), in a flat array, sorted with respect to the
  #                  constraint hierarchy constructed by RCD.
  #--
  # TODO: This method is not Excel-specific; it should be moved to some module
  # containing functionality supporting general formatting.
  def sort_rcd_results(rcd_result)
    # Sort each "stratum" of ercs by the order of the constraints in the
    # corresponding constraint hierarchy stratum, so that the ercs assigned
    # a W by the first constraint in the stratum appear first, etc.
    # This uses a SyncEnumerator, which iterates in parallel through
    # multiple Enumerable objects (here, the explained ercs and the constraint
    # hierarchy).
    ercs_by_stratum = []
    gen = REXML::SyncEnumerator.new(rcd_result.ex_ercs, rcd_result.hierarchy)
    gen.each do |ercs,stratum|
      ercs_by_stratum << sort_by_constraint_order(ercs,stratum)
    end
    explained_ercs = ercs_by_stratum.flatten # in sorted order
    
    # add any unexplained ercs as the last level of the stratified erc list.
    ercs_by_stratum << rcd_result.unex_ercs unless rcd_result.unex_ercs.empty?

    # create flattened versions of
    # the "stratified" erc list (in the form of a CT).
    sorted_ercs = Comparative_tableau.new(rcd_result.label)
    sorted_ercs.concat(explained_ercs)
    sorted_ercs.concat(rcd_result.unex_ercs)
    
    return sorted_ercs, ercs_by_stratum, explained_ercs
  end
  
  # Writes the results of RCD in sorted comparative tableau format to a new
  # Excel worksheet.
  # The new worksheet is inserted after the currently active one.
  # The sheet name is <project_name>.RCD.
  def put_rcd_results(rcd_result) 
    # Add the unranked constraints as a "final stratum" to the hierarchy.
    hier_with_unranked = Hierarchy.new
    hier_with_unranked.concat(rcd_result.hierarchy)
    hier_with_unranked << rcd_result.unranked unless rcd_result.unranked.empty?
    sorted_cons = hier_with_unranked.flatten
    # sort the ercs with respect to the RCD constraint hierarchy
    sorted_ercs, ercs_by_stratum, explained_ercs = sort_rcd_results(rcd_result)
    # Set some table index values
    row_first = 2
    col_first = 1
    pre_erc_rows = 1 # the column heading row with the constraint names
    pre_con_columns = 4 # assumed by wl_pairs_to_ws()
    last_ex_row = row_first - 1 + pre_erc_rows + explained_ercs.size # row of last explained erc
    row_last = last_ex_row + rcd_result.unex_ercs.size

    @excel.screenUpdating=false # turn off updating while writing to a worksheet
    ws = @excel.sheets.add('After'=>@excel.activesheet)
    ws.name = name_sheet("#{@proj_name}.RCD")
    ws.cells(1,1).value = @proj_name
    status_stamp(ws,"RCD")
    
    # write the main CT to the new worksheet
    wl_pairs_to_ws(ws,sorted_cons,sorted_ercs,row_first,col_first)
    # Create borders demarking stratified constraints, ercs
    stratified_tableau_borders(ws,row_first,col_first,pre_erc_rows,
      pre_con_columns,hier_with_unranked,ercs_by_stratum)
    # Apply RCD-based tableau coloring
    rcd_tableau_coloring(ws,row_first,col_first,pre_erc_rows,
      pre_con_columns,hier_with_unranked,ercs_by_stratum,explained_ercs)
    # Extra formatting if the data were inconsistent
    unless rcd_result.consistent? then
      row_extra = row_last+1
      fail_cell = range(ws,row_extra,col_first,row_extra,col_first)
      fail_cell.value = "FAIL!"
      fail_cell.interior.colorindex = RED
    end

    return ws
  ensure # make sure screen updating is turned back on, even if an exception is raised.
    @excel.screenUpdating=true
  end

  # Puts stratified tableau borders on a tableau, based on stratification
  # of both the constraint hierarchy and the ercs.
  def stratified_tableau_borders(ws, row_first, col_first, pre_erc_rows, pre_con_columns,
                                 hier_stratified, ercs_stratified)
    con_count = hier_stratified.flatten.size
    erc_count = ercs_stratified.flatten.size
    col_count = pre_con_columns + con_count
    row_count = pre_erc_rows + erc_count
    col_last = col_first + col_count - 1
    row_last = row_first + row_count - 1
    #
    vl_col_index = col_first + pre_con_columns - 1
    hl_row_index = row_first + pre_erc_rows - 1
    hier_by_stratum = hier_stratified.to_enum
    ercs_by_stratum = ercs_stratified.to_enum
    loop do
      # Find the next stratum of constraints, and draw a border to the right
      vl_col_index += hier_by_stratum.next.size
      range(ws,row_first,vl_col_index,row_last,vl_col_index).borders(XlEdgeRight).weight =
        XlMedium # vertical line between strata
      # Find the set of ercs accounted for by the stratum, and draw a border beneath
      hl_row_index += ercs_by_stratum.next.size
      range(ws,hl_row_index,col_first,hl_row_index,col_last).borders(XlEdgeBottom).weight =
        XlMedium # horizontal line between "stratified" clusters of ercs
    end
  end

  # Apply RCD-determined tableau coloring, based on stratification,
  # which ercs are explained by which constraints, and the presence/absence
  # of unexplained ercs.
  def rcd_tableau_coloring(ws,row_first,col_first,pre_erc_rows,pre_con_columns,hier_stratified,ercs_stratified,explained_ercs)
    con_count = hier_stratified.flatten.size
    erc_count = ercs_stratified.flatten.size
    col_count = pre_con_columns + con_count
    row_count = pre_erc_rows + erc_count
    col_last = col_first + col_count - 1
    row_last = row_first + row_count - 1
    last_ex_row_index = row_first - 1 + pre_erc_rows + explained_ercs.size
    hier_by_stratum = hier_stratified.to_enum
    ercs_by_stratum = ercs_stratified.to_enum
    # Set up the rotation of regular cell colors
    cell_colors = Fiber.new do
      loop do
        Fiber.yield BRIGHTGREEN
        Fiber.yield BRIGHTCYAN
      end
    end

    vl_col_index = col_first + pre_con_columns - 1 # col prior to next stratum
    hl_row_index = row_first + pre_erc_rows - 1 # row prior to next stratum
    loop do
      stratum_con_count = hier_by_stratum.next.size
      stratum_col_first = vl_col_index + 1
      stratum_col_last = vl_col_index + stratum_con_count
      vl_col_index += stratum_con_count # increment by num. of constraints in stratum
      #
      stratum_erc_count = ercs_by_stratum.next.size
      stratum_row_first = hl_row_index 
      stratum_row_first += 1 unless (stratum_erc_count==0)
      stratum_row_last = hl_row_index + stratum_erc_count
      hl_row_index += stratum_erc_count # increment by num. of ercs for stratum
      # Select the next stratum color
      cur_color = cell_colors.resume
      # Set color for the constraint names (column headings) of the stratum.
      head_row = row_first + pre_erc_rows - 1
      con_name_range = range(ws,head_row,stratum_col_first,head_row,stratum_col_last)
      if hl_row_index <= last_ex_row_index then
        cur_color = GOLD if (vl_col_index==col_last) # the last stratum is colored gold.
        con_name_range.interior.colorindex = cur_color
      else
        con_name_range.interior.colorindex = RED
      end
      # Color the cells with W indicating that a constraint of the current
      # stratum accounts for an erc.
      unless (stratum_erc_count == 0) then
        target_range = range(ws,stratum_row_first,stratum_col_first,stratum_row_last,stratum_col_last)
        target_range.each do |cell|
          next if cell.value.nil?
          if hl_row_index <= last_ex_row_index then
            cell.interior.colorindex = cur_color
          else
            cell.interior.colorindex = RED # cell for an unexplained erc.
            ws.cells(cell.row,col_first).interior.colorindex = RED # also mark the erc label
          end
        end
        # For the ercs explaining the next to last stratum, color the L's of
        # the last stratum of constraints.
        if (hl_row_index==row_last) && (hl_row_index<=last_ex_row_index) then
          target_range = range(ws,stratum_row_first,stratum_col_last+1,stratum_row_last,col_last)
          target_range.each do |cell|
            cell.interior.colorindex = GOLD if cell.value == "L"
          end
        end
      end
    end
  end

  # Takes a list of ercs and sorts them with respect to a list of constraints,
  # such that all ercs assigned a W by the first constraint occur first in
  # the sorted erc list, followed by all the ercs assigned a W by the second
  # constraint (but not the first), and so forth. Ercs that are not assigned
  # a W by any of the constraints in the list occur last.
  #
  # :call-seq:
  #   sort_by_constraint_order(erc_list, con_list) -> array
  #
  # This is used for display purposes, to create a clean "W boundary" in
  # formatted comparative tableaux.
  #--
  # TODO: This method is not Excel-specific; it should be moved to some module
  # containing functionality supporting general formatting.
  def sort_by_constraint_order(ercs,cons)
    no_w_ercs = ercs
    sorted_ercs = Array.new
    cons.each do |con|
      w_ercs, no_w_ercs = no_w_ercs.partition{|e| e.w?(con)}
      sorted_ercs.concat(w_ercs)
    end
    sorted_ercs.concat(no_w_ercs) # could only be ercs with no 'W'
    return sorted_ercs
  end
  
  # Put the MIB to an output worksheet
  def put_mib(fred_result)
    sorted_cons = fred_result.rcd_result.hierarchy.flatten
    mib = Comparative_tableau.new(fred_result.label)
    fred_result.fred_kept_steps.each {|step| mib << step.fusion}

    @excel.screenUpdating=false # turn off updating while writing to a worksheet
    ws = @excel.sheets.add('After'=>@excel.activesheet)
    ws.name = name_sheet("MIB." + mib.label)
    range(ws,1,1,1,1).value = mib.label
    basis_ercs_to_ws(ws,sorted_cons,mib,2,1)
    return ws
  ensure # make sure screen updating is turned back on, even if an exception is raised.
    @excel.screenUpdating=true
  end

  # Put the skeletal basis (SKB) to an output worksheet
  def put_skb(fred_result)
    sorted_steps = fred_result.fred_kept_steps
    sorted_cons = fred_result.rcd_result.hierarchy.flatten

    sorted_ercs = Comparative_tableau.new(fred_result.label)
    sorted_ercs.concat(sorted_steps.map{|s| s.skb_erc })
    
    @excel.screenUpdating=false # turn off updating while writing to a worksheet
    ws = @excel.sheets.add('After'=>@excel.activesheet)
    ws.name = name_sheet("SKB." + sorted_ercs.label)
    range(ws,1,1,1,1).value = sorted_ercs.label    
    basis_ercs_to_ws(ws,sorted_cons,sorted_ercs,2,5)
    # Construct the support ct
    end_row, end_col = ws.usedrange.rows.count, ws.usedrange.columns.count
    skb_support_to_ws(ws,sorted_cons,sorted_steps,end_row+2,1)
    return ws
  ensure # make sure screen updating is turned back on, even if an exception is raised.
    @excel.screenUpdating=true
  end

  def put_kept_residues(fred_result)
    sorted_steps = fred_result.fred_kept_steps
    sorted_cons = fred_result.rcd_result.hierarchy.flatten
    @excel.screenUpdating=false # turn off updating while writing to a worksheet
    ws = @excel.sheets.add('After'=>@excel.activesheet)
    ws.name = name_sheet("Residues." + fred_result.label)
    range(ws,1,1,1,1).value = fred_result.label
    row_next = 2
    sorted_steps.each do |step|
      residue_ercs_to_ws(ws,sorted_cons,step,row_next,1)
      row_next += step.comp_tableau.size + 1
      # put the fusion at the bottom
      row_next += 3 # fusion row, plus two rows between residue tableaus
    end
    return ws
  ensure # make sure screen updating is turned back on, even if an exception is raised.
    @excel.screenUpdating=true    
  end
  
  # Writes a factorial typology to a worksheet, listing each language
  # in a separate row, with different columns corresponding to different
  # inputs, and autofilters on the column headings.
  def put_ft(lang_ct_list, system_label = "")
    @excel.screenUpdating=false # turn off updating while writing to a worksheet
    # create and label the FacType worksheet
    ws = @excel.sheets.add('After'=>@excel.activesheet)
    ws.name = name_sheet("FacTyp.#{system_label}")
    sheet_image = []
    # Fill label cell
    ws.cells(1,1).value = system_label
    ws.cells(1,1).font.bold = true
    # table dimensions and row/col coordinates
    typ_count = lang_ct_list.size
    row_last = 3 + typ_count
    lang_size = lang_ct_list[0].winners.size
    col_last = (lang_size * 2) + 1
    # Write the header row
    sheet_image << (row_image = [])
    row_image << "Inputs"
    cset_num = 0
    lang_ct_list[0].winners.each do |winner|
      cset_num += 1
      row_image << "cset-#{cset_num}" << winner.input.to_s
    end
    # Write the winner labels and outputs for each language
    lang_ct_list.each do |lang|
      sheet_image << (row_image = [])
      win_list = lang.winners
      row_image << "Lang# " + lang.label
      win_list.each do |winner|
        row_image << winner.label << winner.merged_outputs_to_s
      end
    end
    # Set candidate label columns as text, bold the headers of the other columns.
    # Candidate label columns are even-numbered; other are odd-numbered.
    # There is one more odd-numbered column than even-numbered one, hence the
    # loop with an if-statement.
    # The label column formatting as text must be done before the label data
    # are written to the sheets, in order for the data to be read back in
    # properly by other procedures.
    for col_ptr in (1..col_last)
      if col_ptr % 2 == 1 then # odd-numbered column
        ws.cells(3,col_ptr).font.bold = true # bold the header
      else
        range(ws,4,col_ptr,row_last,col_ptr).numberformat = "@" # text format
      end
    end
    # Write the image to the worksheet
    range(ws,3,1,row_last,col_last).value = sheet_image
    range(ws,3,1,3,col_last).autofilter # autofilters on the column headings
    range(ws,3,1,3,col_last).horizontalalignment = XlCenter
    # Autofit the columns last, so that it accommodates all the other formatting.
    range(ws,1,1,row_last,col_last).entirecolumn.autofit # autosize the columns
    return ws
  ensure # make sure screen updating is turned back on, even if an exception is raised.
    @excel.screenUpdating=true    
  end
  
  # Returns the complete filename for the OTWorkPlace excel workbook, by
  # searching for the open workbook with a filename starting with "otwpl".
  # Returns nil if no such workbook is found.
  #--
  # Added 8/7/10 by AP; modified 7/25/11 by BBT.
  def find_otwpl
    @excel.workbooks.each do |wb|
      bookname = wb.name
      return bookname if /^otwpl/ =~ bookname
    end
    return nil # no appropriately named workbook was found.
  end
  
  # Adds the status stamp to the given worksheet, using the given status
  # label. Follows the convention of putting the status stamp in the first
  # row second column cell.
  #--
  # Added by AP.
  def status_stamp(ws, status)
    ws.cells(1,2).value = status
    ws.cells(1,2).font.colorindex = 18
    ws.cells(1,2).font.italic = true
    ws.cells(1,2).font.bold = true
    ws.cells(1,2).font.size = 9
    ws.cells(1,2).horizontalAlignment = XlCenter
  end

  def put_moat(moat)
    @excel.screenUpdating=false # turn off updating while writing to a worksheet

    # TODO: (BBT) ultimately, properties like tabcolor should be stored in some
    # object which is constructed when the command is first begun.
    #
    # Determine the appropriate tab color index for the new sheet.
    # * If the current sheet is a factorial typology sheet, take the color from
    # the preceding sheet.
    # * Otherwise (current sheet is a just-generated FRed sheet), take the color
    # from the current sheet.
    active_ws = @excel.activesheet
    if /FacTyp/ =~ active_ws.name
      cs_index = active_ws.index - 1
    else
      cs_index = active_ws.index
    end
    color_index = @excel.sheets(cs_index).tab.colorindex

    # Create a new output worksheet, and set names, labels, and tab
    # color according to the OTWP standard.
    # Cell 1,1: project name   Cell 1,2: status stamp   Cell 1,3: language name
    ws = @excel.sheets.add('After'=>active_ws)
    ws.name = name_sheet("#{@proj_name}.Moat")
    ws.tab.colorindex = color_index
    ws.cells(1,1).value = @proj_name
    status_stamp(ws,"Moat")
    #range(ws,1,3,1,3).value = "Moat"

    # For each diagram basis, create a cell with a pop-up comment
    # containing the ranking diagram.
    row_last = 2
    col_last = 1
    epo_count = 0
    dist=3
    rw = 0
    moat.epos.keys.each do |key|
      epo_count += 1
      label = key.name
      epo_d = EpoDiagram.new(moat.epos[key], label)
      tempfile = File.join(File.dirname(__FILE__),'..','temp',label)
      epo_d.png(tempfile)
      rw = rw+dist                                                  #AP/NM-added
      @excel.columns(col_last+1).columnwidth = 3                    #AP/NM- added
      cell = range(ws,rw,col_last+2,rw,col_last+2)                  #AP/NM- modified
      cell.value = label
      #cell.value = epo.to_s
      put_png_comment(tempfile+'.png', cell)
      cell.comment.visible = true                                   #AP/NM- added
      dist = (cell.comment.shape.height/12).to_i + 1                #AP/NM- added

      tempfile2 = File.join(File.dirname(__FILE__),'..','MOATS', "#{label}.gv")
      #File.open("#{File.dirname(__FILE__).to_s}/Naz_file_2", 'w') do |f|
      File.open(tempfile2, 'w') do |f|
        f.puts epo_d.get_gv
      end

    end

    return ws

    ensure # make sure screen updating is turned back on, even if an exception is raised.
    @excel.screenUpdating=true
  end
  
  
  def put_ft_hs_ops(lang_list, input_list, ft)
    @excel.screenUpdating=false # turn off updating while writing to a worksheet
    active_ws = @excel.activesheet

    # Create a new output worksheet, and set names, labels, and tab
    # color according to the OTWP standard.
    # Cell 1,1: project name   Cell 1,2: status stamp   Cell 1,3: language name
    ws = @excel.sheets.add('After'=>active_ws)
    ws.name = name_sheet("FacTyp-"+@proj_name)
    #ws.tab.colorindex = color_index
    #ws.cells(1,1).value = "HS.FT" #project name
    ws.cells(1,1).value = @proj_name
    status_stamp(ws,"FacTyp-HS")
    
    # this block lists inputs

    i_num = 0
    input_list.each{|input| range(ws, 3, i_num+2, 3, i_num+2).value = input; i_num += 1}

  
    # row of starting language
    lang_row = 4
    lang_count = 0

    lang_list.each do |lang|
      in_col = 2
      in_count = 0
      max_num_paths = 1
      lang_count += 1
      range(ws, lang_row, 1, lang_row, 1).value = "L#{lang_count}"
      
      # Record a total order tha produces this language
      fr = Fred.new(ft.get_comparative_tableau(lang))
      ord = ""
      constraints = fr.constraint_list
      fr.leg.each {|con| ord = "#{ord.to_s}.#{constraints.index(con)}"}
      range(ws, lang_row, input_list.size + 3, lang_row, input_list.size + 3).value = ord
      
      input_list.each do |input|
        p_count = 0
        lang[input].paths.each do |p|
          range(ws, lang_row+p_count, in_col+in_count, lang_row+p_count, in_col+in_count).value = p.to_s          
          p_count += 1
        end
        max_num_paths = p_count if p_count > max_num_paths        
        in_count += 1
      end
      lang_row += max_num_paths
    end
    
    range(ws,3,1,3,1).value = "Inputs->"
    ws.cells(3,1).autofilter 
    # autosize the columns
    #range(ws).columns.autofit
    ws.columns.autofit

    return ws
  ensure # make sure screen updating is turned back on, even if an exception is raised.
    @excel.screenUpdating=true    
  end
  
  def put_ft_hs(lang_list, comp_list)
    @excel.screenUpdating=false # turn off updating while writing to a worksheet
    active_ws = @excel.activesheet

    # Create a new output worksheet, and set names, labels, and tab
    # color according to the OTWP standard.
    # Cell 1,1: project name   Cell 1,2: status stamp   Cell 1,3: language name
    ws = @excel.sheets.add('After'=>active_ws)
    ws.name = name_sheet("FacTyp-"+@proj_name)
    #ws.tab.colorindex = color_index
    #ws.cells(1,1).value = "HS.FT" #project name
    ws.cells(1,1).value = @proj_name
    status_stamp(ws,"FacTyp-HS")
    
    # this block lists languages & candidate set inputs
    range(ws,3,1,3,1).value = "Inputs"
    i_num = 0
    comp_list.each{|comp| range(ws,i_num+4,1,i_num+4,1).value = comp[0].input; i_num += 1}
    l_num = 0
    lang_list.each {|lang| range(ws,3,l_num+2,3,l_num+2).value = lang.label; l_num += 1} 
    
    # row and col of starting input and language
    input_pos = 4
    lang_pos = 2
    
    lang_list.each do |lang|
      comp_list.each do |comp|
        input = comp[0].input
        input_processed = false
        lang.each do |wlpair|
          if !input_processed
            if wlpair.winner.input == input
              range(ws,input_pos,lang_pos,input_pos,lang_pos).value = wlpair.winner.output
              input_processed = true
            end     
          end
        end
        input_pos += 1
      end
      lang_pos += 1
      input_pos = 4
    end

    return ws
  ensure # make sure screen updating is turned back on, even if an exception is raised.
    @excel.screenUpdating=true    
  end
  
  def put_ft_cand_skb(fred_result)
    rcd_result = fred_result.rcd_result
    # Add the unranked constraints as a "final stratum" to the hierarchy.
    hier_with_unranked = Hierarchy.new
    hier_with_unranked.concat(rcd_result.hierarchy)
    hier_with_unranked << rcd_result.unranked unless rcd_result.unranked.empty?
    sorted_cons = hier_with_unranked.flatten
    # sort the ercs with respect to the RCD constraint hierarchy
    sorted_ercs, ercs_by_stratum, explained_ercs = sort_rcd_results(rcd_result)
    #
    @excel.screenUpdating=false # turn off updating while writing to a worksheet

    # TODO: (BBT) ultimately, properties like tabcolor should be stored in some
    # object which is constructed when the command is first begun.
    # 
    # Determine the appropriate tab color index for the new sheet.
    # * If the current sheet is a factorial typology sheet, take the color from
    # the preceding sheet.
    # * Otherwise (current sheet is a just-generated FRed sheet), take the color
    # from the current sheet.
    active_ws = @excel.activesheet
    if /FacTyp/ =~ active_ws.name
      cs_index = active_ws.index - 1
    else
      cs_index = active_ws.index
    end
    color_index = @excel.sheets(cs_index).tab.colorindex

    # Create a new output worksheet, and set names, labels, and tab
    # color according to the OTWP standard.
    # Cell 1,1: project name   Cell 1,2: status stamp   Cell 1,3: language name
    ws = @excel.sheets.add('After'=>active_ws)
    ws.name = name_sheet("#{@proj_name}.#{fred_result.label}")
    ws.tab.colorindex = color_index
    ws.cells(1,1).value = @proj_name
    status_stamp(ws,"RnkRpt")
    range(ws,1,3,1,3).value = sorted_ercs.label 

    # put SKB
    range(ws,2,5,2,5).value = "SKB"
    row_last, col_last =
      basis_ercs_to_ws(ws,sorted_cons,fred_result.skb,3,5)
    # put the support for the SKB
    range(ws,row_last+2,1,row_last+2,1).value = "Support"
    row_last, col_last =
      skb_support_to_ws(ws,sorted_cons,fred_result.fred_kept_steps,row_last+3,1)
    # put full set of winner-loser pairs
    range(ws,row_last+2,2,row_last+2,2).value = "WL Pairs"
    row_first_wlp = row_last+3
    col_first_wlp = 2
    row_last, col_last =
      wl_pairs_to_ws(ws,sorted_cons,sorted_ercs,row_first_wlp,col_first_wlp)
        
    #put MIB onto rnkrpt sheet  -AP May 3 2013
    range(ws,row_last+2,5,row_last+2,5).value = "MIB"
        row_last, col_last =
          basis_ercs_to_ws(ws,sorted_cons,fred_result.mib,row_last+3,5)
      
    # Format RCD Display with VBA macro "rcdPretty"   -AP
#    @excel.cells(row_last,2).select
#    @excel.run "'#{find_otwpl}'!rcdPretty"
    # Set parameters for the tableau to be formatted
    pre_erc_rows = 1
    pre_con_columns = 4
    # Create borders demarking stratified constraints, ercs
    stratified_tableau_borders(ws,row_first_wlp,col_first_wlp,pre_erc_rows,
      pre_con_columns,hier_with_unranked,ercs_by_stratum)
    # Apply RCD-based tableau coloring
    rcd_tableau_coloring(ws,row_first_wlp,col_first_wlp,pre_erc_rows,
      pre_con_columns,hier_with_unranked,ercs_by_stratum,explained_ercs)

    # Find the diagram bases for the SKB.
    #skb_exp = Skb_expansion.new(fred_result.skb)
    #db = skb_exp.diagram_bases

  # Previous version (from 7.4.2013) of diagram expansion *not* using disjunctive expansion.
    # Find the diagram bases for the SKB.
    skb_exp = Skb_expansion.new(fred_result.skb)
    db = skb_exp.diagram_bases
  # End previous version

#AP/NM - new form of disjunctive bases using L/W technique instead of classical e/W rep
=begin
disj_cart_prod = Erc.disj_cart_prod(fred_result.skb)
db = []
disj_cart_prod.each do |disj_e|
    ct = Comparative_tableau.new
    ct.concat(disj_e)
    temp_fred = Fred.new(ct)
    disj_ct = Comparative_tableau.new
    disj_ct.concat(Skb_expansion.new(temp_fred.skb).diagram_bases[0])
    db << disj_ct
end
=end

    # For each diagram basis, create a cell with a pop-up comment
    # containing the ranking diagram.
    row_last += 2
    dcount = 0
    dist=3
    rw = 0
    db.each do |basis|
      dcount += 1
      label = "Diagram#{dcount}"                                   
      rd = RankingDiagram.new(basis, label)
      tempfile = File.join(File.dirname(__FILE__),'..','temp',label)
      rd.png(tempfile)
      rw = rw+dist                                                  #AP/NM-added
      @excel.columns(col_last+1).columnwidth = 3                    #AP/NM- added
      cell = range(ws,rw,col_last+2,rw,col_last+2)                  #AP/NM- modified
      cell.value = label
      put_png_comment(tempfile+'.png', cell)
      cell.comment.visible = true                                   #AP/NM- added
      dist = (cell.comment.shape.height/12).to_i + 1                #AP/NM- added
    end
    
 #ADD hyperlink to lang name [c3] in rnkrpt going back to factyp sheet   -AP
 #But only do it in factyp sequence -- not in shortreport sequence
 if fred_result.label!="ShortReport"
    ws.hyperlinks.add 'anchor'=>ws.cells(1,3), 'Address'=>"",'subaddress'=>"\'FacTyp-#{@proj_name}\'!R1C1",'TextToDisplay'=>ws.cells(1,3).value
 
     #add hyperlink to factyp sheet to lg rnkrpt , nb language name = fred_result.label
     #actual current sheet name = ws.name
     #need to pick up address of language name = [c1].value on factyp sheet
     #note eventual issue wrt join, since that page is unser-named in vba at end of Mojo method
     #REMAINING ISSUE, apparently,sometimes --- discontinuous selections, or at least some prob w. names. May be more general than hlinking
 #ADD hyperlink to factyp page, at lg name in col 1, linking to rnkrpt sheet   
  lang_sht_name=ws.name
  ftSheet= @excel.sheets("FacTyp-"+@proj_name)
  lang_name = ws.cells(1,3).value
  # NNM: the two following lines of code break the FacTyp display for HS because the 
  # layout of the languages on the summary page is different. Would be good to unify the 
  # presentations.
    
  myRow = ftSheet.cells.find('What'=>lang_name,'LookAt'=>XlWhole,'MatchCase'=>true,'after'=>ftSheet.cells(1,1)).row    #ASP FIX 2.24.14
  @excel.sheets(ftSheet.name).hyperlinks.add('anchor'=>ftSheet.cells(myRow,1),'address'=>"",'subaddress'=>"\'#{lang_sht_name}\'!R1C1",'TextToDisplay'=> ftSheet.cells(myRow,1).value)
 end  
    return ws
  ensure # make sure screen updating is turned back on, even if an exception is raised.
    @excel.screenUpdating=true    
  end
  
  def put_ft_hs_cand_skb(fred_result, op_lang=nil, factyp=true)
    rcd_result = fred_result.rcd_result
    # Add the unranked constraints as a "final stratum" to the hierarchy.
    hier_with_unranked = Hierarchy.new
    hier_with_unranked.concat(rcd_result.hierarchy)
    hier_with_unranked << rcd_result.unranked unless rcd_result.unranked.empty?
    sorted_cons = hier_with_unranked.flatten
    # sort the ercs with respect to the RCD constraint hierarchy
    sorted_ercs, ercs_by_stratum, explained_ercs = sort_rcd_results(rcd_result)
    #
    @excel.screenUpdating=false # turn off updating while writing to a worksheet

    # TODO: (BBT) ultimately, properties like tabcolor should be stored in some
    # object which is constructed when the command is first begun.
    # 
    # Determine the appropriate tab color index for the new sheet.
    # * If the current sheet is a factorial typology sheet, take the color from
    # the preceding sheet.
    # * Otherwise (current sheet is a just-generated FRed sheet), take the color
    # from the current sheet.
    active_ws = @excel.activesheet
    if /FacTyp/ =~ active_ws.name
      cs_index = active_ws.index - 1
    else
      cs_index = active_ws.index
    end
    color_index = @excel.sheets(cs_index).tab.colorindex

    # Create a new output worksheet, and set names, labels, and tab
    # color according to the OTWP standard.
    # Cell 1,1: project name   Cell 1,2: status stamp   Cell 1,3: language name
    ws = @excel.sheets.add('After'=>active_ws)
    ws.name = name_sheet("#{@proj_name}.#{fred_result.label}")
    ws.tab.colorindex = color_index
    ws.cells(1,1).value = @proj_name
    status_stamp(ws,"RnkRpt")
    range(ws,1,3,1,3).value = fred_result.label

#ADD hyperlink to lang name [c3] in rnkrpt going back to factyp sheet   -AP
 #But only do it in factyp sequence -- not in shortreport sequence
 #if fred_result.label!="ShortReport" 
 ws.hyperlinks.add 'anchor'=>ws.cells(1,3), 'Address'=>"",'subaddress'=>"\'FacTyp-#{@proj_name}\'!R1C1",'TextToDisplay'=>ws.cells(1,3).value


    # put SKB
    range(ws,2,5,2,5).value = "SKB"
    row_last, col_last =
      basis_ercs_to_ws(ws,sorted_cons,fred_result.skb,3,5)
    # put the support for the SKB
    range(ws,row_last+2,1,row_last+2,1).value = "Support"
    row_last, col_last =
      skb_support_to_ws(ws,sorted_cons,fred_result.fred_kept_steps,row_last+3,1)
    # put full set of winner-loser pairs
    range(ws,row_last+2,2,row_last+2,2).value = "WL Pairs"
    row_first_wlp = row_last+3
    col_first_wlp = 2
    row_last, col_last =
      wl_pairs_to_ws(ws,sorted_cons,sorted_ercs,row_first_wlp,col_first_wlp)
        
    #put MIB onto rnkrpt sheet  -AP May 3 2013
    range(ws,row_last+2,5,row_last+2,5).value = "MIB"
        row_last, col_last =
          basis_ercs_to_ws(ws,sorted_cons,fred_result.mib,row_last+3,5)
      
    # Format RCD Display with VBA macro "rcdPretty"   -AP
#    @excel.cells(row_last,2).select
#    @excel.run "'#{find_otwpl}'!rcdPretty"
    # Set parameters for the tableau to be formatted
    pre_erc_rows = 1
    pre_con_columns = 4
    # Create borders demarking stratified constraints, ercs
    stratified_tableau_borders(ws,row_first_wlp,col_first_wlp,pre_erc_rows,
      pre_con_columns,hier_with_unranked,ercs_by_stratum)
    # Apply RCD-based tableau coloring
    rcd_tableau_coloring(ws,row_first_wlp,col_first_wlp,pre_erc_rows,
      pre_con_columns,hier_with_unranked,ercs_by_stratum,explained_ercs)

    # Find the diagram bases for the SKB.
    skb_exp = Skb_expansion.new(fred_result.skb)
    db = skb_exp.diagram_bases

    # For each diagram basis, create a cell with a pop-up comment
    # containing the ranking diagram.
    row_last += 2
    dcount = 0
    dist=3
    rw = 0
    db.each do |basis|
      dcount += 1
      label = "Diagram#{dcount}"                                   
      rd = RankingDiagram.new(basis, label)
      tempfile = File.join(File.dirname(__FILE__),'..','temp',label)
      rd.png(tempfile)
      rw = rw+dist                                                  #AP/NM-added
      @excel.columns(col_last+1).columnwidth = 3                    #AP/NM- added
      cell = range(ws,rw,col_last+2,rw,col_last+2)                  #AP/NM- modified
      cell.value = label
      put_png_comment(tempfile+'.png', cell)
      cell.comment.visible = true                                   #AP/NM- added
      dist = (cell.comment.shape.height/12).to_i + 1                #AP/NM- added
    end
    
    if !op_lang.nil? 
      label = "Ovpo_#{fred_result.label}"
      ov_po = OpvecPO.new(op_lang)
      od = OpvecPODiagram.new(ov_po, label)
       
      tempfile = File.join(File.dirname(__FILE__),'..','temp',label)
      od.png(tempfile)
      rw = rw+dist                                                  #AP/NM-added
      @excel.columns(col_last+1).columnwidth = 3                    #AP/NM- added
      cell = range(ws,rw,col_last+2,rw,col_last+2)                  #AP/NM- modified
      cell.value = label
      put_png_comment(tempfile+'.png', cell)
      cell.comment.visible = true                                   #AP/NM- added
      dist = (cell.comment.shape.height/12).to_i + 1  
      
      puts "About to output #{label}.gv"
      
      tempfile2 = File.join(File.dirname(__FILE__),'..','OpvecPOs', "#{label}.gv")
      File.open(tempfile2, 'w') do |f|
        f.puts od.get_gv
      end
      
      #put Derivations onto rnkrpt sheet  -NM Oct 2013
      #rw = rw+dist
      #range(ws,rw,col_last+2,rw,col_last+2).value = "Derivations"
      #derivations_to_ws(ws,op_lang,rw+1,col_last+2) 
      #derivations_to_ws_top_to_bottom(ws,op_lang,rw+1,col_last+2) 
      
      put_hs_deriv(fred_result, op_lang, sorted_cons)
    end
    
    
 #ADD hyperlink to lang name [c3] in rnkrpt going back to factyp sheet   -AP
 #   ws.hyperlinks.add 'anchor'=>ws.cells(1,3), 'Address'=>"",'subaddress'=>"\'FacTyp-#{@proj_name}\'!R1C1",'TextToDisplay'=>ws.cells(1,3).value
 
     #add hyperlink to factyp sheet to lg rnkrpt , nb language name = fred_result.label
     #actual current sheet name = ws.name
     #need to pick up address of language name = [c1].value on factyp sheet
     #note eventual issue wrt join, since that page is unser-named in vba at end of Mojo method
     #REMAINING ISSUE, apparently,sometimes --- discontinuous selections, or at least some prob w. names. May be more general than hlinking

#ADD hyperlink to factyp page, at lg name in col 1, linking to rnkrpt sheet

  lang_sht_name=ws.name
  #if factyp                                          -ASP FIX. This doesn't work as conditional, for some reason.
    ftSheet= @excel.sheets("FacTyp-"+@proj_name)
    lang_name = ws.cells(1,3).value
    # NNM: the two following lines of code break the FacTyp display for HS because the 
    # layout of the languages on the summary page is different. Would be good to unify the 
    # presentations.
    
    myRow = ftSheet.cells.find('What'=>lang_name,'LookAt'=>XlWhole,'MatchCase'=>true,'after'=>ftSheet.cells(1,1)).row    #ASP FIX 2.24.14
    @excel.sheets(ftSheet.name).hyperlinks.add('anchor'=>ftSheet.cells(myRow,1),'address'=>"",'subaddress'=>"\'#{lang_sht_name}\'!R1C1",'TextToDisplay'=> ftSheet.cells(myRow,1).value)
  #end          -ASP FIX
   
    return ws
  ensure # make sure screen updating is turned back on, even if an exception is raised.
    @excel.screenUpdating=true    
  end
  
  def put_hs_deriv(fred_result, op_lang, sorted_cons)
    @excel.screenUpdating=false # turn off updating while writing to a worksheet
    active_ws = @excel.activesheet
=begin
    if /FacTyp/ =~ active_ws.name
      cs_index = active_ws.index - 1
    else
      cs_index = active_ws.index
    end
    color_index = @excel.sheets(cs_index).tab.colorindex
=end
    # Create a new output worksheet, and set names, labels, and tab
    # color according to the OTWP standard.
    # Cell 1,1: project name   Cell 1,2: status stamp   Cell 1,3: language name
    ws = @excel.sheets.add('After'=>active_ws)
    ws.name = name_sheet("#{@proj_name}.#{fred_result.label}.dr")
    ws.cells(1,1).value = @proj_name
    status_stamp(ws,"Deriv")
    ws.cells(1,3).value = fred_result.label
    
    #put Derivations onto rnkrpt sheet  -NM Oct 2013
    rw = 2
    col = 2
    range(ws,rw,col,rw,col).value = "Derivations"
    #derivations_to_ws(ws,op_lang,rw+1,col_last+2) 
    derivations_to_ws_top_to_bottom(ws,op_lang,rw+1,col, sorted_cons) 
 
 #ADD hyperlink to lang name [c3] in rnkrpt going back to factyp sheet   -AP
    ws.hyperlinks.add 'anchor'=>ws.cells(1,3), 'Address'=>"",'subaddress'=>"\'FacTyp-#{@proj_name}\'!R1C1",'TextToDisplay'=>ws.cells(1,3).value
 
    return ws
  ensure # make sure screen updating is turned back on, even if an exception is raised.
    @excel.screenUpdating=true    
  end
  
  def derivations_to_ws(ws,lang,rw,col) 
    constraints = lang.values[0].constraints
    lang.keys.reverse.each do |inp|
      #range(ws,rw,col,rw,col).value = inp.to_s
      lang[inp].paths.each do |path| 
        max_row = 0 
        col_offset = 0
        path.each do |comp|
          basis_hs_comp_to_ws(ws, constraints, comp, rw, col+col_offset, comp[0].input)
          # Each competition display has a col for each constraint + opt + output
          col_offset += constraints.size + 3    
          max_row = comp.size+1 if max_row < comp.size+1
        end 
        rw+=max_row+1
      end  
    end
  end
  
  # To Do
  def derivations_to_ws_top_to_bottom(ws,lang,rw,col, sorted_cons)      
    arrow = "\u2193"
    lang.keys.reverse.each do |inp|
      #range(ws,rw,col,rw,col).value = inp.to_s
      lang[inp].paths.each do |path| 
        max_col = 0 
        row_offset = 0
        path.each do |comp|
          basis_hs_comp_to_ws(ws, sorted_cons, comp, rw+row_offset, col, comp[0].input)
          # Each competition display has a col for each constraint + opt + output
          
          if path.index(comp) != path.size - 1
            ws.cells(rw+row_offset+comp.size+1, col).value = arrow
            ws.cells(rw+row_offset+comp.size+1,col).font.bold = true
            ws.cells(rw+row_offset+comp.size+1,col).font.color = -16776961
            ws.cells(rw+row_offset+comp.size+1,col).horizontalalignment = -4108    
          end

          row_offset += comp.size + 2   
        end 
        col+=sorted_cons.size+3
      end  
    end
  end
      
  # Format a comparative tableau of basis ercs, with only one pre-constraint
  # column for Residue#, and write it to the specified section of WS.
  def basis_hs_comp_to_ws(ws,cons,comp,row_first,col_first, inp)
    #range(ws,row_first, col_first, row_first, col_first).value = "VALUE"
    
    # Place the winning candidate last
    winner = comp.winners[0]
    comp.delete(winner)
    comp << winner

    # first row contains the column headers
    sheet_image = [] << (row_image = [])
    row_image.concat([inp, "opt"])
    pre_con_columns = 2
    row_image.concat(cons.map{|con| con.to_s})
    # add the erc rows to the sheet image
    comp.each do |cand|
      sheet_image << (row_image = []) # elements of the eventual output row
      cand.opt? ? opt_v = "x" : opt_v = ""
      row_image.concat([cand.output, opt_v])
      cons.each {|c| row_image.concat([c.eval(cand)])}
    end
    
    # write the sheet_image array to the worksheet
    row_last = row_first + sheet_image.size - 1
    col_last = col_first + pre_con_columns + cons.size - 1
    range(ws,row_first,col_first,row_last,col_last).value = sheet_image
    vt_borders(ws,row_first,col_first,row_last,col_last,pre_con_columns)
    return row_last, col_last

  end
  
  ############################################################## 
  #MOJO SPECIFIC put method. AP. 4-21-12
  
  def put_mojo(fred_result, freds, con_check)
      rcd_result = fred_result.rcd_result
      # Add the unranked constraints as a "final stratum" to the hierarchy.
      hier_with_unranked = Hierarchy.new
      hier_with_unranked.concat(rcd_result.hierarchy)
      hier_with_unranked << rcd_result.unranked unless rcd_result.unranked.empty?
      sorted_cons = hier_with_unranked.flatten
      # sort the ercs with respect to the RCD constraint hierarchy
      sorted_ercs, ercs_by_stratum, explained_ercs = sort_rcd_results(rcd_result)
      #
      @excel.screenUpdating=false # turn off updating while writing to a worksheet
  
      # TODO: (BBT) ultimately, properties like tabcolor should be stored in some
      # object which is constructed when the command is first begun.
      # 
      # Determine the appropriate tab color index for the new sheet.
      # * If the current sheet is a factorial typology sheet, take the color from
      # the preceding sheet.
      # * Otherwise (current sheet is a just-generated FRed sheet), take the color
      # from the current sheet.
      active_ws = @excel.activesheet
      if /FacTyp/ =~ active_ws.name
        cs_index = active_ws.index - 1
      else
        cs_index = active_ws.index
      end
      color_index = @excel.sheets(cs_index).tab.colorindex
  
      # Create a new output worksheet, and set names, labels, and tab
      # color according to the OTWP standard.
      # Cell 1,1: project name   Cell 1,2: status stamp   Cell 1,3: language name
      ws = @excel.sheets.add('After'=>active_ws)
      ws.name = name_sheet("#{@proj_name}.#{fred_result.label}")
      ws.tab.colorindex = color_index
      ws.cells(1,1).value = @proj_name
      status_stamp(ws,"MoJo")
      range(ws,1,3,1,3).value = sorted_ercs.label 
  
      if con_check
        # Conservative join check
        #revised Feb 8, 2014 -AP
        cjoin = CJoin.new(fred_result.constraint_list)
        j_leg, u_leg = cjoin.c_join_counts(freds)
        #range(ws, 1, 5, 1, 5).value = "Join legs #{j_leg}"
        #range(ws, 1, 6, 1, 6).value = "Union legs #{u_leg}"
        #modified location and format Feb 8, 2014.
        ws.cells(3,2).value="Join legs"
        ws.cells(3,3).value=j_leg
        ws.cells(4,2).value="Union legs"
        ws.cells(4,3).value=u_leg
        if j_leg-u_leg>0 
           ws.cells(5,3).value="NONCON"
         else
           ws.cells(5,3).value="CONSERV"
        end
        #ws.cells(5,3).HorizontalAlignment=xlRight    #doesn't work
      end

      # PUT MoJo SKB
      range(ws,2,5,2,5).value = "SKB of Join"
      row_last, col_last =
        basis_ercs_to_ws(ws,sorted_cons,fred_result.skb,3,5)
  
      #PUT MoJo MIB
      range(ws,row_last+2,5,row_last+2,5).value = "MIB of Join"
        row_last, col_last =
          basis_ercs_to_ws(ws,sorted_cons,fred_result.mib,row_last+3,5)
  
  ###cd easily switch to mib in above line by calling fred_result.mib
  
      # DO NOT PUT the support for the MoJo SKB
      ##range(ws,row_last+2,1,row_last+2,1).value = "Support"
      ##row_last, col_last =         ##skb_support_to_ws(ws,sorted_cons,fred_result.fred_kept_steps,row_last+3,1)
      
         # DO NOT PUT full set of winner-loser pairs
     ## range(ws,row_last+2,2,row_last+2,2).value = "WL Pairs"
     ## row_first_wlp = row_last+3
     ## col_first_wlp = 2
     ## row_last, col_last =
       ## wl_pairs_to_ws(ws,sorted_cons,sorted_ercs,row_first_wlp,col_first_wlp)
        
      # Format RCD Display with VBA macro "rcdPretty"   -AP
  #    @excel.cells(row_last,2).select
  #    @excel.run "'#{find_otwpl}'!rcdPretty"
      # Set parameters for the tableau to be formatted
  #    pre_erc_rows = 1
  #    pre_con_columns = 4
      # Create borders demarking stratified constraints, ercs
  #    stratified_tableau_borders(ws,row_first_wlp,col_first_wlp,pre_erc_rows,
  #      pre_con_columns,hier_with_unranked,ercs_by_stratum)
      # Apply RCD-based tableau coloring
  #    rcd_tableau_coloring(ws,row_first_wlp,col_first_wlp,pre_erc_rows,
   #     pre_con_columns,hier_with_unranked,ercs_by_stratum,explained_ercs)
  
      # Find the diagram bases for the SKB.
      skb_exp = Skb_expansion.new(fred_result.skb)
      db = skb_exp.diagram_bases
  
      show_hasse = true
      #show_hasse = false
      if show_hasse
        # For each diagram basis, create a cell with a pop-up comment
        # containing the ranking diagram.
        row_last += 2
        dcount = 0
        dist=3
        rw = 0
        db.each do |basis|
          dcount += 1
          label = "Diagram#{dcount}"                                   
          rd = RankingDiagram.new(basis, label)
          tempfile = File.join(File.dirname(__FILE__),'..','temp',label)
          rd.png(tempfile)
          rw = rw+dist                                           #AP/NM-added
          @excel.columns(col_last+1).columnwidth = 3             #AP/NM- added
          cell = range(ws,rw,col_last+2,rw,col_last+2)           #AP/NM- modified
          cell.value = label
          put_png_comment(tempfile+'.png', cell)
          cell.comment.visible = true                            #AP/NM- added
          dist = (cell.comment.shape.height/12).to_i + 1         #AP/NM- added
        end
      end
  
      return ws
    ensure 
  # make sure screen updating is turned back on, even if an exception is raised.
      @excel.screenUpdating=true    
    end
  
  #########################  Back to your regularly scheduled program
  

  
  def put_learning_results(lang_sim)
    @excel.screenUpdating=false # turn off updating while writing to a worksheet
    hyp = lang_sim.hypothesis
    ws = @excel.sheets.add('After'=>@excel.activesheet)
    ws.name = name_sheet("Learn." + hyp.label)
    range(ws,1,1,1,1).value = hyp.label
    row = 1; col = 1
    #
    lang_sim.results_list.each do |entry|
      row += 3
      range(ws,row,1,row,5).merge
      range(ws,row,1,row,1).value = "#{entry.label}:"
      row, col = learning_result_to_ws(ws, entry, row+1, 1)
    end
    #
  ensure # make sure screen updating is turned back on, even if an exception is raised.
    @excel.screenUpdating=true    
  end
  
  def learning_result_to_ws(ws, gram_test_result, row_first, col_first)
    # Test result components are frozen, so dup before updating.
    hyp = gram_test_result.hypothesis.dup
    rcd_result = hyp.update_grammar{|ercs| OTLearn::RcdFaithLow.new(ercs)}
    # Add the unranked constraints as a "final stratum" to the hierarchy.
    hier_with_unranked = Hierarchy.new
    hier_with_unranked.concat(rcd_result.hierarchy)
    hier_with_unranked << rcd_result.unranked unless rcd_result.unranked.empty?
    sorted_cons = hier_with_unranked.flatten
    # sort the ercs with respect to the RCD constraint hierarchy
    sorted_ercs, ercs_by_stratum, explained_ercs = sort_rcd_results(rcd_result)
    # write the main CT to the new worksheet
    wl_pairs_to_ws(ws,sorted_cons,sorted_ercs,row_first,col_first)

    # Set some table index values
    pre_con_columns = 4
    col_count = pre_con_columns + sorted_cons.size
    pre_erc_rows = 1
    last_ex_row = pre_erc_rows + explained_ercs.size # row of last explained erc
    row_count = last_ex_row + rcd_result.unex_ercs.size
    
    # put vertical lines between the strata
    vl_col_count = pre_con_columns
    hier_with_unranked.each do |stratum|
      vl_col_count += stratum.size
      vl_row_last = row_first-1+row_count
      vl_col_last = col_first-1+vl_col_count
      range(ws,row_first,vl_col_last,vl_row_last,vl_col_last).borders(XlEdgeRight).weight =
        XlMedium # vertical line between strata
    end
    # put horizontal lines between the "stratified" clusters of ercs
    hl_row_count = pre_erc_rows
    ercs_by_stratum.each do |ercs|
      hl_row_count += ercs.size
      hl_row_last = row_first-1+hl_row_count
      hl_col_last = col_first-1+vl_col_count
      range(ws,hl_row_last,col_first,hl_row_last,hl_col_last).borders(XlEdgeBottom).weight =
        XlMedium
    end
    # Flag any unexplained ercs with color
    range(ws,row_first+last_ex_row,1,row_first-1+row_count,1).interior.colorindex = RED unless row_count == last_ex_row
    # Extra formatting if the data were inconsistent
    range(ws,row_first+row_count,1,row_first+row_count,1).value = "FAIL!" unless rcd_result.consistent?

    # Put the lexicon to the worksheet
    first_lex_row = row_first-1 + row_count + 2
    lex = hyp.grammar.lexicon
    r_row = first_lex_row-1
    r_row, r_col = morphs_to_ws(ws, lex.get_prefixes, r_row+1, col_first) unless lex.get_prefixes.empty?
    r_row, r_col = morphs_to_ws(ws, lex.get_roots, r_row+1, col_first)
    r_row, r_col = morphs_to_ws(ws, lex.get_suffixes, r_row+1, col_first) unless lex.get_suffixes.empty?
    
    test_row = r_row + 2
    range(ws,test_row,col_first,test_row,col_first+2).merge
    range(ws,test_row,col_first,test_row,col_first).value = "Learned: #{gram_test_result.all_correct?}"
    return test_row, col_first-1+col_count
  end
  
  # Puts a set of morpheme lexical entries into a single excel row.
  def morphs_to_ws(ws, morphs, row_first, col_first)
    row = row_first
    col = col_first - 3
    morphs.each do |morph|
      col += 3
      range(ws,row,col,row,col).value = morph.label.to_s
      range(ws,row,col+1,row,col+1).value = morph.uf.to_s      
    end
    return row_first, col+1
  end
  
private  

  # Creates a new workbook, deletes the two extra worksheets (workbooks are
  # created with 3 by default), renames the remaining worksheet to the parameter, 
  # and returns the workbook.
  # Normally, this method should only be called by other methods in class Excel.
  def add_output_workbook ws_name
    @output_wb = @excel.workbooks.add()
    @output_wb.worksheets("Sheet2").delete
    @output_wb.worksheets("Sheet3").delete
    ws = @output_wb.worksheets("Sheet1")
    ws.name = ws_name
    return @output_wb
  end

  # Converts the raw data from the worksheet, raw_input,
  # into a competition list. Each competition in the list contains all
  # of the candidates for a particular input.
  # TODO: more actual input verification: are the format assumptions obeyed?
  def convert_input_data(raw_input)
    # The label for the datasheet
    data_sheet_label = raw_input[0][0]
    competition_list = Competition_list.new(data_sheet_label)
    # define the array ranges for the different parts of the raw data
    first_cand = 2
    last_cand = raw_input.size-1
    # check if the candidates are already numbered
    if !raw_input[1][0].nil? and raw_input[1][0].to_str.strip.upcase=='CAND#' then
      user_numbering = true # use the existing numbering
      input_col = 1 # column 0 contains the existing candidate numbers
    else
      user_numbering = false # generate candidate numbers
      input_col = 0 # no candidate number column is present.
    end
    # Set the other numbers for defined columns
    output_col = input_col + 1
    opt_col = output_col + 1
    first_con = opt_col + 1
    # Constraint columns go from first_con up to the first following empty cell
    # in row 1.
    last_con = first_con-1
    last_con += 1 until raw_input[1][last_con+1].nil?
    # check for no candidates / no constraints
    return competition_list if (last_cand < first_cand) or (last_con < first_con)
    # set the column range for constraints and the row range for candidates
    con_range = (first_con..last_con)
    cand_range = (first_cand..last_cand)
    vt_to_comp_list(raw_input, con_range, cand_range, input_col, output_col, opt_col, user_numbering, competition_list)
    competition_list.auto_number_candidates unless user_numbering
    return competition_list
  end

  # Extract the violation tableau data from a raw input array, construct constraints,
  # candidates, and competitions, and return a competition list.
  def vt_to_comp_list(raw_input, con_range, cand_range, input_col, output_col, opt_col, user_numbering, competition_list)
    # Get the constraint labels from the second row (index 1)
    constraints = extract_constraints(raw_input[1], con_range)
    # Construct a candidate for each candidate row
    competition = Competition.new
    input_form = raw_input[cand_range.first][input_col] # need to define variables outside the block for persistency.
    cand_range.each do |row|
      # If a new input is found, close one competition and start the next one.
      if raw_input[row][input_col] && (raw_input[row][input_col] != input_form)
        input_form = raw_input[row][input_col] # update the input
        competition_list << competition # add the previous competition to the list
        competition = Competition.new # start a new competition
      end
      # Create a candidate for the current row
      cand = Candidate.new(input_form, raw_input[row][output_col], raw_input[row][opt_col], constraints)
      if user_numbering then
        cand.label = raw_input[row][0] # Use the candidate number from the first column.
      end
      con_range.each do |col|
        raw_input[row][col] = 0 if !(raw_input[row][col]) # replace blanks with zeroes
        cand.set_viols(constraints[col-con_range.first], raw_input[row][col].to_i) # convert counts to integers
      end
      cand.remark = (raw_input[row][con_range.last+2]).to_s
      competition.push(cand)
    end
    competition_list << competition # add the final competition to the list
    return competition_list
  end
  
  # Extract the constraint information from the header row of a tableau.
  # Input is the header row (an array of text values), and the range within
  # that row that contains constraint labels. If the constraints aren't already
  # numbered, applies automatic numbering to them.
  def extract_constraints(raw_input_row, con_range)
    constraints = []
    con_range.each{|col| constraints[col-con_range.first] = raw_input_row[col]}
    # Construct constraint objects
    constraints.each_index do |i|
      # If constraints are already numbered, keep that numbering.
      con_str = constraints[i].to_s
      if (con_str =~ /^(\d+):/) then
        constraints[i] = Constraint.new($POSTMATCH,$1)
      else # otherwise, number them from 1 in the order they appear.
        constraints[i] = Constraint.new(con_str,i+1)
      end
    end
    return constraints
  end
  
  # Format a violation count tableau, and write it to the specified section
  # of the worksheet.
  def viol_counts_to_ws(ws, competition_list, row_first, col_first)
    constraints = competition_list.constraint_list
    # currently six non-constraint columns: 4 before the constraints, 2 after
    col_count = 4 + constraints.size + 2
    sheet_image = []
    # Column header row first
    sheet_image << (row_image = [])
    row_image.concat(["Cand\#", "Input", "Output", "Opt"])
    row_image.concat(constraints.map{|con| con.to_s})
    row_image.concat([nil, "Remarks"])
    # Candidate rows
    competition_list.each do |comp|
      comp.each do |cand|
        cand_a = cand.to_a # convert the candidate to an array, one element per cell
        sheet_image << (row_image = [])
        row_image.concat(cand_a) # put the candidate array into the next row
        row_image[3] = cand.merged_outputs_to_s if cand.merged? # combine outputs of merged candidates
        (col_count-cand_a.size).times{row_image << nil} # pad the row with nil
      end
    end
    # Compute the row/column of the last cell.
    row_count = sheet_image.size
    row_last = row_first + row_count - 1
    col_last = col_first + col_count - 1
    # Format the label column as text. NOTE: this MUST be done BEFORE writing
    # the sheet image to the worksheet, or the labels won't be formatted correctly.
    range(ws,row_first,col_first,row_first,col_first).EntireColumn.NumberFormat = "@"
    # write the sheet image to the worksheet
    range(ws,row_first,col_first,row_last,col_last).value = sheet_image
    # autosize the columns
    range(ws,row_first,col_first,row_last,col_last).entirecolumn.autofit
    # apply the standard tableau formatting, leaving out the remarks columns
    col_no_remarks = row_first+4+constraints.size-2
    ct_borders(ws,row_first,col_first,row_last,col_no_remarks,4)
    # Add horizontal lines separating the competitions
    hl_row_count = row_first
    competition_list.each do |comp|
      hl_row_count += comp.size
      range(ws,hl_row_count,col_first,hl_row_count,col_no_remarks).borders(XlEdgeBottom).weight = XlMedium
    end
    return row_last, col_last
  end

  # Format a comparative tableau, and write it to the specified section of the worksheet.
  def wl_pairs_to_ws(ws, sorted_cons, sorted_ercs, row_first, col_first)
    # first row contains the column headers
    sheet_image = [] << (row_image = [])
    row_image.concat(["ERC\#", "Input", "Winner", "Loser"])
    pre_con_columns = 4
    row_image.concat(sorted_cons.map{|con| con.to_s})
    # add the erc rows to the sheet image
    sorted_ercs.each do |erc|
      sheet_image << (row_image = []) # elements of the eventual output row
      row_image << erc.label
      if erc.respond_to?(:winner) then # pair contains a winner and a loser
        row_image << erc.winner.input.to_s << erc.winner.merged_outputs_to_s
        row_image <<  erc.loser.output.to_s
      else
        3.times{row_image << nil} # base ercs don't have winner or loser
      end
      add_prefs_to_row(erc, sorted_cons, row_image)
    end
    # write the sheet_image array to the worksheet
    row_last = row_first + sheet_image.size - 1
    col_last = col_first + pre_con_columns + sorted_cons.size - 1
    range(ws,row_first,col_first,row_last,col_last).value = sheet_image
    ct_borders(ws,row_first,col_first,row_last,col_last,pre_con_columns)
    return row_last, col_last
  end

  # Format a comparative tableau, and write it to the specified section of WS.
  def skb_support_to_ws(ws, sorted_cons, steps, row_first, col_first)
    # first row contains the column headers
    sheet_image = [] << (row_image = [])
    row_image.concat(["Residue\#", "ERC\#", "Input", "Winner", "Loser"])
    pre_con_columns = 5
    row_image.concat(sorted_cons.map{|con| con.to_s})
    # add the erc rows to the sheet image
    row_cur = row_first
    steps.each do |step|
      step.support.each do |erc|
        row_cur += 1
        sheet_image << (row_image = []) # elements of the eventual output row
        row_image << step.comp_tableau.label << erc.label
        if erc.respond_to?(:winner) then # pair contains a winner and a loser
          row_image << erc.winner.input.to_s << erc.winner.output.to_s
          row_image <<  erc.loser.output.to_s
        else
          3.times{row_image << nil} # base ercs don't have winner or loser
        end
        add_prefs_to_row(erc, sorted_cons, row_image)
        # NOTE: next loop colors actual cells of the worksheet, while the
        # preceding code wrote content data to an image array; it won't be
        # written to the worksheet until later.
        col_cur = col_first + pre_con_columns - 1
        sorted_cons.each do |con|
          col_cur += 1
          # If the evaluation doesn't survive into the SKB erc, color it.
          if !erc.e?(con) && step.skb_erc.e?(con) then
            range(ws,row_cur,col_cur,row_cur,col_cur).interior.colorindex = PALEYELLOW
          end
        end
      end
    end
    # write the sheet_image array to the worksheet
    row_last = row_first + sheet_image.size - 1
    col_last = col_first + pre_con_columns + sorted_cons.size - 1
    range(ws,row_first,col_first,row_last,col_last).value = sheet_image
    ct_borders(ws,row_first,col_first,row_last,col_last,pre_con_columns)
    return row_last, col_last
  end

  # Format the kept residues generated by FRed, and write them to the WS.
  def residue_ercs_to_ws(ws, sorted_cons, step, row_first, col_first)
    # Sort the residue ercs with respect to the sorted constraints.
    residue = Comparative_tableau.new(step.comp_tableau.label)
    residue.concat(sort_by_constraint_order(step.comp_tableau,sorted_cons))
    # first row contains the column headers
    sheet_image = [] << (row_image = [])
    row_image.concat([residue.label])
    pre_con_columns = 1
    row_image.concat(sorted_cons.map{|con| con.to_s})
    range(ws,row_first,col_first,row_first,col_first).font.bold = true
    # add the erc rows to the sheet image
    row_cur = row_first
    residue.each do |erc|
      row_cur += 1
      sheet_image << (row_image = []) # elements of the eventual output row
      row_image << erc.label
      add_prefs_to_row(erc, sorted_cons, row_image)
      # NOTE: next loop colors actual cells of the worksheet, while the
      # preceding code wrote content data to an image array; it won't be
      # written to the worksheet until later.
      if step.support.find{|e| erc.label==e.label} then
        range(ws,row_cur,col_first,row_cur,col_first).interior.colorindex = MIDYELLOW
        col_cur = col_first + pre_con_columns - 1
        sorted_cons.each do |con|
          col_cur += 1
          # Color the cells relative to the corresponding SKB erc
          if erc.w?(con) && step.skb_erc.w?(con) then
            range(ws,row_cur,col_cur,row_cur,col_cur).interior.colorindex = LIGHTGREEN
          elsif erc.l?(con) && step.skb_erc.l?(con) then
            range(ws,row_cur,col_cur,row_cur,col_cur).interior.colorindex = MAGENTA
          elsif !erc.e?(con) then
            range(ws,row_cur,col_cur,row_cur,col_cur).interior.colorindex = PALEYELLOW
          end
        end
      end
    end
    # write the fusion in the row below
    sheet_image << (row_image = []) # elements of the eventual output row
    row_image << step.fusion.label
    add_prefs_to_row(step.fusion, sorted_cons, row_image)
    # write the sheet_image array to the worksheet
    row_last = row_first + sheet_image.size - 1
    col_last = col_first + pre_con_columns + sorted_cons.size - 1
    range(ws,row_first,col_first,row_last,col_last).value = sheet_image
    ct_borders(ws,row_first,col_first,row_last,col_last,pre_con_columns)
    range(ws,row_last,col_first,row_last,col_last).borders(XlEdgeTop).weight = XlMedium # above fusion
    first_con = col_first + pre_con_columns
    range(ws,row_first+1,first_con,row_last-1,col_last).font.bold = false
    # Add Status column header after constraint headers, the Keep indicator after the fusion
    range(ws,row_first,col_last+1,row_first,col_last+1).value = "Status"
    range(ws,row_last,col_last+1,row_last,col_last+1).value = "KEEP"
    range(ws,row_last,col_last+1,row_last,col_last+1).font.bold = true
    range(ws,row_last,col_last+1,row_last,col_last+1).interior.colorindex = BRIGHTGREEN
    return row_last, col_last+1 # return the lowest row and further column used
  end

  # Format a comparative tableau of basis ercs, with only one pre-constraint
  # column for Residue#, and write it to the specified section of WS.
  def basis_ercs_to_ws(ws,sorted_cons,sorted_ercs,row_first,col_first)
    # first row contains the column headers
    sheet_image = [] << (row_image = [])
    row_image.concat(["Residue\#"])
    pre_con_columns = 1
    row_image.concat(sorted_cons.map{|con| con.to_s})
    # add the erc rows to the sheet image
    sorted_ercs.each do |erc|
      sheet_image << (row_image = []) # elements of the eventual output row
      row_image << erc.label
      add_prefs_to_row(erc, sorted_cons, row_image)
    end
    # write the sheet_image array to the worksheet
    row_last = row_first + sheet_image.size - 1
    col_last = col_first + pre_con_columns + sorted_cons.size - 1
    range(ws,row_first,col_first,row_last,col_last).value = sheet_image
    ct_borders(ws,row_first,col_first,row_last,col_last,pre_con_columns)
    return row_last, col_last
  end

  # Draws standard violation tableau formatting
  def vt_borders(ws,row_first,col_first,row_last,col_last,pre_con_columns)
    # autosize the columns
    range(ws,row_first,col_first,row_last,col_last).entirecolumn.autofit
    # format the erc evaluation cells
    first_con = col_first + pre_con_columns # first column for a constraint
    #range(ws,row_first+1,first_con,row_last,col_last).font.bold = true
    range(ws,row_first+1,first_con,row_last,col_last).horizontalAlignment = XlCenter # center justification
    # Set the borders
    range(ws,row_first,col_first,row_last,col_last).borders.weight = XlThin  # basic grid
    range(ws,row_first,col_first,row_first,col_last).borders(XlEdgeTop).weight = XlThin # top of table
    range(ws,row_last,col_first,row_last,col_last).borders(XlEdgeBottom).weight = XlMedium # bottom of table
    range(ws,row_first,col_first,row_first,col_last).borders(XlEdgeBottom).linestyle = XlDouble # top double line
    range(ws,row_first,first_con-1,row_last,first_con-1).borders(XlEdgeRight).linestyle = XlDouble # vertical double line
    return ws    
  end

  # Draws standard comparative tableau formatting
  def ct_borders(ws,row_first,col_first,row_last,col_last,pre_con_columns)
    # autosize the columns
    range(ws,row_first,col_first,row_last,col_last).entirecolumn.autofit
    # format the erc evaluation cells
    first_con = col_first + pre_con_columns # first column for a constraint
    range(ws,row_first+1,first_con,row_last,col_last).font.bold = true
    range(ws,row_first+1,first_con,row_last,col_last).horizontalAlignment = XlCenter # center justification
    # Set the borders
    range(ws,row_first,col_first,row_last,col_last).borders.weight = XlThin  # basic grid
    range(ws,row_first,col_first,row_first,col_last).borders(XlEdgeTop).weight = XlMedium # top of table
    range(ws,row_last,col_first,row_last,col_last).borders(XlEdgeBottom).weight = XlMedium # bottom of table
    range(ws,row_first,col_first,row_first,col_last).borders(XlEdgeBottom).linestyle = XlDouble # top double line
    range(ws,row_first,first_con-1,row_last,first_con-1).borders(XlEdgeRight).linestyle = XlDouble # vertical double line
    return ws    
  end
  
  def add_prefs_to_row(ct_row, sorted_cons, row_image)
    sorted_cons.each do |con|
      if ct_row.l?(con)
        row_image << "L"
      elsif ct_row.w?(con)
        row_image << "W"
      else
        row_image << nil  # Leave 'e' cells blank, for readability
      end
    end
    row_image
  end

  # Returns an excel range object from the begin cell to the end cell,
  # relative to the parameter base, which can be a worksheet or a range.
  def range(base, row_begin, col_begin, row_end, col_end)
    #ws.range(ws.cells(row_begin,col_begin),ws.cells(row_end,col_end))
    base.range(range_to_s(row_begin, col_begin, row_end, col_end))
  end

  # Converts an integer column designation to the equivalent letter string
  # for Excel: 1 -> "A", 26 -> "Z", 27 -> "AA", 29 -> "AC", etc.
  def col_to_letter(col)
    raise "col_to_letter(): Cannot have a nonnegative column index." if col<1
    letcol = "A"
    (col-1).times{letcol.succ!}
    return letcol
  end

  # Converts row/column integer designations of a range into Excel letter/number
  # format. (1,2,3,4) -> "B1:D3"
  def range_to_s(row_first, col_first, row_last, col_last)
    "#{col_to_letter(col_first)}#{row_first}:#{col_to_letter(col_last)}#{row_last}"
  end

  # Returns the size of the PNG image stored in the file of name _fname_,
  # as an array [width, height] giving the values in points.
  def png_size(fname)
    # Read the first 24 bytes of the image file into a string.
    prefix = nil
    File.open(fname, mode: "rb"){|fh| prefix = fh.read(24)}
    # A PNG file has bytes 1-3 with the value "PNG".
    raise "file is not PNG" unless prefix[1..3] == "PNG"
    # Read the picture size (in points) from bytes 16-24.
    return prefix[16..24].unpack('NN')
  end

  # Adds a pop-up comment to _cell_, and fills it with the PNG image
  # stored in file _fname_.
  def put_png_comment(fname, cell)
    # get the dimensions (in points) of the image
    width, height = png_size(fname)
    cell.clearcomments
    cbox = cell.addcomment # returns a reference to the comment box
    cbox.shape.fill.userpicture(fname) # fill with the picture
    cbox.text " " # to blank out the inserted username.
    # resize the comment box shape to match the picture
      #cbox.shape.shadow.visible = false                         #AP-added
    cbox.shape.width = width/2                                #AP/NM-added
    cbox.shape.height = height/2                              #AP/NM-added
  end

public
  #two functions for licit sheet naming -AP
  #format is NAME or NAME{n} for nth repetition
  def name_sheet(tryname)
    allsheets=[]
    @excel.sheets.each {|sht| allsheets << sht.name} #all sheet names in active workbook
    pref_name = tryname.split("{")[0] #what we're trying to name it minus any numbering
    pref_name.delete!('?/:[]*\\') #dump invalid characters (double backslash must be last)
    pref_name = pref_name[0...20] #keep it short: only 31 characters max in sheet name
    # Collect every sheet name of the same name prior to '{'
    all_dups = allsheets.find_all {|x| x.upcase =~ /^#{Regexp.escape(pref_name.upcase)}(\{|$)/ }
    return pref_name if all_dups.empty?  # if no duplicate names, no problem
    # pick out the highest duplicate index, and add 1 to get the new index
    biggest = find_max_rep_index(all_dups) + 1
    return pref_name += "\{#{biggest.to_s}\}"
  end
  
  # find the largest sheet name index of the form {N}.
  def find_max_rep_index (bilge)
    indices = bilge.map do |x|
      if x =~ /\{(\d+)/ then # if there are digits following '{'
        $1.to_i # convert those digits to an integer
      else
        1 # convert no match to 1
      end
    end
    return indices.max # return the largest index
  end

  def nameMOJO
    @excel.run "'#{find_otwpl}'!nameMOJO"
  end

  def nameMyLanguage
    @excel.run "'#{find_otwpl}'!nameMyLanguage"
  end

  def my_method (val)
    ws = @excel.activesheet
    range(ws,1,1,1,1).value = val
  end

def fixRnkRptsAll                                             #ASP FIX
    @excel.run "'#{find_otwpl}'!fixRnkRptsAll"
  end

def myMethod  #test method  -AP
  #@excel.activesheet.cells(2,1).value = @excel.activesheet.cells.find('What'=>"WL Pairs", 'after'=>@excel.activesheet.cells(1,1), 'SearchOrder'=>'xlByColumns', 'lookat'=>'xlByPart').row 
  ws=@excel.activesheet
  lang_sht_name=ws.name
  ftSheet= @excel.sheets(ws.index-1)
  lang_name = ws.cells(1,3).value
  myRow = ftSheet.cells.find('What'=>lang_name,'LookAt'=>XlWhole,'MatchCase'=>true,'after'=>ftSheet.cells(1,1)).row   #ASP FIX
  @excel.sheets(ftSheet.name).hyperlinks.add('anchor'=>ftSheet.cells(myRow,1),'address'=>"",'subaddress'=>"\'#{lang_sht_name}\'!R1C1",'TextToDisplay'=> ftSheet.cells(myRow,1).value)
          

  #vba find parameters
  #(What:="L", after:=testCell, SearchOrder:=xlByColumns, lookAt:=xlPart)
  #@excel.activesheet.cells(1,3).value="whoop\'"
  #@excel.sheets.add 'after'=>@excel.activesheet
#@excel.activesheet.hyperlinks.add 'anchor'=>@excel.activesheet.cells(1,1) , 'address'=>"", 'subaddress'=>"\'INDEX\'!R1C1",'TextToDisplay'=>"Whoopee"
end


end # class Excel_session
