# Author: Naz Merchant
#

# A HS operation consisting of a source string and a target string
#
# Operations take an input string and produce an array of candidates,
# one for each location in the input string that the operation can apply to.

require_relative 'constraint.rb'
require_relative 'candidate_hs.rb'

class Operation
  
  # Format for operations:
  #
  # => For type A, source must be an array strings & target must be 
  # => an array of strings. Source and target must have the same number of elements.
  #
  #    type source     target
  #    A    abc        xyz         <-- each instance of 'a' is replaced with 'x', 'b' w/ 'y', etc
  #    B    [ab]_[cd]  x           <-- 'x' is inserted between a&c, a&d, b&c, and b&d
  #    C    abc        ""          <-- each a,b,&c is deleted
  
  # Type A&C are identical in structure.
  
  # source & target must be arrays of string
  
  def initialize(name, source, target, constraints)
    @name = name
    @source = source
    @target = target
    #Need constraints to create candidates
    @cons = constraints
  end
  
  def create_candidates(input)
    cands = Array.new
     
    (0..@source.length-1).each do |so_i|
      so = @source[so_i]
      so_l = so.length
      (0..input.length-1).each do |i|
        # Is it even poss. to match source?
        if i+so_l-1 < input.length
          if input[i..i+so_l-1] == so
            output = input.dup
            output[i..i+so_l-1] = @target[so_i]
            cands << HS_Candidate.new(input, output, false, @cons, self)
          end
        end
      end
    end


=begin
    # Type A
    for source_index in 0..@source.length-1
      source_str = @source[source_index]
      target_str = @target[source_index]  # Requires that source.length = target.length
     
      # Deal with source_str in initial position
      if input.index(source_str) == 0
        output = target_str + input[source_str.length..input.length]
        cands << HS_Candidate.new(input, output, false, @cons, self)
      end
     
      # Deal with source_str elsewhere
      pos = 1
      while pos != nil && pos < input.length
        pos = input.index(source_str, pos)
        if pos != nil
          output = input[0..pos-1] + target_str + input[pos+source_str.length..input.length]
          pos += source_str.length
          cands << HS_Candidate.new(input, output, false, @cons, self)
        end
      end
    end
=end
    
    set_violations(cands)
    return cands
  end
  
  # Requires that constraints are HS constraints
  def set_violations(cands)
    cands.each{|cand| @cons.each{|con| cand.set_viols(con, con.eval(cand))}}
  end
  
  # Returns the name of the constraint.
  def name() @name end
  
  # Two operations are equivalent if their names, sources & targets are equivalent.
  # This needs to be modified to include source & target equivalence.
  def ==(other)
    @name == other.name
  end

  # The same as ==
  def eql?(other)
    self == other
  end

  def to_s()
    @name.to_s + ": " + @source.to_s + " --> " + @target.to_s
  end
  
end