# Author: Naz Merchant
#

# A basic HS constraint, consisting of a name, an id, and a regular expression object
# that allows for evaluation of violations of a candidate.

# Constraint_HS subclasses Constraint, a basic OT constraint. The regex is added
# since candidates in HS can be created on the fly in Ruby contra what happens in the OT system.

require_relative 'constraint.rb'
require_relative 'candidate.rb'

class Constraint_HS < Constraint
  
  # Returns a constraint with the given name and id.
  # The id is an abbreviated label used for constructing labels for
  # complex objects, e.g., residues in *FRed*.
  def initialize(name, id, regex)
    super(name, id)
    @regex = regex
  end
  
  def eval(candidate)
    return candidate.output.scan(@regex).length
  end
end