# Author: Naz Merchant 
#

require_relative 'competition'
require_relative 'competition_list'
require_relative 'comparative_tableau'
require_relative 'derivation'
require_relative 'path'
require_relative 'operation'
require_relative 'ot_utils'

# A FactorialTypologyHS2 object determines the set of languages defined
# by a constraint set, an operation set, and a set of inputs. It does so by calculating
# the set of derivations from the inputs for each total order, aggregating those
# total orders that produce the same set of derivations.
#
# Wanted: augment this so that it also calculates the operation vectors defined by a 
# typology and the partial orders imposed on the opvects by the languages in the typo.

class Factorial_Typology_HS_Op
  
  attr_reader :langs, :derivs
  
  def initialize(inputs, ops, cons)
    @inputs = inputs
    @ops = ops
    @cons = cons
    
    @derivs = Array.new
    @langs = Array.new
    calculate_all_langs
    determine_unique_langs
  end
  
  # This method produces a hashtable, one for each total order, consisting of
  # derivations, one for each input. The hashtable is the language for that total order.
  def calculate_all_langs
    # This should be changed so that *all* total orders aren't calculated, if possible
    hiers = OT_Utilities.create_all_permutations_as_hierarchies(@cons)
    hiers.each do |h| 
      # @derivs stores hashes of derivations. The keys of the hash table are the inputs.
      deriv_h = Hash.new
      @inputs.each {|inp| deriv_h[inp] = Derivation.new(inp, h, @ops)}
      @derivs << deriv_h
    end
  end
  
  # A language is a unique set of derivations, so determining the languages
  # consists of determining the unique derivations -- ranking information can then
  # be extracted from the derivs. The array @derivs may have duplicate derivations for
  # diff total orders -- only one is needed in the factyp.
  def determine_unique_langs
    @derivs.each do |dh|
      contains_dh = false
      @langs.each {|lang| contains_dh = true if derivation_hash_equal?(lang, dh)}   
      @langs << dh unless contains_dh
    end
  end
  
  def derivation_hash_equal? (dh1, dh2)
    dh1.keys.each {|key|return false unless dh1[key] == dh2[key]}
    return true
  end
  
  def get_comparative_tableau(lang)     
    ct = Comparative_tableau.new("", @cons)
    lang.values.each {|d| d.get_competition_list.each {|c| ct.add_competition(c)}}
    return ct
  end
  
=begin
  # Returns an array of comparative tableau, one for each language in the typology
  def get_languages_as_comparative_tableau
    cts = Array.new
    @langs.each {|d_ht| cts << get_comparative_tableau(d_ht)}
    return cts
  end
=end
 
  def to_s
    # Only display the derivations for testing purposes
    
    #s = "Derivations: \n"
    #@derivs.each {|ds| ds.each {|d| s += d.to_s + "\n"}}
    
    # Languages are simply derivations
    s = "Languages: \n"
    @langs.each {|l| s += l.to_s + "\n"}
    return s
  end
  
end
