# Author: Bruce Tesar
# 
 
require_relative 'competition.rb'
 
# A stratified constraint hierarchy. The top stratum of the hierarchy
# is first in the list, and so forth. Each stratum is a list of constraints.
class Hierarchy < Array
  
  # Returns a duplicate hierarchy, with distinct stratum objects,
  # but containing references to the very same constraint objects.
  def dup
    copy = Hierarchy.new
    self.each {|strat| copy << strat.dup}
    return copy
  end

  # Returns a string representation of the hierarchy, with strata
  # delimited by square brackets.
  def to_s
    out_str = ""
    self.each do |stratum|
      out_str << "["
      stratum.each {|con| out_str << con.to_s << " "}
      out_str.chomp!(" ") # remove trailing space
      out_str << "] << "
    end
    out_str.chomp!(" << ")
  end
  
  # NNM: Given a set of candidates (cand_set is an instrance of Competition class) produces
  # an array containing all the winning candidates based on this hierarchy, setting 
  # the winners to 'winner'. 
  #
  # Uses pooling of the marks to assess winners from strata with more than one constraint
  
  def winners (cand_set)
    win_cands = Array.new
    win_cands.concat(cand_set)
    self.each {|strat| win_cands = strat_winners(strat, win_cands)}
    return win_cands
  end
  
  def strat_winners(strat, cands)
    winners = Array.new
    min_viol = nil
    cands.each do |cand|
      viol_count = 0
      strat.each {|con| viol_count += cand.get_viols(con)}
      min_viol = viol_count unless min_viol != nil
      
      if viol_count == min_viol
        winners << cand
      elsif viol_count < min_viol
        min_viol = viol_count
        winners = Array.new
        winners << cand
      end
    end  
    return winners
  end

end # class Hierarchy
