# Author: Naz Merchant
# With mods by Alan Prince
#--
# This class implements the join algorithm (Merchant JOin) on a set of sets of ERCs.
# 
# References:
#
# Merchant 2008. {Contrast Pairs and Ranking.}[http://roa.rutgers.edu/view.php3?id=XXX]
# Merchant 2011. {Learning ranking information from unspecified overt forms using the join}
# [http://roa.rutgers.edu/view.php3?id=XXX]
#

require_relative 'erc'
require_relative 'fred'
require_relative 'comparative_tableau'

class Mojo

  attr_reader :fred_result
  attr_reader :join_result
  
  # Runs Mojo on a set of freds, using the MIB, producing a Fred object
  # as the result. 
  def initialize(freds)
    @label = ""                                    #Altered AP 5-21-13 and AP 6-18-13
    freds.each {|f| @label.concat(f.label + "_")}  #Ruby screws up on +
    @label = "JN." + @label.chomp("_")             #final demarcator dropped

    @fred_result = freds[0]
    freds.each do |fred|
      two_freds = Array.new
      two_freds << @fred_result
      two_freds << fred
      @fred_result = join_freds(two_freds)
    end
  end
  
  def join_freds(freds) 
    # Need to produce fusional closures
    fc_a = Array.new
    
    freds.each do |f|
      ct = Comparative_tableau.new
      f.fred_steps_list.each {|f_step| ct << f_step.fusion }
      fc_a << ct
    end
    
    #@freds.each {|f| @fc_a << fusionally_close(f.mib)}
    
    # Get all combinations of ERCs into arrays, selecting one
    # from each fusionally closed set.  TODO: iterate over constraints
    # and choose only those ERCs that have an L in the constraint
    # and only when all @fc_a sets have such an ERC.
    # 
    fc_product = Array.new
    fc_a.each { |ct| 
      if fc_product.empty?
        fc_product = ct
      else
        fc_product = fc_product.product(ct)
      end  
    }

    # Hack! If Mojo is called on only one language, then @mibs_product
    # is an array consisting of ERCs, and it needs to be an array
    # consisting of arrays of ERCs for conversion to cts to work.
    if freds.size == 1
      fc_product = [fc_product]
    end

    # array.product does not produce arrays of depth 1, which are needed
    # for the construction of Comparative_tableau
    fc_product.each {|ary| ary.flatten!}

    # Convert the sets of ERCs contained in @mibs_product to
    # Comparative tableau so that the Mojo.join can operate
    # over them.
    fc_product_ct = Array.new
    fc_product.each do |ary|
      ct = Comparative_tableau.new("temp label")
      ary.each {|erc| ct << erc}
      ct_label = ""
      ary.each {|erc| ct_label.concat (".#{erc.label}")}
      ct.label=ct_label
      fc_product_ct << ct
    end
    
    # Join each of the sets of Comparative tableau and put the result
    # in a FRed.
    join_result = Comparative_tableau.new(@label)
    fc_product_ct.each {|p| join_result << Mojo.join(p)}

    return Fred.new(join_result)
  end
  
  # Returns an erc that is the join of the ercs of the comparative
  # tableau _ct_.
  #
  # Each constraint assessment of the join is determined from the assessments
  # to that same constraint of each of the factor ercs.
  # * If a constraint is 'W' for any factor, it is 'W' in the join.
  # * If a constraint is 'e' for any factor, but not 'W' for any factor,
  #   then it is 'e' in the join.
  # * Otherwise, it must be 'L' in all factors, and is 'L' in the join.
  def Mojo.join(ct)
    return nil if ct.empty?
    constraints = ct.constraint_list
    join = Erc.new(constraints, "j#{ct.label}")
    constraints.each do |con|
      if ct.any?{|erc| erc.w?(con)} then
        join.set_w(con)
      elsif ct.any?{|erc| erc.e?(con)} then
        join.set_e(con) # constraints are e by default
      else
        join.set_l(con)
      end  
    end
    return join
  end

  def fusionally_close (ct)
    
    # Start with the ERCs from the original ct
    closure = ct.dup
    
    new_fusand = true
    while new_fusand
      new_fusand = false
      closure_dup = closure.dup

      closure.each do |erc_1|
        closure.each do |erc_2|
          ct_t = Comparative_tableau.new
          ct_t << erc_1 << erc_2
          fus = Fred.fuse(ct_t)
          if !closure_dup.include?(fus)
            closure_dup << fus
            new_fusand = true
          end
        end
      end
      closure = closure_dup
    end

    return closure
  end
  
end
