# Author: Nazarre Merchant
#


require 'set'
require_relative 'derivation.rb'
require_relative 'opvec_po_diagram.rb'

class OpvecPO

  attr_reader :lang
  attr_reader :opvecs
  attr_reader :pos

  # A language should be a hashtable storing a set of derivations, whose keys
  # are the inputs to the derivations, one for each input in the HS system.
  #
  # The partial ordering relation only keeps immediate dominance relations.
  # The transitive closure is omitted.
  def initialize(language)
    @lang = language
    @opvecs = Array.new
    @pos = Set.new
    create_pos
  end
  
  def create_pos
    @lang.each_value do |d|
      d.get_competition_list.each do |comp|
        # Each competition in a derivation has exactly one winner. For tied optima,
        # each winner appears in its own, identical (except winner-marked), candidate set 
        # so paths can correctly identify outputs and inputs at each step.
        winner = comp.winners[0]
        base = get_base(comp)
        ov_winner = get_opvec(winner, base)
        
        comp.each do |cand|
          ov_loser = get_opvec(cand, base)
          @pos << [ov_winner, ov_loser] unless (ov_winner == ov_loser) or omit(ov_loser)
        end
      end
    end
  end
  
  # TODO: modify to omit all harmonically bounded opvecs 
  def omit(ov)
    contains_pos = false
    contains_neg = false
    ov.constraints.each do |c|
      contains_pos = true if ov.get_val(c) > 0
      contains_neg = true if ov.get_val(c) < 0
    end
    
    return (contains_pos && !contains_neg)
  end
  
  def get_opvec(cand, base)
    ov = OpVec.new(cand, base, cand.operation, cand.constraint_list)
    @opvecs << ov unless @opvecs.include?(ov)
    return @opvecs[@opvecs.index(ov)]
  end

  def get_base(competition)
    base = nil
    competition.each {|cand| base = cand if cand.base?}
    # Arb. select first candidate for the base if none are marked (faith cand is default marked)
    base = competition[0] if base.nil?
    base.base=true
    return base
  end

  def to_s
    opv_po_s = ""
    @pos.each do |po|
      opv_po_s << "#{po[0]} \t>\t #{po[1]}"
      opv_po_s << "\n"
    end

    return opv_po_s
  end

  def create_file(label)
    ov_d = OpvecPODiagram.new(self, label)
        #epo_d = EpoDiagram.new(moat.epos[key], label)
    tempfile = File.join(File.dirname(__FILE__),'..','temp',label)
    ov_d.png(tempfile)
    
    tempfile2 = File.join(File.dirname(__FILE__),'..','OpvecPOs', "#{label}.gv")
    #File.open("#{File.dirname(__FILE__).to_s}/Naz_file_2", 'w') do |f|
    File.open(tempfile2, 'w') do |f|
      f.puts ov_d.get_gv
    end
  end
  
end # end class OpvecPO
