# Author: Naz Merchant
#

# For testing the HS factorial typology and various HS parts.


require_relative '../../lib/constraint_hs.rb'
require_relative '../../lib/candidate.rb'
require_relative '../../lib/constraint.rb'
require_relative '../../lib/operation.rb'
require_relative '../../lib/operation_faith.rb'
require_relative '../../lib/operation_epen.rb'
require_relative '../../lib/hierarchy.rb'
require_relative '../../lib/derivation.rb'
require_relative '../../lib/factorial_typology_hs_op.rb'
require_relative '../../lib/opvec_po.rb'

require_relative 'operation_bst.rb'
require_relative 'operation_bst_epen.rb'
require_relative 'operation_bst_faith.rb'
require_relative 'bst_parser.rb'
require_relative 'constraint_bst_hs.rb'

class BST_Test
  
  def initialize    
    @parser = Parser.new(["V", "w"], ["C", "k"], ["V", "V"], ["C", "C"])
        
    @cons1_On = Constraint_BST_HS.new("m.Ons", "", Regexp.new("\\.[Vw]"), @parser, false)
    @cons2_NC = Constraint_BST_HS.new("m.NoCod", "", Regexp.new("[Ck]\\."), @parser, false)
    @cons3_Cx = Constraint_BST_HS.new("m.*Cmplx", "", Regexp.new("CC+|C*kC+|C+kC*"), @parser, false)
    @cons4_Mx = Constraint_BST_HS.new("f.Max", "", Regexp.new("x"), @parser)
    @cons5_Dp = Constraint_BST_HS.new("f.Dep", "", Regexp.new("[kw]"), @parser)
    @constraints = Array.new
    @constraints << @cons1_On << @cons2_NC << @cons3_Cx << @cons4_Mx << @cons5_Dp
    

    @op1_Del = Operation_BST.new(           "Del",      ["V", "C"], ["x", "x"], @constraints, @parser)
    @op2_Epn = Epenthesis_Operation_BST.new("e.p", ["", ""],   ["w", "k"], @constraints, @parser)
    @op3_Fai = Faithful_Operation_BST.new(  "F",       [""],       [""],       @constraints, @parser)
    @bst_ops = Array.new
    @bst_ops << @op1_Del << @op2_Epn << @op3_Fai
  end
  
  def view_single_competition_and_path
    comp = Competition.new.concat(@op1_Del.create_candidates("V"))
    comp[0].opt=true
    puts "Comp 1: " + comp.to_s
  end
  
  def test_competition_and_path_equality
    comp1 = Competition.new.concat(@op1_Del.create_candidates("CVCCV"))
    comp2 = Competition.new.concat(@op1_Del.create_candidates("CVCCV"))    
    comp3 = Competition.new.concat(@op2_Epn.create_candidates("VC"))  
    comp4 = Competition.new.concat(@op2_Epn.create_candidates("VC")) 
    comp5 = Competition.new.concat(@op3_Fai.create_candidates("CVCCV"))
    comp6 = Competition.new.concat(@op3_Fai.create_candidates("CVCCV"))

    # Set first candidate (arbitrarily) as optimal
    comp1[0].opt=true
    comp2[0].opt=true
    comp3[0].opt=true
    comp4[0].opt=true
    comp5[0].opt=true
    comp6[0].opt=true
    
    path1 = Path.new
    path2 = Path.new
    path3 = Path.new
    
    path1 << comp1 << comp1 << comp3 << comp5
    path2 << comp1 << comp2 << comp3 << comp5
    path3 << comp1 << comp1 << comp3 << comp3

    puts "test_competition_and_path_equality \t#{comp1 != comp3 && comp1 != comp4 && comp1 != comp5 && comp1 != comp6 &&
                                                 comp1 == comp2 && comp3 == comp4 && comp5 == comp6 &&
                                                 path1 == path2 && path1 != path3}"
     
    #puts "Comp 1: " + comp1.to_s
    #puts "Comp 2: " + comp2.to_s
    #puts "Comp 3: " + comp3.to_s
    #puts "Comp 4: " + comp4.to_s
    #puts "Comp 5: " + comp5.to_s
    #puts "Comp 6: " + comp6.to_s
    
    #puts "Path 1: " + path1.to_s
     
    #puts "Equality of comp1 and comp2: \t\ttrue  == #{comp1 == comp2}"
    #puts "Equality of comp1 and comp3: \t\ttrue  == #{comp1 == comp3}"
    #puts "Equality of path1 and path2: \t\ttrue  == #{path1 == path2}"
    #puts "Equality of path1 and path3: \t\tfalse  == #{path1 == path3}" 
  end
  
  def test_derivation_equality
    hier1 = Hierarchy.new
    hier2 = Hierarchy.new
    hier3 = Hierarchy.new
    hier4 = Hierarchy.new
    hier5 = Hierarchy.new
    hier6 = Hierarchy.new
    hier1 << [@cons1_On] << [@cons2_NC] << [@cons3_Cx] << [@cons4_Mx] << [@cons5_Dp] 
    hier2 << [@cons1_On] << [@cons2_NC] << [@cons3_Cx] << [@cons4_Mx] << [@cons5_Dp] 
    hier3 << [@cons4_Mx] << [@cons5_Dp] << [@cons3_Cx] << [@cons1_On] << [@cons2_NC] 
    hier4 << [@cons4_Mx] << [@cons5_Dp] << [@cons3_Cx] << [@cons1_On] << [@cons2_NC] 
    hier5 << [@cons1_On] << [@cons2_NC] << [@cons3_Cx] << [@cons5_Dp] << [@cons4_Mx]
    hier6 << [@cons1_On] << [@cons2_NC] << [@cons3_Cx] << [@cons5_Dp] << [@cons4_Mx]
        
    deriv_1 = Derivation.new("CVC", hier1, @bst_ops)
    deriv_2 = Derivation.new("CVC", hier2, @bst_ops)
    deriv_3 = Derivation.new("CVC", hier3, @bst_ops)
    deriv_4 = Derivation.new("CVC", hier4, @bst_ops)
    deriv_5 = Derivation.new("CVCC", hier5, @bst_ops)
    deriv_6 = Derivation.new("CVCC", hier6, @bst_ops)
                
    puts "test_derivation_equality \t\t#{deriv_1 == deriv_2 && deriv_3 == deriv_4 && deriv_5 == deriv_6 && 
                                         deriv_1 != deriv_3 && deriv_1 != deriv_5 && deriv_3 != deriv_5}"
                                         
    #puts deriv_1
    #puts deriv_2
    #puts deriv_3
    #puts deriv_4
    #puts deriv_5
    #puts deriv_6
    
    #puts deriv_1.get_competition_list
    #puts deriv_2.get_candidate_sets
  end
  
  def test_factorial_typology
    ft_V = Factorial_Typology_HS_Op.new(["V"], @bst_ops, @constraints)
    ft_VC = Factorial_Typology_HS_Op.new(["V", "VC"], @bst_ops, @constraints)
    #ft_VCC = Factorial_Typology_HS_Op.new(["V", "VC", "VCC"], @bst_ops, @constraints)
    #ft_VCCC = Factorial_Typology_HS_Op.new(["V", "VC", "VCC", "VCCC"], @bst_ops, @constraints)
    #ft_VCCCC = Factorial_Typology_HS_Op.new(["VCCCC"], @bst_ops, @constraints)
    #ft_VCCCCC = Factorial_Typology_HS_Op.new(["VCCCCC"], @bst_ops, @constraints)    
    #ft_VCCCCCC = Factorial_Typology_HS_Op.new(["VCCCCCC"], @bst_ops, @constraints) 

    #ft_VCC_only = Factorial_Typology_HS_Op.new(["VCC"], @bst_ops, @constraints)

    puts "test_factorial_typology \t\t#{120 == ft_V.derivs.length && 120 == ft_VC.derivs.length && 
                                        3 == ft_V.langs.length && 10 == ft_VC.langs.length}"

    #puts "Lang count (V)\t\t #{ft_V.langs.length}"
    #puts "Lang count (VC)\t\t #{ft_VC.langs.length}"
    #puts "Lang count (VCC)\t #{ft_VCC.langs.length}"
    #puts "Lang count (VCCC)\t #{ft_VCCC.langs.length}"
    #puts "Lang count (VCCCC)\t #{ft_VCCCC.langs.length}"
    #puts "Lang count (VCCCCC)\t #{ft_VCCCCC.langs.length}"
    #puts "Lang count (VCCCCCC)\t #{ft_VCCCCCC.langs.length}"
    
    #puts "Lang count (VCC only)\t #{ft_VCC_only.langs.length}"
    
    # with 0-0 c's: 3 langs.
    # with 0-1 c's: 10 langs.
    # with 0-2 c's: 36 langs.
    # with 0-3 c's: 43 langs.    
    # with 0-4 c's: 47 langs.
    # with 0-5 c's: 47 langs. 
    # with 0-6 c's: 47 langs.
    # with 0-7 c's: 47 langs.
    # with 0-8 c's: 47 langs.

    # One derivation for each total order
    #puts "Derivation count with input 'V' \t6 == #{ft_V.derivs.length}"
    #puts "Derivation count with input 'VC' \t6 == #{ft_VC.derivs.length}" 
    #puts "Language count with input 'o':  \t2 == #{ft_o.langs.length}"
    #puts "Language count with input 'oo': \t3 == #{ft_oo.langs.length}"
    
    #puts ft_all.to_s
    #ft_VCC.langs.each {|d_ht| puts; d_ht.values.each {|d| puts d.get_comparative_tableau}}
    #ft_VC.get_languages_as_comparative_tableau.each {|ct| puts; puts ct}
    #ft_VCC.get_languages_as_comparative_tableau.each {|ct| puts; puts ct}
    
  end
  
  def test_opvec
    hier1 = Hierarchy.new
    hier2 = Hierarchy.new
    hier1 << [@cons1_On] << [@cons2_NC] << [@cons3_Cx] << [@cons4_Mx] << [@cons5_Dp] 
    hier2 << [@cons4_Mx] << [@cons5_Dp] << [@cons3_Cx] << [@cons1_On] << [@cons2_NC] 

    deriv_1 = Derivation.new("CVC", hier1, @bst_ops)
    deriv_2 = Derivation.new("CVC", hier2, @bst_ops)
    deriv_3 = Derivation.new("CVC", hier2, @bst_ops)
      
    lang1 = Hash.new; lang1[deriv_1.input] = deriv_1
    lang2 = Hash.new; lang2[deriv_2.input] = deriv_2
    lang3 = Hash.new; lang3[deriv_3.input] = deriv_3
    
    opvecs1 = deriv_1.get_opvecs
    opvecs2 = deriv_2.get_opvecs
    opvecs3 = deriv_3.get_opvecs
    ov_po1 = OpvecPO.new(lang1)    
    ov_po2 = OpvecPO.new(lang2)
    ov_po3 = OpvecPO.new(lang3)
    
    puts "Opvecs: for input CVC and hier #{hier1.to_s}"
    opvecs1.each {|ov| puts "#{ov.to_s}"}
    puts; puts deriv_1; puts; puts ov_po1

    ov_po1.create_file("OV_1")
    ov_po2.create_file("OV_2")
    ov_po3.create_file("OV_3")
    
    ft_VC = Factorial_Typology_HS_Op.new(["CVC"], @bst_ops, @constraints)
    count = 1
    ft_VC.langs.each do |l|
      ov_po = OpvecPO.new(l)
      ov_po.create_file("CVC#{count}")
      count += 1
    end
      
    puts
    puts ft_VC
    #puts "Opvecs: for input CVC and hier #{hier2.to_s}"
    #opvecs2.each {|ov| puts "#{ov.to_s}"}
    #puts; puts deriv_2; puts; puts ov_po2
  end
end

bst = BST_Test.new
#bst.view_single_competition_and_path
bst.test_competition_and_path_equality
bst.test_derivation_equality
bst.test_factorial_typology
bst.test_opvec


