# Author: Naz Merchant
#

# A HS operation consisting of a source string and a target string
#
# Operations take an input string and produce an array of candidates,
# one for each location in the input string that the operation can apply to.
#
# Bst operations differ from the usual HS ops as they must create parses for each
# outputs string. 

require_relative '../../lib/operation.rb'

require_relative 'bst_parser.rb'
require_relative 'candidate_bst.rb'

class Operation_BST < Operation
  
  # Format for operations:
  #
  # => For type A, source must be an array strings & target must be 
  # => an array of strings. Source and target must have the same number of elements.
  #
  #    type source     target
  #    A    abc        xyz         <-- each instance of 'a' is replaced with 'x', 'b' w/ 'y', etc
  #    B    [ab]_[cd]  x           <-- 'x' is inserted between a&c, a&d, b&c, and b&d
  #    C    abc        ""          <-- each a,b,&c is deleted
  
  # Type A&C are identical in structure.
  
  # source & target must be arrays of string
  
  def initialize(name, source, target, constraints, parser)
    super(name, source, target, constraints)
    @parser = parser
  end
  
  def create_candidates(input)
    cands = Array.new
    super(input).each do |c|
      @parser.parse(c.output).each {|o| cands << BST_Candidate.new(input, o, false, @cons, self, @parser)}
    end

    set_violations(cands)
    return cands
  end
  
end