# Author: Bruce Tesar
# 

require_relative '../lib/constraint'
require_relative '../lib/erc'

# This module provides a method for simple construction of example
# ERCs, for use in testing. It is designed as a mix-in, to be included
# in Test classes. The constants L, E, and W allow the definition of an ERC
# in a compact and easy-to-read fashion.
#   erc = quick_erc([W, L, E], "erc_label")
# The returned object is of class Erc, and contains references to constructed
# constraints. The constraints of the ERC are autonamed, left-to-right:
# "C1", "C2", ...
module QuickErc

  # Constants for evaluation by constraints
  L = "L"  # prefers the loser
  E = "e"  # no preference
  W = "W"  # prefers the winner

  # Constants for the first few constraints
  C1 = Constraint.new("C1", 1)
  C2 = Constraint.new("C2", 1)
  C3 = Constraint.new("C3", 1)
  C4 = Constraint.new("C4", 1)
  C5 = Constraint.new("C5", 1)

  # Takes a list of evaluation values (one per implicit constraint) and an
  # optional label, and constructs an Erc with constraints auto-labeled
  # C1, C2, ..., with the evaluation values assigned in the order that
  # they are listed.
  def quick_erc(evals, label="")
    constraints = Array.new
    erc = Erc.new(constraints, label)
    id = 0
    evals.each do |val|
      id += 1
      con = Constraint.new("C#{id.to_s}", id)
      constraints << con
      if val==W then erc.set_w(con)
      elsif val==E then erc.set_e(con)
      elsif val==L then erc.set_l(con)
      else raise "quick_erc: invalid evaluation value #{val}. Use the constants defined in module QuickErc."
      end
    end
    return erc
  end

end # module QuickErc