# 
# To change this template, choose Tools | Templates
# and open the template in the editor.
 

#$:.unshift File.join(File.dirname(__FILE__),'..','lib')

require 'test/unit'
require_relative '../lib/candidate'

class Test_candidate < Test::Unit::TestCase
  # Set up test candidates
  def setup
    @new_cand = Candidate.new("in1", "out1", true, ["C1", "C2"])
    @new_cand.set_viols("C1", 0)
    @new_cand.set_viols("C2", 3)

    @new_equiv = Candidate.new("in_equiv", "out_equiv", true, ["C1", "C2"])
    @new_equiv.set_viols("C1", 0)
    @new_equiv.set_viols("C2", 3)

    @nonopt_cand = Candidate.new("in2", "out2", false, ["C1", "C2"])
    @nonopt_cand.set_viols("C1", 5)
    @nonopt_cand.set_viols("C2", 2)

    @undef_cand = Candidate.new("in", "out", false, ["C3", "C4"])
    @undef_cand.set_viols("C4", 1)
  end

  def teardown
    # teardown code here
  end

  def test_to_s
    assert_equal("in1 --> out1 opt:Y  C1:0 C2:3", @new_cand.to_s)
    assert_equal("in2 --> out2 opt:N  C1:5 C2:2", @nonopt_cand.to_s)
    # labels and remarks
    @new_cand.label = "1a"
    @nonopt_cand.label = "1b"
    @nonopt_cand.remark = "Can\'t win"
    assert_equal("1a: in1 --> out1 opt:Y  C1:0 C2:3", @new_cand.to_s)
    assert_equal("1b: in2 --> out2 opt:N  C1:5 C2:2  Can\'t win", @nonopt_cand.to_s)
    # test to_s with undefined violation counts
    assert_equal("in --> out opt:N  C3:? C4:1", @undef_cand.to_s)
    @undef_cand.set_viols("C3", 9)
    assert_equal("in --> out opt:N  C3:9 C4:1", @undef_cand.to_s)
  end

  def test_to_a
    expected_new_cand = ["","in1","out1",1,0,3,nil,nil]
    assert_equal(expected_new_cand, @new_cand.to_a)
  end

  def test_readables
    assert_equal("in1", @new_cand.input)
    assert_equal("out1", @new_cand.output)
  end

  def test_optimal
    assert(@new_cand.opt?(), "@new_cand should be optimal.")
    assert(!@nonopt_cand.opt?(), "@nonopt_cand should NOT be optimal.")
  end

  def test_constraint_access
    conlist0 = ["*PA", "*GA"]
    cand0 = Candidate.new("i", "o", false, conlist0)
    con_list = cand0.constraint_list
    assert_equal(conlist0, con_list)
    assert_same(conlist0, con_list) # should be the same object
  end

  def test_constraint_violations
    # Test constraint violation setting and getting.
    assert_equal(0, @new_cand.get_viols("C1"))
    assert_equal(3, @new_cand.get_viols("C2"))
    @new_cand.set_viols("C1", 2)
    assert_equal(2, @new_cand.get_viols("C1"))
    @new_cand.set_viols("C2", 1)
    assert_equal(1, @new_cand.get_viols("C2"))
  end
  
  def test_ident_viols
    assert(@new_cand.ident_viols?(@new_cand))
    assert(@nonopt_cand.ident_viols?(@nonopt_cand))
    assert(!@new_cand.ident_viols?(@nonopt_cand))
    assert(!@nonopt_cand.ident_viols?(@new_cand))
    #
    assert(!@new_cand.equal?(@new_equiv))
    assert(@new_cand.ident_viols?(@new_equiv))
    assert(@new_equiv.ident_viols?(@new_cand))
  end
  
  def test_harmonically_bounds
    cand1 = Candidate.new("in1", "out1", true, ["C1", "C2", "C3"])
    cand1.set_viols("C1", 0); cand1.set_viols("C2", 3); cand1.set_viols("C3", 1)
    assert(!cand1.harmonically_bounds?(cand1))
    #
    other = Candidate.new("in1", "out1", true, ["C1", "C2", "C3"])
    other.set_viols("C1", 1); other.set_viols("C2", 3); other.set_viols("C3", 1)
    assert(!other.harmonically_bounds?(other))
    assert(cand1.harmonically_bounds?(other))
    assert(!other.harmonically_bounds?(cand1))
    #
    other = Candidate.new("in1", "out1", true, ["C1", "C2", "C3"])
    other.set_viols("C1", 1); other.set_viols("C2", 1); other.set_viols("C3", 1)
    assert(!other.harmonically_bounds?(other))
    assert(!cand1.harmonically_bounds?(other))
    assert(!other.harmonically_bounds?(cand1))
    #
    other = Candidate.new("in1", "out1", true, ["C1", "C2", "C3"])
    other.set_viols("C1", 1); other.set_viols("C2", 4); other.set_viols("C3", 7)
    assert(!other.harmonically_bounds?(other))
    assert(cand1.harmonically_bounds?(other))
    assert(!other.harmonically_bounds?(cand1))
  end
  
  def test_opt_set
    assert(@new_cand.opt?)
    @new_cand.opt = false
    assert(!@new_cand.opt?)
  end
  
  def test_equality
    assert_not_equal(@new_cand,@new_equiv)
    diff_label = @new_cand.dup
    diff_label.label = "diff_label"
    assert_equal(@new_cand, diff_label)
    assert(@new_cand==diff_label)
    assert(@new_cand.eql?(diff_label))
    diff_input = Candidate.new("diff_input", @new_cand.output, @new_cand.opt?, @new_cand.constraint_list)
    assert_not_equal(@new_cand,diff_input)
    assert(@new_cand!=diff_input)
    assert(!@new_cand.eql?(diff_input))
  end
  
  def test_merge_flag
    assert(!@new_cand.merged?)
    assert_nothing_raised(RuntimeError){@new_cand.add_merge_candidate(@new_equiv)}
    assert(@new_cand.merged?)
  end
  
  def test_merged_candidates
    assert_equal("in1 --> out1 opt:Y  C1:0 C2:3", @new_cand.to_s)
    assert_equal("out1", @new_cand.merged_outputs_to_s)
    assert_nothing_raised(RuntimeError){@new_cand.add_merge_candidate(@new_equiv)}
    assert_equal("in1 --> out1 opt:Y  C1:0 C2:3  MERGED\n --> out_equiv", @new_cand.to_s)
    assert_equal("out1 MER out_equiv", @new_cand.merged_outputs_to_s)
    assert_raise(RuntimeError){@new_cand.add_merge_candidate(@nonopt_cand)}
    assert_equal("in1 --> out1 opt:Y  C1:0 C2:3  MERGED\n --> out_equiv", @new_cand.to_s)
    assert_equal("out1 MER out_equiv", @new_cand.merged_outputs_to_s)
  end
  
  def test_get_merged_candidates
    @new_cand.add_merge_candidate(@new_equiv)
    assert(@new_cand.merged?)
    mc_list = @new_cand.get_merge_candidates
    assert_equal(1,mc_list.size)
    assert("in2 --> out_equiv opt:N  C1:0 C2:3", mc_list[0].to_s)
  end
  
  def test_dup
    @new_cand.label = '3.1'
    assert_equal("3.1: in1 --> out1 opt:Y  C1:0 C2:3", @new_cand.to_s)
    copy = @new_cand.dup
    assert_equal("3.1: in1 --> out1 opt:Y  C1:0 C2:3", copy.to_s)
    assert_equal(@new_cand.label, copy.label)
  end

  def test_freeze
    @new_cand.freeze
    assert(@new_cand.frozen?)
    dup_cand = @new_cand.dup
    assert(!dup_cand.frozen?)
    # Exception is raised when an attempt is made to modify a frozen object.
    if RUBY_VERSION =~ /^1\.9/ then
      # Ruby 1.9 raises a RuntimeError.
      assert_raise(RuntimeError){@new_cand.opt=false}
      assert_nothing_raised(RuntimeError){@new_cand.get_viols("C1")}
      assert_raise(RuntimeError){@new_cand.set_viols("C1",7)}
    else
      # Ruby 1.8 raises a TypeError.
      assert_raise(TypeError){@new_cand.opt=false}
      assert_nothing_raised(TypeError){@new_cand.get_viols("C1")}
      assert_raise(TypeError){@new_cand.set_viols("C1",7)}
    end
  end

end # class Test_candidate
