# Author: Bruce Tesar
 

require 'test/unit'
require_relative '../lib/comparative_tableau'
require_relative '../lib/constraint'
require_relative '../lib/competition_list'

class Test_comparative_tableau < Test::Unit::TestCase
  def setup
    @win1 = Candidate.new("in1", "out1", true, ["C1", "C2", "C3"])
    @win1.label = "1.1"
    @win1.set_viols("C1", 0)
    @win1.set_viols("C2", 3)
    @win1.set_viols("C3", 2)

    @win1_alt = Candidate.new("in1", "out6", true, ["C1", "C2", "C3"])
    @win1_alt.label = "1f"
    @win1_alt.set_viols("C1", 0)
    @win1_alt.set_viols("C2", 2)
    @win1_alt.set_viols("C3", 3)

    @los1_2 = Candidate.new("in1", "out2", false, ["C1", "C2", "C3"])
    @los1_2.label = "1.2"
    @los1_2.set_viols("C1", 1)
    @los1_2.set_viols("C2", 1)
    @los1_2.set_viols("C3", 2)

    @los1_3 = Candidate.new("in1", "out3", false, ["C1", "C2", "C3"])
    @los1_3.label = "1.3"
    @los1_3.set_viols("C1", 2)
    @los1_3.set_viols("C2", 1)
    @los1_3.set_viols("C3", 1)

    @comp1 = Competition.new
    @comp1.push(@win1)
    @comp1.push(@los1_2)

    # Competition 2
    @win2 = Candidate.new("in2", "out1", true, ["C1", "C2", "C3"])
    @win2.label = "2.1"
    @win2.set_viols("C1", 0)
    @win2.set_viols("C2", 3)
    @win2.set_viols("C3", 2)

    @los2_2 = Candidate.new("in2", "out2", false, ["C1", "C2", "C3"])
    @los2_2.label = "2.2"
    @los2_2.set_viols("C1", 1)
    @los2_2.set_viols("C2", 1)
    @los2_2.set_viols("C3", 2)

    @los2_3 = Candidate.new("in2", "out3", false, ["C1", "C2", "C3"])
    @los2_3.label = "2.3"
    @los2_3.set_viols("C1", 2)
    @los2_3.set_viols("C2", 4)
    @los2_3.set_viols("C3", 1)

    @comp2 = Competition.new
    @comp2.push(@win2)
    @comp2.push(@los2_2)
    @comp2.push(@los2_3)
    
    # Competition 3 (viols match Competition 2)
    @win3 = Candidate.new("in3", "out1", true, ["C1", "C2", "C3"])
    @win3.label = "3.1"
    @win3.set_viols("C1", 0)
    @win3.set_viols("C2", 3)
    @win3.set_viols("C3", 2)

    @los3_2 = Candidate.new("in3", "out2", false, ["C1", "C2", "C3"])
    @los3_2.label = "3.2"
    @los3_2.set_viols("C1", 1)
    @los3_2.set_viols("C2", 1)
    @los3_2.set_viols("C3", 2)

    @comp3 = Competition.new
    @comp3.push(@win3)
    @comp3.push(@los3_2)

    # competition list
    @clist = Competition_list.new
    @clist.label = "WS7"
    @clist << @comp1 << @comp2
  end

  def teardown
    # teardown code here
  end

  def test_add_competition
    ct = Comparative_tableau.new()
    ct.add_competition(@comp1)
    expected1 = "1.1>2 in1 out1 out2 C1:W C2:L C3:e"
    assert_equal(expected1, ct.to_s)
    #
    @comp1.push(@los1_3)
    ct = Comparative_tableau.new()
    ct.add_competition(@comp1)
    expected2 = expected1 + "\n" + "1.1>3 in1 out1 out3 C1:W C2:L C3:L"
    assert_equal(expected2, ct.to_s)
    #
    ct.add_competition(@comp2)
    expected3 = expected2 + "\n" + "2.1>2 in2 out1 out2 C1:W C2:L C3:e" +
      "\n" + "2.1>3 in2 out1 out3 C1:W C2:W C3:L"
    assert_equal(expected3, ct.to_s)
  end

  def test_add_competition_list
    ct = Comparative_tableau.new(@clist.label)
    ct.add_competition_list(@clist)
    assert_equal("WS7", ct.label)
  end
  
  def test_integrity_checks
    ct = Comparative_tableau.new
    assert_nothing_raised(RuntimeError){ct.add_competition(@comp1)}
    # add a second winner to the competition, to trigger an exception.
    ct = Comparative_tableau.new
    ct.add_competition(@comp2)
    @comp1.push(@win1_alt)
    assert_raise(RuntimeError){ct.add_competition(@comp1)}
    # An empty competition has no winner
    assert_raise(RuntimeError){ct.add_competition(Competition.new)}
  end
  
  def test_merged_winners
    ct = Comparative_tableau.new
    cand_m = Candidate.new(@win1.input, "out_m", @win1.opt?, @win1.constraint_list)
    @win1.constraint_list.each{|con| cand_m.set_viols(con, @win1.get_viols(con))}
    cand_m.label = "1.m"; cand_m.remark = ""
    @comp1.push(cand_m)
    assert_nothing_raised(RuntimeError){ct.add_competition(@comp1)}
    assert(ct.winners[0].merged?)
    assert_equal("MERGED", ct.winners[0].remark)
    expected1 = "1.1>2 in1 out1 MER out_m out2 C1:W C2:L C3:e"
    assert_equal(expected1, ct.to_s)
  end
  
  def test_constraints
    ct = Comparative_tableau.new
    ct.add_competition_list(@clist)
    assert_equal(["C1", "C2", "C3"], ct.constraint_list)
    ct = Comparative_tableau.new
    assert_equal([], ct.constraint_list)
  end
  
  def test_partition
    ct = Comparative_tableau.new
    ct.add_competition_list(@clist)
    assert_equal(3, ct.size)
    ct_true, ct_false = ct.partition{|erc| erc.label == "2.1>2"}
    assert_instance_of(Comparative_tableau, ct_true)
    assert_instance_of(Comparative_tableau, ct_false)
    assert_equal(1, ct_true.size)
    assert_equal(2, ct_false.size)
    assert_equal(ct[0], ct_false[0])
    assert_equal(ct[1], ct_true[0])
    assert_equal(ct[2], ct_false[1])
  end
  
  def test_reject
    ct = Comparative_tableau.new
    ct.add_competition_list(@clist)
    assert_equal(3, ct.size)
    assert_equal(2, ct.winners.size)
    ctr = ct.reject{|erc| erc.label == "2.1>2"}
    ctr.label = "After reject"
    assert_equal(2, ctr.size)
    assert_equal(2, ctr.winners.size)
    ctr2 = ctr.reject{|erc| erc.label == "2.1>3"}
    ctr2.label = "After second reject"
    assert_equal(1, ctr2.size)
    assert_equal(1, ctr2.winners.size)
  end
  
  def test_constructor_with_conlist
    con_list = [Constraint.new('C1',1),Constraint.new('C2',2)]
    ct = Comparative_tableau.new("Empty", con_list)
    assert_equal(['1:C1','2:C2'], ct.constraint_list.map{|c| c.to_s})
  end
  
  def test_winner_list
    ct = Comparative_tableau.new(@clist.label)
    ct.add_competition_list(@clist)
    assert_equal(2,ct.winners.size)
    assert_equal([@win1,@win2],ct.winners)
    #
    @comp1.push(@los1_3)
    ct.add_competition(@comp1)
    assert_equal(2,ct.winners.size)
    assert_equal([@win1,@win2],ct.winners)
    #
    ct.add_competition(@comp3)
    assert_equal(3,ct.winners.size)
    assert_equal([@win1,@win2,@win3],ct.winners)
  end
  
end # class Test_comparative_tableau
