# 
# Test the class Competition.
# The class is in source file competition.rb.
 

require 'test/unit'
require_relative '../lib/competition'
require_relative '../lib/competition_list'
require_relative '../lib/candidate'

# Test for class Competition
class Test_competition < Test::Unit::TestCase
  def setup
    @new_cand = Candidate.new("in1", "out1", true, ["C1", "C2"])
    @new_cand.set_viols("C1", 0)
    @new_cand.set_viols("C2", 3)

    @winner2 = Candidate.new("in4", "out4", true, ["C1", "C2"])
    @winner2.set_viols("C1", 1)
    @winner2.set_viols("C2", 4)

    @nonopt_cand = Candidate.new("in2", "out2", false, ["C1", "C2"])
    @nonopt_cand.set_viols("C1", 5)
    @nonopt_cand.set_viols("C2", 2)

    @undef_cand = Candidate.new("in", "out", false, ["C1", "C2"])
    @undef_cand.set_viols("C2", 1)
  end

  def test_add_candidates
    comp = Competition.new
    assert(!comp.optima?, "There should NOT be an optimum.")
    assert(!comp.mult_optima?, "There should NOT be multiple optima.")
    comp.push(@new_cand)
    assert(comp.optima?, "There should be an optimum.")
    assert(!comp.mult_optima?, "There should only be one optimum.")
    comp.push(@nonopt_cand)
    assert(comp.optima?, "There should be an optimum.")
    assert(!comp.mult_optima?, "There should only be one optimum.")
    comp.push(@winner2)
    assert(comp.optima?, "There should be an optimum.")
    assert(comp.mult_optima?, "There should be multiple optima.")
  end

  def test_get_winners_losers
    comp = Competition.new
    comp.push(@new_cand)
    assert_equal([@new_cand], comp.winners)
    assert_equal([], comp.losers)
    comp.push(@nonopt_cand)
    assert_equal([@new_cand], comp.winners)
    assert_equal([@nonopt_cand], comp.losers)    
    comp.push(@winner2)
    assert_equal([@new_cand, @winner2], comp.winners)
    assert_equal([@nonopt_cand], comp.losers)    
  end

  def test_to_s
    comp = Competition.new
    comp.push(@new_cand)
    comp.push(@nonopt_cand)
    comp.push(@winner2)
    expected = "in1 --> out1 opt:Y  C1:0 C2:3\nin2 --> out2 opt:N  C1:5 C2:2\nin4 --> out4 opt:Y  C1:1 C2:4\n"
    assert_equal(expected, comp.to_s)
  end

  def test_constraint_list
    comp = Competition.new
    assert_equal([], comp.constraint_list)
    assert_equal(["C1","C2"], @winner2.constraint_list)
  end
  
  def test_merge_identical_candidates
    comp = Competition.new
    cand1 = Candidate.new("input", "out1", false, ["C1", "C2"])
    cand1.set_viols("C1", 0)
    cand1.set_viols("C2", 3)
    comp.push(cand1)
    cand2 = Candidate.new("input", "out2", false, ["C1", "C2"])
    cand2.set_viols("C1", 1)
    cand2.set_viols("C2", 2)
    comp.push(cand2)
    cand3 = Candidate.new("input", "out3", false, ["C1", "C2"])
    cand3.set_viols("C1", 0)
    cand3.set_viols("C2", 3)
    comp.push(cand3)
    #
    assert_equal(3,comp.size)
    comp.merge_identical_candidates
    assert_equal(2,comp.size)
    out_string = "input --> out1 opt:N  C1:0 C2:3  MERGED\n --> out3\ninput --> out2 opt:N  C1:1 C2:2\n"
    assert_equal(out_string, comp.to_s)
  end
  
  def test_merge_identical_candidates_2
    comp = Competition.new
    cand1 = Candidate.new("input", "out1", false, ["C1", "C2"])
    cand1.set_viols("C1", 0)
    cand1.set_viols("C2", 3)
    comp.push(cand1)
    cand2 = Candidate.new("input", "out2", false, ["C1", "C2"])
    cand2.set_viols("C1", 1)
    cand2.set_viols("C2", 2)
    comp.push(cand2)
    cand3 = Candidate.new("input", "out3", false, ["C1", "C2"])
    cand3.set_viols("C1", 0)
    cand3.set_viols("C2", 3)
    comp.push(cand3)
    #
    cand4 = Candidate.new("input", "out4", false, ["C1", "C2"])
    cand4.set_viols("C1", 2)
    cand4.set_viols("C2", 1)
    comp.push(cand4)
    cand5 = Candidate.new("input", "out5", false, ["C1", "C2"])
    cand5.set_viols("C1", 2)
    cand5.set_viols("C2", 1)
    comp.push(cand5)
    cand6 = Candidate.new("input", "out6", false, ["C1", "C2"])
    cand6.set_viols("C1", 0)
    cand6.set_viols("C2", 3)
    comp.push(cand6)
    assert_equal(6,comp.size)
    comp.merge_identical_candidates
    assert_equal(3,comp.size)
    out_string = "input --> out1 opt:N  C1:0 C2:3  MERGED\n --> out3\n --> out6\n"
    out_string += "input --> out2 opt:N  C1:1 C2:2\n"
    out_string += "input --> out4 opt:N  C1:2 C2:1  MERGED\n --> out5\n"
    assert_equal(out_string, comp.to_s)
  end
  
end # class Test_competition
