# Author: Bruce Tesar
 
require 'test/unit'
require_relative '../lib/fred'
require_relative '../lib/constraint'
require_relative '../lib/comparative_tableau'
require_relative 'quick_erc'

class Test_fred < Test::Unit::TestCase

  # Bring in the quick_erc method and related constants.
  include QuickErc

  def setup
    @con1 = Constraint.new("C1", 1)
    @con2 = Constraint.new("C2", 2)
    @con3 = Constraint.new("C3", 3)
    @con4 = Constraint.new("C4", 4)
    @con5 = Constraint.new("C5", 5)
    @con6 = Constraint.new("C6", 6)
    #
    @erc1 = quick_erc([W,L,E],"erc1")
    @erc2 = quick_erc([E,W,E],"erc2")
    @erc3 = quick_erc([L,W,W],"erc3")
    @erc4 = quick_erc([L,E,L],"erc4")
    #
    @fuse1_2 = quick_erc([W,L,E],"fuse1_2")
    @fuse1_2_3 = quick_erc([L,L,W],"fuse1_2_3")
    #
    @ct1_2 = Comparative_tableau.new("CT1_2") << @erc1 << @erc2
    @ct1_2_3 = Comparative_tableau.new("CT1_2_3") << @erc1 << @erc2 << @erc3
    @ct3_2_1 = Comparative_tableau.new("CT3_2_1") << @erc3 << @erc2 << @erc1
  end

  def teardown
    # teardown code here
  end

  def test_fuse
    assert(@fuse1_2.eql?(Fred.fuse(@ct1_2)), 'The fusion should be equivalent.')
    assert(@fuse1_2_3.eql?(Fred.fuse(@ct1_2_3)), 'The fusion should be equivalent.')
    assert(Fred.fuse(@ct1_2_3).eql?(Fred.fuse(@ct3_2_1)), 'The fusion should be equivalent.')
  end
  
  def test_arrow
    assert(Fred.arrow?(@erc1,@erc1))
    assert(Fred.arrow?(quick_erc([W,L,E]),quick_erc([W,L,E])))
    assert(!Fred.arrow?(@erc1,@erc2))
    assert(!Fred.arrow?(quick_erc([W,L,E]),quick_erc([E,W,E])))
    assert(!Fred.arrow?(quick_erc([E,W,E]),quick_erc([W,L,E])))
    assert(Fred.arrow?(@fuse1_2_3,@erc3))
    assert(!Fred.arrow?(@erc3,@fuse1_2_3))
    assert(Fred.arrow?(quick_erc([W,L,E]),quick_erc([W,W,E])))
    assert(Fred.arrow?(quick_erc([W,L,E,L,L,W]),quick_erc([W,W,E,L,E,W])))
    assert(!Fred.arrow?(quick_erc([W,L,E,L,L,W]),quick_erc([W,W,E,L,E,E])))
    assert(Fred.arrow?(quick_erc([W,L,E,L,L,W]),quick_erc([W,L,E,L,E,W])))
    assert(!Fred.arrow?(quick_erc([W,L,E,L,L,W]),quick_erc([W,L,L,L,E,W])))
  end
  
  def test_find_skb_l
    assert_equal([@con2].to_set, Fred.find_skb_l(quick_erc([W,L,E]),quick_erc([W,L,E])))
    assert_equal([@con2,@con3].to_set, Fred.find_skb_l(quick_erc([W,L,L]),quick_erc([W,L,L])))
    assert_equal([].to_set, Fred.find_skb_l(quick_erc([W,L,E]),quick_erc([W,E,L])))
    assert_equal([@con2].to_set, Fred.find_skb_l(quick_erc([L,L,L]),quick_erc([W,L,E])))
  end
  
  def test_triv_valid
    assert(!quick_erc([W,L,E]).triv_valid?)
    assert(quick_erc([W,E,W]).triv_valid?)
    assert(quick_erc([E,E,E]).triv_valid?)
    assert(quick_erc([E,E,W]).triv_valid?)
  end
  
  def test_triv_invalid
    assert(!quick_erc([W,L,E]).triv_invalid?)
    assert(!quick_erc([W,W,E]).triv_invalid?)
    assert(quick_erc([L,E,L]).triv_invalid?)
    assert(quick_erc([L,L,L]).triv_invalid?)
    assert(!quick_erc([E,E,E]).triv_invalid?)
  end
  
  def test_run_fred
    ct = Comparative_tableau.new("CT")
    ct << quick_erc([W,L,E,W,W,E],"erc1")
    ct << quick_erc([E,W,E,W,E,E],"erc2")
    ct << quick_erc([L,W,W,W,W,E],"erc3")
    ct << quick_erc([W,L,W,W,E,E],"erc4")
    fred_val = Fred.new(ct)
    assert(fred_val.fred_steps_list.size == 4)
    assert(fred_val.fred_steps_list[0].keep?)
    assert(quick_erc([L,L,W,W,W,E]).eql?(fred_val.fred_steps_list[0].fusion))
    assert("CT", fred_val.label)
    assert("MIB", fred_val.mib.label)
    assert("CT", fred_val.mib.to_s)
    fred_steps_expected = <<FRED_STEPS_EXPECTED
Fred step CT KEEP? true
FUS: f.CT 1:C1:L 2:C2:L 3:C3:W 4:C4:W 5:C5:W 6:C6:e
SKB: CT 1:C1:L 2:C2:e 3:C3:W 4:C4:W 5:C5:W 6:C6:e
SKB Support:
erc3 1:C1:L 2:C2:W 3:C3:W 4:C4:W 5:C5:W 6:C6:e
Fred step CT.3 KEEP? true
FUS: f.CT.3 1:C1:W 2:C2:L 3:C3:e 4:C4:W 5:C5:W 6:C6:e
SKB: CT.3 1:C1:W 2:C2:L 3:C3:e 4:C4:W 5:C5:W 6:C6:e
SKB Support:
erc1 1:C1:W 2:C2:L 3:C3:e 4:C4:W 5:C5:W 6:C6:e
Fred step CT.3.1 KEEP? false
FUS: f.CT.3.1 1:C1:e 2:C2:W 3:C3:e 4:C4:W 5:C5:e 6:C6:e
Fred step CT.5 KEEP? true
FUS: f.CT.5 1:C1:W 2:C2:L 3:C3:W 4:C4:W 5:C5:e 6:C6:e
SKB: CT.5 1:C1:W 2:C2:L 3:C3:W 4:C4:W 5:C5:e 6:C6:e
SKB Support:
erc4 1:C1:W 2:C2:L 3:C3:W 4:C4:W 5:C5:e 6:C6:e
FRED_STEPS_EXPECTED
    assert_equal(fred_steps_expected, fred_val.fred_steps_list.join)
  end
  
  def test_run_fred_single
    ct = Comparative_tableau.new("CTsing")
    ct << quick_erc([W,L,E,W,W,E],"erc1")
    fred_val = Fred.new(ct)
    assert(fred_val.fred_steps_list.size == 1)
    assert(fred_val.fred_steps_list[0].keep?)
    assert(quick_erc([W,L,E,W,W,E]).eql?(fred_val.fred_steps_list[0].fusion))
    assert("CTsing", fred_val.label)
    assert("MIB", fred_val.mib.label)
    assert("CTsing", fred_val.mib.to_s)
    fred_steps_expected = <<FRED_STEPS_EXPECTED
Fred step CTsing KEEP? true
FUS: f.CTsing 1:C1:W 2:C2:L 3:C3:e 4:C4:W 5:C5:W 6:C6:e
SKB: CTsing 1:C1:W 2:C2:L 3:C3:e 4:C4:W 5:C5:W 6:C6:e
SKB Support:
erc1 1:C1:W 2:C2:L 3:C3:e 4:C4:W 5:C5:W 6:C6:e
FRED_STEPS_EXPECTED
    assert_equal(fred_steps_expected, fred_val.fred_steps_list.join)
  end

  def test_run_fred_empty
    ct = Comparative_tableau.new("CTsing")
    fred_val = Fred.new(ct)
    assert(fred_val.fred_steps_list.empty?)
    assert(fred_val.mib.empty?)
  end
  
  # Brasovaneau & Prince pp. 28-29
  def test_bp
    ct = Comparative_tableau.new("BP")
    ct << quick_erc([W,L,E,W],"erc1")
    ct << quick_erc([E,W,L,W],"erc2")
    ct << quick_erc([W,W,L,E],"erc3")
    ct << quick_erc([W,L,W,E],"erc4")
    fred_val = Fred.new(ct)
    assert_equal("BP", fred_val.label)
    assert_equal("MIB.BP", fred_val.mib.label)

    # Check the computed mib by comparing its string representation.
    mib_output_expected =  <<MIB_OUTPUT
f.BP.4 1:C1:W 2:C2:L 3:C3:L 4:C4:e
f.BP.1 1:C1:e 2:C2:W 3:C3:L 4:C4:W
MIB_OUTPUT
    assert_equal(mib_output_expected, fred_val.mib.to_s+"\n")

    # Check each of the FRed steps.
    fred_steps_expected = <<FRED_STEPS_OUTPUT
Fred step BP KEEP? false
FUS: f.BP 1:C1:W 2:C2:L 3:C3:L 4:C4:W
Fred step BP.1 KEEP? true
FUS: f.BP.1 1:C1:e 2:C2:W 3:C3:L 4:C4:W
SKB: BP.1 1:C1:e 2:C2:W 3:C3:L 4:C4:W
SKB Support:
erc2 1:C1:e 2:C2:W 3:C3:L 4:C4:W
Fred step BP.4 KEEP? true
FUS: f.BP.4 1:C1:W 2:C2:L 3:C3:L 4:C4:e
SKB: BP.4 1:C1:W 2:C2:L 3:C3:L 4:C4:e
SKB Support:
erc3 1:C1:W 2:C2:W 3:C3:L 4:C4:e
erc4 1:C1:W 2:C2:L 3:C3:W 4:C4:e
FRED_STEPS_OUTPUT
    assert_equal(fred_steps_expected, fred_val.fred_steps_list.join)
  end
  
end # class Test_fred
