# Author: Naz Merchant
 
require 'test/unit'
require_relative '../lib/mojo'
require_relative '../lib/fred'
require_relative '../lib/constraint'
require_relative '../lib/comparative_tableau'
require_relative 'quick_erc'

class Test_fred < Test::Unit::TestCase

  # Bring in the quick_erc method and related constants.
  include QuickErc

  def setup
    @erc11 = quick_erc([W,L,E],"erc11")
    @erc12 = quick_erc([E,W,L],"erc12")
    @erc21 = quick_erc([W,E,L],"erc21")
    @erc22 = quick_erc([E,L,W],"erc22")
    #
    @join11_12 = quick_erc([W,W,E],"join11_12")
    @join12_21 = quick_erc([W,W,L],"join12_21")
    @join21_22 = quick_erc([W,E,W],"join21_22")
    #
    @erc4 = quick_erc([W,L,L],"erc4")
    @erc5 = quick_erc([L,L,W],"erc5")
    #
    @ct1 = Comparative_tableau.new("CT1") << @erc11 << @erc12
    @ct2 = Comparative_tableau.new("CT2") << @erc21 << @erc22
    @ct3 = Comparative_tableau.new("CT3") << @erc12 << @erc21
    @ct4 = Comparative_tableau.new("CT4") << @erc4
    @ct5 = Comparative_tableau.new("CT5") << @erc5
  end

  def teardown
    # teardown code here
  end

  def test_join
    assert(@join11_12.eql?(Mojo.join(@ct1)), 'The join should be equivalent.')
    assert(@join21_22.eql?(Mojo.join(@ct2)), 'The join should be equivalent.')
    assert(@join12_21.eql?(Mojo.join(@ct3)), 'The join should be equivalent.')
  end
  
  def test_triv_valid
    assert(!quick_erc([W,L,E]).triv_valid?)
    assert(quick_erc([W,E,W]).triv_valid?)
    assert(quick_erc([E,E,E]).triv_valid?)
    assert(quick_erc([E,E,W]).triv_valid?)
  end
  
  def test_triv_invalid
    assert(!quick_erc([W,L,E]).triv_invalid?)
    assert(!quick_erc([W,W,E]).triv_invalid?)
    assert(quick_erc([L,E,L]).triv_invalid?)
    assert(quick_erc([L,L,L]).triv_invalid?)
    assert(!quick_erc([E,E,E]).triv_invalid?)
  end
  
  def test_run_mojo
    mojo1 = Mojo.new([Fred.new(@ct1), Fred.new(@ct2)])    
    fred1 = mojo1.fred_result
    
    assert_equal(4, mojo1.fc_product.size)
    assert_equal(4, mojo1.fc_product_ct.size)
    assert(quick_erc([W,L,L]).eql?(fred1.mib[0]))

    mojo2 = Mojo.new([Fred.new(@ct4), Fred.new(@ct5)])    
    fred2 = mojo2.fred_result
    
    assert_equal(1, mojo2.fc_product.size)
    assert_equal(1, mojo2.fc_product_ct.size)
    assert(quick_erc([W,L,W]).eql?(fred2.mib[0]))
  end

end # class Test_mojo
