# Author: Bruce Tesar
# 

require 'test/unit'
require_relative 'quick_erc'
require_relative '../lib/constraint'

class Test_QuickErc < Test::Unit::TestCase

  # Bring in the quick_erc method and related constants.
  include QuickErc

  def setup
  end

  def test_quick_erc_object
    erc1 = quick_erc([W, L, E], "erc1")
    assert(erc1.kind_of?(Erc))
  end

  def test_constraint_list
    erc1 = quick_erc([W, L, E], "erc1")
    assert(C1 == erc1.constraint_list[0])
    assert(C2 == erc1.constraint_list[1])
    assert(C3 == erc1.constraint_list[2])
  end

  def test_constraint_eval_access
    erc1 = quick_erc([W, L, E], "erc1")
    assert(erc1.w?(C1))
    assert(!erc1.l?(C1))
    assert(!erc1.e?(C1))
    
    assert(erc1.l?(C2))
    assert(!erc1.w?(C2))
    assert(!erc1.e?(C2))

    assert(erc1.e?(C3))
    assert(!erc1.w?(C3))
    assert(!erc1.l?(C3))
  end

  def test_eql
    # equiv if same evals, even if labels are different.
    erc1 = quick_erc([L, E, W], "erc1")
    erc1a = quick_erc([L, E, W], "erc1a")
    assert(erc1.eql?(erc1a))
    # not equiv if the evals are different
    erc2 = quick_erc([E, W, L], "erc2")
    assert(!erc1.eql?(erc2))
    erc3 = quick_erc([E, W, E], "erc3")
    assert(!erc3.eql?(erc2))
    # Make sure the presence of a label isn't considered by #eql?().
    erc1_nolabel = quick_erc([L,E,W])
    assert(erc1_nolabel.eql?(erc1))
  end

  def test_invalid_eval_object
    assert_raise(RuntimeError) {quick_erc([L, E, "$"])}
    assert_nothing_raised(RuntimeError) {quick_erc([L, E, "W"])}
  end

end
