# Author: Bruce Tesar
# 

require 'test/unit'
require_relative 'quick_erc'
require_relative '../lib/ranking_diagram'
require_relative '../lib/comparative_tableau'
require_relative '../lib/erc'

class Test_RankingDiagram < Test::Unit::TestCase
  include QuickErc
  
  def setup
    
  end
  
  def test_total_order
    graph_name = 'total_order_test'
    tempfile = File.join(File.dirname(__FILE__),'..','temp',graph_name)
    ct = Comparative_tableau.new
    ct << quick_erc([W,L,E])
    ct << quick_erc([E,W,L])
    rd = RankingDiagram.new(ct, graph_name)
    rd.png(tempfile)
  end

  def test_no_order
    graph_name = 'no_order_test'
    tempfile = File.join(File.dirname(__FILE__),'..','temp',graph_name)
    ct = Comparative_tableau.new
    ct << quick_erc([E,E,E])
    rd = RankingDiagram.new(ct, graph_name)
    rd.png(tempfile)
  end

  def test_two_part_order
    graph_name = 'two_part_order_test'
    tempfile = File.join(File.dirname(__FILE__),'..','temp',graph_name)
    ct = Comparative_tableau.new
    ct << quick_erc([W,L,E,E,E])
    ct << quick_erc([E,E,L,W,L])
    rd = RankingDiagram.new(ct, graph_name)
    rd.png(tempfile)
  end

  def test_multi_rank
    graph_name = 'multi_rank_test'
    tempfile = File.join(File.dirname(__FILE__),'..','temp',graph_name)
    ct = Comparative_tableau.new
    ct << quick_erc([W,L,E,E,E])
    ct << quick_erc([E,W,E,E,L])
    ct << quick_erc([E,E,L,W,E])
    rd = RankingDiagram.new(ct, graph_name)
    rd.png(tempfile)
  end

  def test_multi_parent
    graph_name = 'multi_parent_test'
    tempfile = File.join(File.dirname(__FILE__),'..','temp',graph_name)
    ct = Comparative_tableau.new
    ct << quick_erc([W,L,E,E,E])
    ct << quick_erc([E,L,E,E,W])
    ct << quick_erc([E,E,L,W,E])
    rd = RankingDiagram.new(ct, graph_name)
    rd.png(tempfile)
  end

  def test_diffrank_parent
    graph_name = 'diffrank_parent_test'
    tempfile = File.join(File.dirname(__FILE__),'..','temp',graph_name)
    ct = Comparative_tableau.new
    ct << quick_erc([W,L,E,E,E,E])
    ct << quick_erc([E,W,L,E,E,E])
    ct << quick_erc([E,E,L,W,L,E])
    rd = RankingDiagram.new(ct, graph_name)
    rd.png(tempfile)
  end

  def test_spaced_constraint_names
    graph_name = 'spaced_constraint_names_test'
    tempfile = File.join(File.dirname(__FILE__),'..','temp',graph_name)

    constraints = Array.new
    erc = Erc.new(constraints, graph_name+"_ERC1")

    con = Constraint.new("Con:1", 1)
    constraints << con
    erc.set_w(con)

    con = Constraint.new("Con 2 Left", 2)
    constraints << con
    erc.set_l(con)

    con = Constraint.new("Con 3 Right", 3)
    constraints << con
    erc.set_e(con)

    ct = Comparative_tableau.new
    ct << erc
    rd = RankingDiagram.new(ct, graph_name)
    rd.png(tempfile)
  end

end
