# Author: Bruce Tesar


require 'test/unit'
require_relative '../lib/rcd'
require_relative '../lib/comparative_tableau'
require_relative '../lib/competition_list'

class Test_rcd < Test::Unit::TestCase
  def setup
    @win1 = Candidate.new("in1", "out1", true, ["C1", "C2", "C3"])
    @win1.label = "1a"
    @win1.set_viols("C1", 0)
    @win1.set_viols("C2", 3)
    @win1.set_viols("C3", 2)

    @win1_alt = Candidate.new("in1", "out6", true, ["C1", "C2", "C3"])
    @win1_alt.label = "1f"
    @win1_alt.set_viols("C1", 0)
    @win1_alt.set_viols("C2", 2)
    @win1_alt.set_viols("C3", 3)

    @los1_2 = Candidate.new("in1", "out2", false, ["C1", "C2", "C3"])
    @los1_2.label = "1b"
    @los1_2.set_viols("C1", 1)
    @los1_2.set_viols("C2", 1)
    @los1_2.set_viols("C3", 2)

    @los1_3 = Candidate.new("in1", "out3", false, ["C1", "C2", "C3"])
    @los1_3.label = "1c"
    @los1_3.set_viols("C1", 2)
    @los1_3.set_viols("C2", 1)
    @los1_3.set_viols("C3", 1)

    @comp1 = Competition.new
    @comp1.push(@win1)
    @comp1.push(@los1_2)

    # Competition 2
    @win2 = Candidate.new("in2", "out1", true, ["C1", "C2", "C3"])
    @win2.label = "2a"
    @win2.set_viols("C1", 0)
    @win2.set_viols("C2", 3)
    @win2.set_viols("C3", 2)

    @los2_2 = Candidate.new("in2", "out2", false, ["C1", "C2", "C3"])
    @los2_2.label = "2b"
    @los2_2.set_viols("C1", 1)
    @los2_2.set_viols("C2", 1)
    @los2_2.set_viols("C3", 2)

    @los2_3 = Candidate.new("in2", "out3", false, ["C1", "C2", "C3"])
    @los2_3.label = "2c"
    @los2_3.set_viols("C1", 2)
    @los2_3.set_viols("C2", 4)
    @los2_3.set_viols("C3", 1)

    @comp2 = Competition.new
    @comp2.push(@win2)
    @comp2.push(@los2_2)
    @comp2.push(@los2_3)
    
    # competition list
    @clist1 = Competition_list.new
    @clist1.label = "rcd test 1"
    @clist1 << @comp1 << @comp2

    # comparative tableau 
    @ct = Comparative_tableau.new()
    @ct.add_competition_list(@clist1)
  end

  def teardown
    # teardown code here
  end

  def test_rcd_consistent
    rcd_result = Rcd.new(@ct)
    assert_equal("NoLabel", rcd_result.label)
    assert(rcd_result.consistent?, "This ERC set should be consistent.")
    hier_expect = [["C1"],["C2", "C3"]]
    assert_equal(hier_expect, rcd_result.hierarchy)
    unranked_expect = []
    assert_equal(unranked_expect, rcd_result.unranked)
    remaining_expect = []
    assert_equal(remaining_expect, rcd_result.unex_ercs)
  end

  def test_additional_erc
    erc_new = Erc.new(@ct.constraint_list, "incon")
    erc_new.set_e("C1"); erc_new.set_w("C2"); erc_new.set_l("C3")
    @ct << erc_new
    @ct.label = "test_additional_erc"
    rcd_result = Rcd.new(@ct)
    assert_equal("test_additional_erc", rcd_result.label)
    assert(rcd_result.consistent?, "This ERC set should be consistent.")
    hier_expect = [["C1"],["C2"],["C3"]]
    assert_equal(hier_expect, rcd_result.hierarchy)
    unranked_expect = []
    assert_equal(unranked_expect, rcd_result.unranked)
    remaining_expect = []
    assert_equal(remaining_expect, rcd_result.unex_ercs)
  end

  def test_inconsistent
    erc_new = Erc.new(@ct.constraint_list, "incon1")
    erc_new.set_e("C1"); erc_new.set_w("C2"); erc_new.set_l("C3")
    @ct << erc_new
    remaining_expect = Comparative_tableau.new("RCD remaining") << erc_new
    erc_new = Erc.new(@ct.constraint_list, "incon2")
    erc_new.set_e("C1"); erc_new.set_l("C2"); erc_new.set_w("C3")
    @ct << erc_new
    remaining_expect << erc_new
    rcd_result = Rcd.new(@ct)
    assert(!rcd_result.consistent?, "This ERC set should be INconsistent.")
    hier_expect = [["C1"]]
    assert_equal(hier_expect, rcd_result.hierarchy)
    unranked_expect = ["C2","C3"]
    assert_equal(unranked_expect, rcd_result.unranked)
    assert_equal(remaining_expect, rcd_result.unex_ercs)    
  end

  def test_alle_inconsistent
    erc_new = Erc.new(@ct.constraint_list, "incon1")
    erc_new.set_e("C1"); erc_new.set_w("C2"); erc_new.set_l("C3")
    @ct << erc_new
    remaining_expect = Comparative_tableau.new("RCD remaining") << erc_new
    erc_new = Erc.new(@ct.constraint_list, "incon2")
    erc_new.set_e("C1"); erc_new.set_l("C2"); erc_new.set_w("C3")
    @ct << erc_new
    remaining_expect << erc_new
    erc_new = Erc.new(@ct.constraint_list, "alle")
    erc_new.set_e("C1"); erc_new.set_e("C2"); erc_new.set_e("C3")
    @ct << erc_new
    remaining_expect << erc_new
    rcd_result = Rcd.new(@ct)
    assert(!rcd_result.consistent?, "This ERC set should be INconsistent.")
    hier_expect = [["C1"]]
    assert_equal(hier_expect, rcd_result.hierarchy)
    unranked_expect = ["C2","C3"]
    assert_equal(unranked_expect, rcd_result.unranked)
    assert_equal(remaining_expect, rcd_result.unex_ercs)    
  end
  
end
