# Author: Bruce Tesar
# 

require 'test/unit'
require 'shoulda'
require_relative '../lib/skb_expansion'
require_relative 'quick_erc'
require_relative '../lib/fred'

class Test_skb_expansion < Test::Unit::TestCase

  include QuickErc

  context "A single ERC with one L and one W" do
    setup do
      @skb = [quick_erc([W,E,L])]
      @expansion = Skb_expansion.new(@skb)
    end
    should "have a conjunctive expansion with that same single erc" do
      ce = @expansion.conj_expansion
      assert_equal(1, ce.size)
      assert_equal([quick_erc([W,E,L])], ce)
    end
    should "have a single cover" do
      cover_groups = @expansion.cover_groups
      assert_equal(1, cover_groups.size) # only one L
      group3 = cover_groups[0]
      assert_equal(1, group3.size) # only one cover
      cover_3_0 = group3[0]
      assert_equal(C3, cover_3_0.l_constraint)
      assert_equal([C1], cover_3_0.w_constraints)
      assert_equal([quick_erc([W,E,L])], cover_3_0.reduced_ercs)
    end
    should "have a single diagram basis [[W,E,L]]" do
      assert_equal(1,@expansion.diagram_bases.size)
      basis = @expansion.diagram_bases[0]
      assert_equal([quick_erc([W,E,L])],basis)
    end
    #TODO: drop this test, remove outside access to cartesian product.
    should "have a cartesian product of size one" do
      cp = @expansion.get_cart_prod
      assert_equal(Array, cp[0].class)
      assert_equal(W_cover, cp[0][0].class)
    end
  end

  context "A single ERC with two Ls and one W" do
    setup do
      @skb = [quick_erc([W,L,L])]
      @expansion = Skb_expansion.new(@skb)
    end
    should "have a conjunctive expansion with two ercs" do
      ce = @expansion.conj_expansion
      assert_equal(2, ce.size)
      assert(ce.find{|e| e==quick_erc([W,L,E])})
      assert(ce.find{|e| e==quick_erc([W,E,L])})
    end
    should "have two groups, each with one cover" do
      cover_groups = @expansion.cover_groups
      assert_equal(2, cover_groups.size) # two L's
      #
      group2 = cover_groups[0]
      assert_equal(1,group2.size)
      cover_2_0 = group2[0]
      w_covers_group2 = group2.map{|cov| cov.w_constraints}
      assert_equal([[C1]], w_covers_group2) # only one W, C1
      assert_equal([quick_erc([W,L,E])], cover_2_0.reduced_ercs)
      #
      group3 = cover_groups[1]
      assert_equal(1,group3.size)
      cover_3_0 = group3[0]
      w_covers_group3 = group3.map{|cov| cov.w_constraints}
      assert_equal([[C1]], w_covers_group3) # only one W, C1
      assert_equal([quick_erc([W,E,L])], cover_3_0.reduced_ercs)
    end
    should "have a single diagram basis with one erc" do
      assert_equal(1,@expansion.diagram_bases.size)
      basis = @expansion.diagram_bases[0]
      assert_equal([quick_erc([W,L,L])],basis)
    end
  end

  context "An skb [W,W,L,L]" do
    setup do
      @skb = [quick_erc([W,W,L,L])]
      @expansion = Skb_expansion.new(@skb)
    end
    should "have two cover groups, each with two covers" do
      cover_groups = @expansion.cover_groups
      assert_equal(2, cover_groups.size)
      group3 = cover_groups[0]
      assert_equal(2, group3.size)
      group4 = cover_groups[1]
      assert_equal(2, group4.size)
    end
    should "have reduced erc groups [[W,E,L,E],[E,W,L,E]] and [[W,E,E,L],[E,W,E,L]]" do
      cover_groups = @expansion.cover_groups
      #
      group3 = cover_groups[0]
      cover_3_0 = group3[0]
      cover_3_1 = group3[1]
      assert_equal([quick_erc([W,E,L,E])], cover_3_0.reduced_ercs)
      assert_equal([quick_erc([E,W,L,E])], cover_3_1.reduced_ercs)
      #
      group4 = cover_groups[1]
      cover_4_0 = group4[0]
      cover_4_1 = group4[1]
      assert_equal([quick_erc([W,E,E,L])], cover_4_0.reduced_ercs)
      assert_equal([quick_erc([E,W,E,L])], cover_4_1.reduced_ercs)
    end
    should "have ranking erc groups [[W,L,E,E],[L,W,E,E]] and [[W,L,E,E],[L,W,E,E]]" do
      cover_groups = @expansion.cover_groups
      #
      group3 = cover_groups[0]
      cover_3_0 = group3[0]
      cover_3_1 = group3[1]
      assert_equal([quick_erc([W,L,E,E])], cover_3_0.ranking_ercs)
      assert_equal([quick_erc([L,W,E,E])], cover_3_1.ranking_ercs)
      #
      group4 = cover_groups[1]
      cover_4_0 = group4[0]
      cover_4_1 = group4[1]
      assert_equal([quick_erc([W,L,E,E])], cover_4_0.ranking_ercs)
      assert_equal([quick_erc([L,W,E,E])], cover_4_1.ranking_ercs)
    end
    should "have two diagram bases" do
      assert_equal(2,@expansion.diagram_bases.size)
      basis1 = @expansion.diagram_bases[0]
      assert_equal([quick_erc([W,E,L,L])],basis1)
      basis2 = @expansion.diagram_bases[1]
      assert_equal([quick_erc([E,W,L,L])],basis2)
    end
  end

  context "An skb [W,E,W,L],[W,W,E,L]" do
    setup do
      @skb = [quick_erc([W,E,W,L])]
      @skb << quick_erc([W,W,E,L])
      @expansion = Skb_expansion.new(@skb)
    end
    should "have two diagram bases" do
      assert_equal(2,@expansion.diagram_bases.size)
      basis1 = @expansion.diagram_bases[0]
      assert_equal([quick_erc([W,E,E,L])],basis1)
      basis2 = @expansion.diagram_bases[1]
      assert_equal([quick_erc([E,W,E,L]),quick_erc([E,E,W,L])],basis2)
    end
  end

  context "An skb [W,E,W,L],[E,W,L,W]" do
    setup do
      @skb = [quick_erc([W,E,W,L])]
      @skb << quick_erc([E,W,L,W])
      @expansion = Skb_expansion.new(@skb)
    end
    should "have three diagram bases" do
      assert_equal(3,@expansion.diagram_bases.size)      
      basis1 = @expansion.diagram_bases[0]
      assert_equal([quick_erc([W,E,E,L]),quick_erc([E,W,L,E])],basis1)
      basis2 = @expansion.diagram_bases[1]
      assert_equal([quick_erc([W,E,E,L]),quick_erc([E,E,L,W])],basis2)
      basis3 = @expansion.diagram_bases[2]
      assert(basis3.member?(quick_erc([E,E,W,L])))
      assert(basis3.member?(quick_erc([E,W,L,E])))
    end
  end

  context "An skb [W,E,L,L],[E,W,W,L]" do
    setup do
      @skb = [quick_erc([W,E,L,L])]
      @skb << quick_erc([E,W,W,L])
      @expansion = Skb_expansion.new(@skb)
    end
    should "have three diagram bases" do
      assert_equal(2,@expansion.diagram_bases.size)      
      basis1 = @expansion.diagram_bases[0]
      assert_equal([quick_erc([W,E,L,L]),quick_erc([E,W,E,L])],basis1)
      basis2 = @expansion.diagram_bases[1]
      assert_equal([quick_erc([W,E,L,E]),quick_erc([E,E,W,L])],basis2)
    end
  end

  #**************************************
  # Tests specific to #count_and_sort_w()

  context "An erc group [W E L][W W L]" do
    setup do
      @ercs = [quick_erc([W,E,L]),quick_erc([W,W,L])]
      @constraint_list = @ercs.first.constraint_list
      @expansion = Skb_expansion.new([quick_erc([E,E,E])]) # dummy to allow instance method call
    end
    should "have W list [C1, C2]" do
      w_count_sorted = @expansion.count_and_sort_w(@ercs,@constraint_list)
      assert_equal([C1,C2],w_count_sorted)
    end
  end

  context "Erc group [W E E L W]" do
    setup do
      @ercs = [quick_erc([W,E,E,L,W])]
      @constraint_list = @ercs.first.constraint_list
      @expansion = Skb_expansion.new([quick_erc([E,E,E,E,E])]) # dummy to allow instance method call
    end
    should "have W list [C1,C5]" do
      w_count_sorted = @expansion.count_and_sort_w(@ercs,@constraint_list)
      assert_equal([C1,C5],w_count_sorted)
    end
    context "& [E W E L W]" do
      setup do
        @ercs << quick_erc([E,W,E,L,W])
      end
      should "have W list [C5, C1, C2]" do
        w_count_sorted = @expansion.count_and_sort_w(@ercs,@constraint_list)
        assert_equal([C5,C1,C2],w_count_sorted)
      end
      context "& [E W W L E]" do
        setup do
          @ercs << quick_erc([E,W,W,L,E])
        end
        should "have W list [C2,C5,C1,C3]" do
          w_count_sorted = @expansion.count_and_sort_w(@ercs,@constraint_list)
          assert_equal([C2,C5,C1,C3],w_count_sorted)
        end
      end
    end
  end

  #*********************************************
  # Tests specific to #check_possible_w_covers()

  context "An empty set of uncovered ercs" do
    setup do
      @uncovered_ercs = []
      @w_cons = []
      @expansion = Skb_expansion.new([]) # dummy to allow instance method call
    end
    should "have a single, empty cover" do
      covers = @expansion.check_possible_w_covers(W_cover.new, @uncovered_ercs, @w_cons, [])
      assert_equal(1,covers.size)
      assert(covers[0].w_constraints.empty?)
    end
  end

  context "Ercs [W E E L]" do
    setup do
      @uncovered_ercs = [quick_erc([W,E,E,L])]
      @w_cons = [C1]
      @expansion = Skb_expansion.new([]) # dummy to allow instance method call
    end
    should "have a single cover [C1]" do
      covers = @expansion.check_possible_w_covers(W_cover.new, @uncovered_ercs, @w_cons, [])
      assert_equal(1,covers.size)
      assert_equal([C1],covers[0].w_constraints)
    end
    context "& [E E W L]" do
      setup do
        @uncovered_ercs << quick_erc([E,E,W,L])
        @w_cons = [C1,C3]
      end
      should "have one cover [[C1,C3]]" do
        covers = @expansion.check_possible_w_covers(W_cover.new, @uncovered_ercs, @w_cons, [])
        assert_equal(1,covers.size)
        assert_equal([C1,C3],covers[0].w_constraints)
      end
      context "& [W W E L]" do
        setup do
          @uncovered_ercs << quick_erc([W,W,E,L])
          @w_cons = [C1,C2,C3]
        end
        should "have one cover [[C1,C3]]" do
          covers = @expansion.check_possible_w_covers(W_cover.new, @uncovered_ercs, @w_cons, [])
          assert_equal(1,covers.size)
          assert_equal([C1,C3],covers[0].w_constraints)
        end
      end
    end
    context "& [E W W L]" do
      setup do
        @uncovered_ercs << quick_erc([E,W,W,L])
        @w_cons = [C1,C2,C3]
      end
      should "have two covers [C1,C2] and [C1,C3]" do
        covers = @expansion.check_possible_w_covers(W_cover.new, @uncovered_ercs, @w_cons, [])
        assert_equal(2,covers.size)
        assert_equal([C1,C2],covers[0].w_constraints)
        assert_equal([C1,C3],covers[1].w_constraints)
      end
    end
  end

  context "Ercs [[E,W,W,L],[E,E,W,L],[W,W,E,L],[W,E,E,L]]" do
    setup do
      @uncovered_ercs = []
      @uncovered_ercs << quick_erc([E,W,W,L])
      @uncovered_ercs << quick_erc([E,E,W,L])
      @uncovered_ercs << quick_erc([W,W,E,L])
      @uncovered_ercs << quick_erc([W,E,E,L])
      @w_cons = [C1,C2,C3]
      @expansion = Skb_expansion.new([]) # dummy to allow instance method call
    end
    should "yield only cover [C1,C3]" do
      covers = @expansion.check_possible_w_covers(W_cover.new, @uncovered_ercs, @w_cons, [])
      assert_equal(1,covers.size)
      assert_equal([C1,C3],covers[0].w_constraints)
    end
  end

end
