# Author: Bruce Tesar
#
# In Ruby 1.9, SyncEnumerator was moved to REXML (I don't know why).
# These tests are to ensure that the new version behaves the way I
# expect it to.

require 'test/unit'
require 'REXML/syncenumerator'

class Test_syncenumerator < Test::Unit::TestCase
  def test_sync_equal
    arr1 = [1,2,3]
    arr2 = ['a','b','c']
    count = 0
    letter = 'a'
    s = REXML::SyncEnumerator.new(arr1, arr2)
    s.each {|a1, a2| count+=1; assert_equal(count,a1); assert_equal(letter, a2); letter=letter.next}
    assert_equal(3,count)
    assert_equal('d',letter)
  end

  def test_first_smaller
    arr1 = [1,2,3,4]
    arr2 = ['a','b','c','d','e']
    count = 0
    letter = 'a'
    s = REXML::SyncEnumerator.new(arr1, arr2)
    s.each do |a1, a2|
      count+=1
      if count > 4
        assert_nil(a1)
      else
        assert_equal(count,a1)
      end
      assert_equal(letter, a2)
      letter=letter.next
    end
    assert_equal(5,count)
  end

  def test_first_larger
    arr1 = [1,2,3,4,5]
    arr2 = ['a','b']
    count = 0
    letter = 'a'
    s = REXML::SyncEnumerator.new(arr1, arr2)
    s.each do |a1, a2|
      count+=1
      assert_equal(count,a1)
      if count > 2
        assert_nil(a2)
      else
        assert_equal(letter, a2)
        letter=letter.next
      end
    end
    assert_equal(5,count)
    assert_equal('c',letter)
  end

  def test_four
    arr1 = [1,2,3]
    arr2 = ['a','b']
    arr3 = []
    arr4 = [7,8,9]
    count = 0
    s = REXML::SyncEnumerator.new(arr1,arr2,arr3,arr4)
    s.each do |enums|
      count += 1
      case count
      when 1 then assert_equal([1,'a',nil,7],enums)
      when 2 then assert_equal([2,'b',nil,8],enums)
      when 3 then assert_equal([3,nil,nil,9],enums)
      else        raise "Illegal count: none of the containers is this long."
      end
    end
  end

end # class Test_SyncEnumerator