# Author: Bruce Tesar
 

require 'test/unit'
require_relative '../lib/win_lose_pair'

class Test_win_lose_pair < Test::Unit::TestCase
  def setup
    @winner = Candidate.new("in4", "out4", true, ["C1", "C2", "C3"])
    @winner.label = "4.2"
    @winner.set_viols("C1", 1)
    @winner.set_viols("C2", 4)
    @winner.set_viols("C3", 8)

    @loser = Candidate.new("in4", "out6", false, ["C1", "C2", "C3"])
    @loser.label = "4.6"
    @loser.set_viols("C1", 5)
    @loser.set_viols("C2", 2)
    @loser.set_viols("C3", 8)
    
    @wl1 = Win_lose_pair.new(@winner, @loser)
    
    # Reverse winner and loser, for testing eql?
    @loser_rev = Candidate.new("in4", "out4", false, ["C1", "C2", "C3"])
    @loser_rev.label = "4.2_rev"
    @loser_rev.set_viols("C1", 1)
    @loser_rev.set_viols("C2", 4)
    @loser_rev.set_viols("C3", 8)

    @winner_rev = Candidate.new("in4", "out6", true, ["C1", "C2", "C3"])
    @winner_rev.label = "4.6_rev"
    @winner_rev.set_viols("C1", 5)
    @winner_rev.set_viols("C2", 2)
    @winner_rev.set_viols("C3", 8)
    
    @wl1_rev = Win_lose_pair.new(@winner_rev, @loser_rev)
    
    #
    @winner2 = Candidate.new("in42", "out42", true, ["C1", "C2", "C3"])
    @winner2.label = "42.2"
    @winner2.set_viols("C1", 1)
    @winner2.set_viols("C2", 4)
    @winner2.set_viols("C3", 8)

    @loser2 = Candidate.new("in42", "out62", false, ["C1", "C2", "C3"])
    @loser2.label = "42.6"
    @loser2.set_viols("C1", 5)
    @loser2.set_viols("C2", 2)
    @loser2.set_viols("C3", 8)
    
    @wl2 = Win_lose_pair.new(@winner2, @loser2)

  end

  def teardown
    # teardown code here
  end

  def test_matching_inputs
    loser2 = Candidate.new("inx", "out6", false, ["C1", "C2", "C3"])    
    assert_raise(RuntimeError){Win_lose_pair.new(@winner, loser2)}
  end
  
  def test_pref_queries
    assert(@wl1.w?("C1"), "C1 should prefer the winner.")
    assert(!@wl1.l?("C1"), "C1 should not prefer the loser.")
    assert(!@wl1.e?("C1"), "C1 should not be e.")
    
    assert(!@wl1.w?("C2"), "C2 should not prefer the winner.")
    assert(@wl1.l?("C2"), "C2 should prefer the loser.")
    assert(!@wl1.e?("C2"), "C2 should not be e.")

    assert(!@wl1.w?("C3"), "C3 should not prefer the winner.")
    assert(!@wl1.l?("C3"), "C3 should not prefer the loser.")
    assert(@wl1.e?("C3"), "C3 should be e.")
  end

  def test_label
    assert_equal("4.2>6", @wl1.label)
    @loser.label = "55"
    wl2 = Win_lose_pair.new(@winner, @loser)
    assert_equal("NoLabel", wl2.label)
  end

  def test_constraints
    assert_equal(["C1", "C2", "C3"], @wl1.constraint_list)
  end
  
  def test_to_s
    expected = "4.2>6 in4 out4 out6 C1:W C2:L C3:e"
    assert_equal(expected, @wl1.to_s)
  end
  
  def test_equiv
    assert(@wl1.eql?(@wl1))
    assert(!@wl1.eql?(@wl1_rev))
    assert(@wl1.eql?(@wl2))
    assert(@wl2.eql?(@wl1))
  end
  
end # class Test_win_lose_pair
