# Author: Naz Merchant
#


require_relative '../../lib/constraint_hs.rb'
require_relative '../../lib/candidate.rb'

class Constraint_HS_AFR < Constraint_HS
  
  def initialize(name, id, label)
    super(name, id, Regexp.new(label))
    
    # Note that a Constraint_HS uses 'label' as a regular expression to match
    # the output of 'candidate' in the method eval. Presumably this obj won't be 
    # doing that.
  end
  
  def eval(candidate)
    out_n = candidate.output.reverse
    # Only faithful segs
    out_n = out_n.gsub("uX", "Xu")
    out_n = out_n.gsub("uH", "Hu")
    out_n = out_n.gsub("hX", "Xh")
    out_n = out_n.gsub("hH", "Hh") 
    
    # With unfaithful heads only
    out_n = out_n.gsub("uY", "Yu") 
    out_n = out_n.gsub("hY", "Yh") 
    out_n = out_n.gsub("uK", "Ku") 
    out_n = out_n.gsub("hK", "Kh") 

    # With unfaithful non-heads only    
    out_n = out_n.gsub("yX", "Xy")
    out_n = out_n.gsub("yH", "Hy")
    out_n = out_n.gsub("kX", "Xk")
    out_n = out_n.gsub("kH", "Hk") 
    
    # With both head & non-head unfaithful
    out_n = out_n.gsub("yY", "Yy") 
    out_n = out_n.gsub("kY", "Yk") 
    out_n = out_n.gsub("yK", "Ky") 
    out_n = out_n.gsub("kK", "Kk") 
    
    out_n = out_n.gsub(/[^a-zA-Z]/, "")

    viols = 0
    (0..out_n.length-1).each {|i| viols += i if out_n[i] =~ /[A-Z]/}

    return viols
  end
end

=begin
input = "in"
output = ".-H--Xu-."
con = Constraint_HS_AFR.new("t", "t", "t")
viols = con.eval(Candidate.new(input, output, false, nil))

puts "Output: \t#{output}"
puts "Viols:  \t#{viols}"
=end



