# Author: Naz Merchant
#

# A basic HS constraint, consisting of a name, an id, and a regular expression object
# that allows for evaluation of violations of a candidate.

# Constraint_HS subclasses Constraint, a basic OT constraint. The regex is added
# since candidates in HS can be created on the fly in Ruby contra what happens in the OT system.

require_relative '../../lib/candidate.rb'
require_relative '../../lib/constraint.rb'

require_relative 'bst_parser.rb'

class Constraint_BST_HS < Constraint_HS
  
  # Returns a constraint with the given name and id.
  # The id is an abbreviated label used for constructing labels for
  # complex objects, e.g., residues in *FRed*.
  def initialize(name, id, regex, parser, incl_ineff=true)
    super(name, id, regex)
    @parser = parser
    @incl_ineff = incl_ineff
  end
  
  def eval(candidate)
    return super(candidate) if @incl_ineff
    return @parser.remove_ineffables(candidate.output).scan(@regex).length
  end
end